// APIFont.cpp: implementation of the CAPIFont class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "APIFont.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

#include "..\api3d\APITexture.h"
#include "..\globals.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CAPIFont::CAPIFont()
{
	texture = NULL;
	strcpy(filename, "");
	valid = false;
}

CAPIFont::~CAPIFont()
{
	Destroy();
	if (charData)
		delete [] charData;
}

void CAPIFont::Destroy(void)
{
	// This might cause a problem on program end
	// ??
	APITextures.DeleteTexture(texture);
}

void CAPIFont::Initialize(char *p_filename)
{
	// Immediately load in the binary data for the
	//   font and prepare the data for the texture, 
	//   don't make the texture until it's needed
	strcpy(filename, p_filename);

	FILE *fp;
	fp = fopen(filename, "rb");
	if (fp)
	{
		char fileID[9];
		fseek(fp, 0, SEEK_SET);
		fread(fileID, 8, 1, fp);
		fileID[8] = '\0';
		if (strcmp(fileID, "GEFONT01"))
		{
			TRACE("* Warning: Binary font file %s is not of the custom format\n", filename);
		}
		else
		{
			fseek(fp, 8, SEEK_SET);
			fread(&height, 4, 1, fp);
			fread(&width, 4, 1, fp);
			fread(charValid, 256, 1, fp);
			fread(charWidth, 256, 1, fp);
			charData = new BYTE[width*height];
			fread(charData, width*height, 1, fp);

			texture = APITextures.CreateNewTexture();

			// Allow texture to load data
			texture->SetTexture(filename, filename, APITEXTUREFORMAT_ALPHA8, false, APITEXTUREFILEFORMAT_FONT);
			texture->LoadFontIntoMemory(this);

			valid = true;
		}

		fclose (fp);
	}
	else
		TRACE ("* Warning: Failed to open font file %s", filename);

}

unsigned CAPIFont::GetTextWidth(char *p_text)
{
    static unsigned totalWidth;
    static BYTE chr;

	if (!p_text) return 0;

    unsigned i;
    totalWidth = 0;
    
    for (i=1; i<=strlen(p_text); i++)
	{
        chr = p_text[i-1];
        if (charValid[chr])
            totalWidth = totalWidth + charWidth[chr];
    }
    
	return totalWidth;
}

void CAPIFont::GetTextDimensions(char *p_text, unsigned &p_widthOut, unsigned &p_heightOut)
{
    p_widthOut = GetTextWidth(p_text);
    p_heightOut = height;
}


