// APITexture.cpp: implementation of the CAPITexture class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "APITexture.h"
#include "..\globals.h"
#include <math.h> // sqrt
#include "APIFont.h"

//#include <stdio.h>	// FILE, fread, etc.

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CAPITexture::CAPITexture()
{
	textureData = NULL;
	valid = true;
	loaded = false;

	name[0] = '\0';
	filename[0] = '\0';

	GLid = 0;
	DXid = NULL;

	width = 0;
	height = 0;
	widthf = 0.0f;
	heightf = 0.0f;
	textureData = NULL;
	mipmap = false;
	forceClamp = false;

	lightmap = false;
	renderBuffer = false;
	font = false;

	update = false;

	inList = false;
}

CAPITexture::~CAPITexture()
{
	Destroy(Render->GetAPIType());
}

void CAPITexture::Destroy(defRenderAPIType p_APIType)
{
	ReleaseAPI(p_APIType);
	DeallocateMemory();
}

void CAPITexture::DeallocateMemory(void)
{
	if (textureData != NULL)
	{
		delete [] textureData;
		textureData = NULL;

		if (lightmap)
			Metrics.textureLMMemory -= textureSysMemorySize;
		else
			Metrics.textureMemory -= textureSysMemorySize;
		textureSysMemorySize = 0;

	}
}

void CAPITexture::ReleaseAPI(defRenderAPIType p_APIType)
{
	switch (p_APIType)
	{
	case APITYPE_OPENGL:
		if (GLid != 0)
		{
			glDeleteTextures(1, &GLid);
			GLid = 0;
			if (lightmap)
			{
				Metrics.textureLMQty--;
			}
			else
			{
				Metrics.textureQty--;
			}
		}
		break;
	case APITYPE_DIRECTX8:
		if (DXid != NULL)
		{
			DXid->Release();
			DXid = NULL;
		}
		break;
	}

	loaded = false;
}

bool CAPITexture::LoadIntoMemory(void)
{
	//BYTE *row = NULL;	// Single pixel row of data from file
	static BYTE row[4096];
	long start;			// Beginning of data

	bool result = false;

	if (textureData == NULL)
	{
		FILE *fp;

		fp = fopen(filename, "rb");
		if (fp != NULL)
		{
			// Load it in as RGBA

			switch (fileFormat)
			{
			case APITEXTUREFILEFORMAT_BMP:

				// bitmap (.BMP)

				// For a 0-index file (NOT 1 - c++ uses 0, VB's Seek uses 1)
				// Byte 18-21 has width
				// Byte 22-25 has height
				// Byte 10-13 has start of data

				// Get width and height
				fseek(fp, 18, SEEK_SET);
				fread(&width, 4, 1, fp);
				fseek(fp, 22, SEEK_SET);
				fread(&height, 4, 1, fp);
				widthf = float(width);
				heightf = float(height);

				// Allocate storage
				textureData = new BYTE[width*4*height];
				Metrics.textureMemory += (width*4*height);
				textureSysMemorySize = width*4*height;

				// Get start of data
				fseek(fp, 10, SEEK_SET);
				fread (&start, 4, 1, fp);

				// Set up row storage
				//row = new BYTE[width*3];

				// Load it in
				fseek(fp, start, SEEK_SET);
				int x, y;
				for (y=0; y<height; y++)
				{
					fread(row, width*3, 1, fp);

					for (x=0; x<width; x++)
					{
						// In a BMP, the data is stored as BGR
						textureData[x*4+y*width*4] = row[x*3+2];
						textureData[x*4+1+y*width*4] = row[x*3+1];
						textureData[x*4+2+y*width*4] = row[x*3];
						if (format == APITEXTUREFORMAT_ALPHA8)
							textureData[x*4+3+y*width*4] = row[x*3];
						else
							textureData[x*4+3+y*width*4] = 255;
					}
				}

				break;

			}	// switch fileformat

			fclose(fp);

			result = true;
		}
		else
		{
			// Report file failed to load
			TRACE("* Warning: Unable to load texture '%s', filename '%s'\n", name, filename);
		}
	}

	// Cleanup
	//if (row != NULL)
	//	delete [] row;

	return result;
}

void CAPITexture::SetTexture(char *p_name, char *p_filename, defAPITextureFormat p_format, bool p_mipmap, defAPITextureFileFormat p_fileFormat, bool p_forceClamp)
{
	strcpy(name, p_name);
	strcpy(filename, p_filename);

	format = p_format;
	mipmap = p_mipmap;

	fileFormat = p_fileFormat;
	forceClamp = p_forceClamp;
}

void CAPITexture::LoadTextureGL(void)
{
	if (!valid) return;
	if (loaded) return;

	// Load into memory if not already
	if (!IsInMemory())
	{
		if (!LoadIntoMemory())
		{
			valid = false;
			return;
		}
	}

	glGenTextures(1, &GLid);
	Metrics.textureQty++;
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, GLid);

	// Set filter and wrap parameters
    glPixelStorei (GL_PACK_ALIGNMENT, 1);
    // GL_CLAMP is good for singular icons rendered inside a polygon
    //   but bad for a repeating tiled texture across a loarge number of polygons (like terrain)
    glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
    glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
    // GL_NEAREST makes them look pixelated (good for icons, bad for surfaces)
    glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    if (mipmap)
        glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
    else
        glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    // Multiply by color (GL_DECAL ignores polygon colors, and would be unaffected by lighting)
    glTexEnvf (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
    
    GLuint internalformat;
    switch (format)
	{
		case APITEXTUREFORMAT_RGB:
			internalformat = GL_RGB;
			break;
        case APITEXTUREFORMAT_ALPHA8:
            internalformat = GL_ALPHA8;
			break;
	}
    
	// Load it up
    if (!mipmap)
        glTexImage2D (GL_TEXTURE_2D, 0, internalformat, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, textureData);
    else
        gluBuild2DMipmaps (GL_TEXTURE_2D, internalformat, width, height, GL_RGBA, GL_UNSIGNED_BYTE, textureData);

	glDisable(GL_TEXTURE_2D);

	loaded = true;
}

/////////////////////////////////////////////////////////////////
// Lightmap functions

void CAPITexture::ApplyRadiusLight(float p_x, float p_y, float p_radiusMap, float p_baseDistance, CColorRGB p_color)
{
	if (!lightmap) return;
	if (!textureData) return;

	float totalRadius = float(sqrt(pow(p_baseDistance,2) + pow(p_radiusMap,2)));

    // p_x, p_y are in 0.0-1.0
    // p_radius is in Units, p_texelPerUnit converts it
    int xMin, xMax;
    int yMin, yMax;
    int x, y;
    
    float tx, ty;
	float k1, k2;

    int r, g, b;
    float v;
    
    // Adjust downward by 0.5 to put us in the middle of the texel
    tx = float(p_x * width) - 0.5f;
    ty = float(p_y * height) - 0.5f;
    xMin = int(p_x * float(width) - p_radiusMap * texelPerUnit - 0.5f);
    xMax = int(p_x * float(width) + p_radiusMap * texelPerUnit - 0.5f);
    yMin = int(p_y * float(height) - p_radiusMap * texelPerUnit - 0.5f);
    yMax = int(p_y * float(height) + p_radiusMap * texelPerUnit - 0.5f);
    if (xMin < 0) xMin = 0;
    if (xMax > width - 1) xMax = width - 1;
    if (yMin < 0) yMin = 0;
    if (yMax > height - 1) yMax = height - 1;
    
    if ((xMax >= xMin) && (yMax >= yMin))
		update = true;
    
    for (x=xMin; x<=xMax; x++)
    for (y=yMin; y<=yMax; y++)
	{
		k1 = (x-tx)/texelPerUnit;
		k2 = (y-ty)/texelPerUnit;

        v = 1.0f - float(sqrt(k1*k1 + k2*k2 + pow(p_baseDistance,2)) / totalRadius);

        if (v >= 0)
		{
            r = textureData[x * 4 + 2 + y*width*4];
            g = textureData[x * 4 + 1 + y*width*4];
            b = textureData[x * 4 + 0 + y*width*4];
            
			// 1-dest color blend
            r = int(r + p_color.red * v * float(255-r)/255.0f);
            g = int(g + p_color.green * v * float(255-g)/255.0f);
            b = int(b + p_color.blue * v * float(255-b)/255.0f);
            if (r > 255) r = 255;
            if (g > 255) g = 255;
            if (b > 255) b = 255;
            
            textureData[x * 4 + 2 + y*width*4] = r;
            textureData[x * 4 + 1 + y*width*4] = g;
            textureData[x * 4 + 0 + y*width*4] = b;
        }
	}
}

// Load lightmap into memory if necessary
void CAPITexture::UpdateLightmapGL(void)
{
    if ((update) && (textureData))
	{
        if ((width > 0) && (height > 0))
		{
        
            if (!loaded)
			{
                glGenTextures (1, &GLid);
                //Debug.Print "Made Lightmap texture " & GLid
                loaded = true;
				Metrics.textureLMQty++;
			}

            //If Not p_texEnabled Then
                glEnable (GL_TEXTURE_2D);
            //End If
            
            glBindTexture (GL_TEXTURE_2D, GLid);
            // Storage boundary for data (1, 2, 4, 8)
            glPixelStorei (GL_PACK_ALIGNMENT, 1);
            // GL_CLAMP is good for singular icons rendered inside a polygon
            //   but bad for a repeating tiled texture across a loarge number of polygons (like terrain)
            glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
            glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
            // GL_NEAREST makes them look pixelated (good for icons, bad for surfaces)
            glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
            glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
           // Multiply by color (GL_DECAL ignores polygon colors, and would be unaffected by lighting)
            glTexEnvf (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
        
            //glTexImage2D (GL_TEXTURE_2D, 0, GL_RGB, width, height, 0, tiBGRAExt, GL_UNSIGNED_BYTE, textureData);
			// I hope GL_BGRA works in C++!
            glTexImage2D (GL_TEXTURE_2D, 0, GL_RGB, width, height, 0, GL_BGRA, GL_UNSIGNED_BYTE, textureData);
            
            update = false;
            valid = true;
            
            glDisable (GL_TEXTURE_2D);
        }
    }
}

void CAPITexture::SetLightmapDimensions(unsigned p_width, unsigned p_height, float p_texelPerUnit)
{
    if ((lightmap) && (!textureData))
	{
        width = p_width;
        height = p_height;
		widthf = float(width);
		heightf = float(height);
        texelPerUnit = p_texelPerUnit;
        textureData = new BYTE[width * 4 * height]; // set up as BGRA
		Metrics.textureLMMemory += (width * 4 * height);
		textureSysMemorySize = width*4*height;
        
        update = true;
    }
}

// This can be sped up significantly with doubled use of
//   Copymemory
// ??
void CAPITexture::ClearLightmap(CColorRGB p_color)
{
    int i, j;
    if ((lightmap) && (width > 0) && (height > 0) && (textureData))
	{
        for (i=0; i<width; i++)
        for (j=0; j<height; j++)
		{
            textureData[i*4 + j*4*width] = p_color.blue;
            textureData[i*4 + 1 + j*4*width] = p_color.green;
            textureData[i*4 + 2 + j*4*width] = p_color.red;
            textureData[i*4 + 3 + j*4*width] = 0; // can be anything
		}	
    }
    update = true;
}

////////////////////////////////////////////////////////////////////
// Render buffer functions
void CAPITexture::SetupFrameBuffer(defRenderAPIType p_APIType, int p_width, int p_height)
{
    width = p_width;
    height = p_height;
	widthf = float(width);
	heightf = float(height);
    
    // This is only here for GL's benefit, the data won't be used,
    //   it's just that TexImage2D won't accept a NULL pointer
    /*Dim texData() As Byte
    ReDim texData(0 To width * 3 - 1, 0 To height)
    Dim x As Integer, y As Integer
    For x = 0 To width - 1
    For y = 0 To height - 1
        texData(x * 3, y) = 0
        texData(x * 3 + 1, y) = 0
        texData(x * 3 + 2, y) = 0
    Next y
    Next x*/
    
    switch (p_APIType)
	{
	case APITYPE_OPENGL:
            glGenTextures (1, &GLid);
            glEnable (GL_TEXTURE_2D);
            glBindTexture (GL_TEXTURE_2D, GLid);
            glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
            glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
            glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
            glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
            glTexImage2D (GL_TEXTURE_2D, 0, GL_RGB, width, height, 0, GL_RGB, GL_UNSIGNED_BYTE, NULL);
            glDisable (GL_TEXTURE_2D);
            renderBuffer = true;
            loaded = true;
            valid = true;
            mipmap = false;
			break;
        //Case APITYPE_DIRECTX:
        //    ' We did the meat stuff in Initialize
        //    renderbuffer = True
        //   mipmap = False
        //    loaded = True
        //    valid = True
			// break;
	default:
            // Nothing else to do
            renderBuffer = true;
            mipmap = false;
            loaded = false;
            valid = false;
    }
}

//////////////////////////////////////////////
// Font functions

// Need to send void * because of the annoying circular dependence
//   in .h files
void CAPITexture::LoadFontIntoMemory(void *p_vfont)
{
	CAPIFont *p_font = (CAPIFont *)p_vfont;

	if (p_font)
	{
		// Flag texture as a font
		font = true;

		// We still need to load it as ARGB so that the API can load it
		//   without any problems
		// Fonts are always loaded into a 256x256 texture, for now
		width = 256;
		height = 256;
		textureData = new BYTE[width*4*height];
		Metrics.textureMemory += (width*4*height);
		textureSysMemorySize = width*4*height;

		bool *charValid;
		BYTE *charWidth;
		BYTE *charData;
		charWidth = p_font->GetCharWidth();
		charValid = p_font->GetCharValid();
		charData = p_font->GetCharData();

		unsigned startX, startY;
		unsigned indexX, indexY;
		unsigned x, y;
		unsigned i;

		startX = 0;
		startY = 0;
		indexX = 0;
		indexY = 0;
		for (i=0; i<=255; i++)
		{
			if (charValid[i])
			{
				if (startX + charWidth[i] > 255)
				{
					startX = 0;
					startY = startY + p_font->GetHeight();
				}
				p_font->SetCharStartS(i, float(startX));
				p_font->SetCharStartT(i, float(startY));
        
				// Load it up
				for (y=0; y<p_font->GetHeight(); y++)
				for (x=0; x<charWidth[i]; x++)
				{
					// Put it everywhere in ARGB, even though we will only be using ALPHA8
					textureData[(startX + x) * 4 +     width * 4 * (startY + y)] = charData[indexX + indexY*p_font->GetWidth()];
					textureData[(startX + x) * 4 + 1 + width * 4 * (startY + y)] = charData[indexX + indexY*p_font->GetWidth()];
					textureData[(startX + x) * 4 + 2 + width * 4 * (startY + y)] = charData[indexX + indexY*p_font->GetWidth()];
					textureData[(startX + x) * 4 + 3 + width * 4 * (startY + y)] = charData[indexX + indexY*p_font->GetWidth()];
					//textureData[(startX + x) * 4 +     width * 4 * (startY + y)] = charData[indexY + indexX*p_font->GetHeight()];
					//textureData[(startX + x) * 4 + 1 + width * 4 * (startY + y)] = charData[indexY + indexX*p_font->GetHeight()];
					//textureData[(startX + x) * 4 + 2 + width * 4 * (startY + y)] = charData[indexY + indexX*p_font->GetHeight()];
					//textureData[(startX + x) * 4 + 3 + width * 4 * (startY + y)] = charData[indexY + indexX*p_font->GetHeight()];
					//textureData[(startX + x) * 4 + 256 * 4 * (startY + y)] = 255;
					//textureData[(startX + x) * 4 + 1 + 256 * 4 * (startY + y)] = 255;
					//textureData[(startX + x) * 4 + 2 + 256 * 4 * (startY + y)] = 255;
					//textureData[(startX + x) * 4 + 3 + 256 * 4 * (startY + y)] = 255;
            
					indexX = indexX + 1;
					if (indexX >= p_font->GetWidth())
					{
						indexX = 0;
						indexY = indexY + 1;
					}
				}

				startX = startX + charWidth[i];

			}	// charValid
		}
	}
}