// APITexture.h: interface for the CAPITexture class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_APITEXTURE_H__04BA645C_5665_4589_BFAF_328A19F7B59E__INCLUDED_)
#define AFX_APITEXTURE_H__04BA645C_5665_4589_BFAF_328A19F7B59E__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include <gl/gl.h>
#include <gl/glu.h>
#include <d3d8.h>
#include "enums.h"
#include "ColorRGB.h"

enum defAPITextureFormat
{
	APITEXTUREFORMAT_NONE = 0,
	APITEXTUREFORMAT_RGB,
	APITEXTUREFORMAT_RGBA,
	APITEXTUREFORMAT_ALPHA8
};

enum defAPITextureFileFormat
{
	APITEXTUREFILEFORMAT_NONE = 0,
	APITEXTUREFILEFORMAT_BMP,
	APITEXTUREFILEFORMAT_FONT
};

class CAPITexture  
{
public:
	CAPITexture();
	virtual ~CAPITexture();

private:
	// Reference data
	char					name[50];
	char					filename[50];
	defAPITextureFormat		format;
	defAPITextureFileFormat	fileFormat;

	DWORD textureSysMemorySize;	// Amount allocated for system memory

	// Reference for the relevent API
	// If these are populated, it's assumed to be loaded and ready
	//   unless for some reason the texture needs to be restored
	//   the loaded flag serves the same purpose
	// If these values are zero, the texture isn't loaded
	GLuint					GLid;
	LPDIRECT3DTEXTURE8		DXid;

	// Texture data storage (array)
	// This is an in-memory RGBA version of the texture
	// Lightmaps are stored as BGRA for compatibility with
	//    off-screen rendering
	BYTE *textureData;

	bool valid;	    // If true, still possible to load and use
	bool loaded;	// loaded and ready to use (memory could be deallocated
					//    but texture is still usable in the API)

	long width;		// Dimensions
	long height;	// Valid only if loaded is true
	float widthf;	// For quick conversion during rendering
	float heightf;

	bool mipmap;	// Do we want it mipmapped when loaded?
	bool forceClamp;	// Always clamp?  (Usually for textures meant to be projected)

	// Special texture types
	bool lightmap;		// Is it a lightmap?
	bool renderBuffer;	// Is it used for offscreen rendering?
	bool font;			// Is it a font?

	float texelPerUnit;	// How many texture pixels per world unit of 1.0?
						//   (used for lightmaps)
	bool update;		// Does the lightmap need to be loaded into the API
						//    from memory (memory storage changed)?

	bool inList;		// Is this texture in the dynamic list?

	/////////////////////

public:
	void LoadTextureGL(void);
	void LoadTextureDX(void);

	GLuint GetGLid(void) {return GLid;}

	inline void SetTexture(char *p_name, char *p_filename, defAPITextureFormat p_format, bool p_mipmap, defAPITextureFileFormat p_fileFormat)
	{
		SetTexture(p_name, p_filename, p_format, p_mipmap, p_fileFormat, false);
	}
	void SetTexture(char *p_name, char *p_filename, defAPITextureFormat p_format, bool p_mipmap, defAPITextureFileFormat p_fileFormat, bool p_forceClamp);
	void SetMipmap(bool p_mipmap) {mipmap = p_mipmap;}
	void SetInList(bool p_inList) { inList = p_inList;}
	void SetValid(bool p_valid) { valid = p_valid; }

	char * GetName(void) {return name;}

	inline bool IsLoaded(void) {return loaded;}
	inline bool IsValid(void) {return valid;}
	inline bool IsInMemory(void) {return (textureData != NULL);}
	inline bool IsLightmap(void) {return lightmap;}
	inline bool IsRenderBuffer(void) {return renderBuffer;}
	inline bool IsFont(void) {return font;}
	inline bool IsInList(void) {return inList;}

	inline BYTE * GetTextureData(void) {return textureData;};
	inline long GetWidth(void) {return width;}
	inline long GetHeight(void) {return height;}
	inline float GetWidthf(void) {return widthf;}
	inline float GetHeightf(void) {return heightf;}
	void GetDimensions(long &p_widthOut, long &p_heightOut)
	{
		p_widthOut = width;
		p_heightOut = height;
	}
	inline bool GetMipMap(void) {return mipmap;}
	inline bool GetForceClamp(void) {return forceClamp;}
	inline defAPITextureFormat GetFormat(void) {return format;}
	inline float GetTexelPerUnit(void) {return texelPerUnit;}

	bool LoadIntoMemory(void);	// Transfer from file into system memory
	void Destroy(defRenderAPIType p_APIType);	// Get rid of memory and API reference
	void ReleaseAPI(defRenderAPIType p_APIType); // Detach the texture from the API
	void DeallocateMemory(void);	// Get rid of system memory for texture

	////////////////////////////////////////////
	// Lightmap functions
	inline void ApplyRadiusLight(float p_x, float p_y, float p_radius, CColorRGB p_color)
	{
		ApplyRadiusLight(p_x, p_y, p_radius, 0.0f, p_color);
	}
	void ApplyRadiusLight(float p_x, float p_y, float p_radiusMap, float p_baseDistance, CColorRGB p_color);
	void UpdateLightmapGL(void);
	void SetLightmapDimensions(unsigned p_width, unsigned p_height, float p_texelPerUnit);
	void ClearLightmap(CColorRGB p_color);
	void SetLightmap(bool p_lightmap) {lightmap = p_lightmap;}

	////////////////////////////////////////////
	// Renderbuffer functions
	void SetupFrameBuffer(defRenderAPIType p_APIType, int p_width, int p_height);

	////////////////////////////////////////////
	// Font functions
	void LoadFontIntoMemory(void *p_vfont);
};

typedef CAPITexture * LPAPITEXTURE;

#endif // !defined(AFX_APITEXTURE_H__04BA645C_5665_4589_BFAF_328A19F7B59E__INCLUDED_)
