// APITextureRegistry.cpp: implementation of the CAPITextureRegistry class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "APITextureRegistry.h"
#include "..\globals.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CLinkedList<CAPITexture> CLinkedList<CAPITexture>::DeletedList;

CAPITextureRegistry::CAPITextureRegistry()
{

}

CAPITextureRegistry::~CAPITextureRegistry()
{
	DestroyAll();
}

void CAPITextureRegistry::Initialize(unsigned p_textureQty)
{
	textureQty = p_textureQty;

	textures = new CAPITexture[textureQty];
}

void CAPITextureRegistry::DestroyAll(void)
{
	unsigned i;
	for (i=0; i<textureQty; i++)
	{
		textures[i].Destroy(Render->GetAPIType());
	}

	CLinkedListNode<CAPITexture> *node;
	node = textureList.GetFirst();
	if (node)
		while (node->next != NULL)
		{
			node->Destroy(Render->GetAPIType());
			node = node->next;
		}
	textureList.EmptyList();

	if (textures)
	{
		delete [] textures;
		textures = NULL;
		textureQty = 0;
	}

	dynamicLightmap.Destroy(Render->GetAPIType());

	// sphereLightmaps are only references, no need to destroy them
}

void CAPITextureRegistry::ReleaseAPIAll(void)
{
	unsigned i;
	for (i=0; i<textureQty; i++)
	{
		textures[i].ReleaseAPI(Render->GetAPIType());
	}
	CLinkedListNode<CAPITexture> *node;
	node = textureList.GetFirst();
	if (node)
		while (node->next != NULL)
		{
			node->ReleaseAPI(Render->GetAPIType());
			node = node->next;
		}

	dynamicLightmap.ReleaseAPI(Render->GetAPIType());
}

CAPITexture * CAPITextureRegistry::GetTexture(char *p_name)
{
	unsigned i;
	for (i=0; i<textureQty; i++)
	{
		if (!strcmp(p_name, textures[i].GetName()))
			return &textures[i];
	}

	TRACE ("* Warning: Texture '%s' not found in registry\n", p_name);
	return NULL;
}

// For compatibility with VB code
int CAPITextureRegistry::GetTextureID(char *p_name)
{
	unsigned i;
	for (i=0; i<textureQty; i++)
	{
		if (!strcmp(p_name, textures[i].GetName()))
			return (int)i;
	}

	TRACE ("* Warning: Texture '%s' not found in registry\n", p_name);
	return -1;
}

CAPITexture * CAPITextureRegistry::GetTexture(unsigned p_id)
{
	return &textures[p_id];
}

CAPITexture * CAPITextureRegistry::CreateNewTexture(void)
{
	CLinkedListNode<CAPITexture> *node;
	node = textureList.NewNode(true);
	node->SetInList(true);
	return node;
}

// Get rid of a dynamically requested texture
void CAPITextureRegistry::DeleteTexture(CAPITexture *p_texture)
{
	if (p_texture)
	if (p_texture->IsInList())
	{
		p_texture->Destroy(Render->GetAPIType());
		textureList.Delete((CLinkedListNode<CAPITexture> *)p_texture);
	}
}
