// APITextureRegistry.h: interface for the CAPITextureRegistry class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_APITEXTUREREGISTRY_H__C265D2B3_60A6_4F34_A667_5F0A93CAA07E__INCLUDED_)
#define AFX_APITEXTUREREGISTRY_H__C265D2B3_60A6_4F34_A667_5F0A93CAA07E__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "APITexture.h"
#include "..\structs\LinkedList.h"

class CAPITextureRegistry  
{
public:
	CAPITextureRegistry();
	virtual ~CAPITextureRegistry();

private:
	CAPITexture	*textures;		// Static textures defined at startup
	unsigned textureQty;

	CLinkedList<CAPITexture> textureList;	// Additional textures
											//   requested during runtime

	CAPITexture	dynamicLightmap;	// Texture used for dynamic lightmaps
									//   (offscreen rendered, then
									//    modulated with surfaces and decals)

	LPAPITEXTURE sphereLightmaps[10];	// References to lightmap splices

public:
	void Initialize(unsigned p_textureQty);
	void DestroyAll(void);
	void ReleaseAPIAll(void);
	CAPITexture * GetTexture(char *p_name);
	CAPITexture * GetTexture(unsigned p_id);

	// For compatibility with VB code
	int GetTextureID(char *p_name);
	CAPITexture * GetTexture(int p_id)
	{
		if (p_id >= 0)
			return &textures[p_id];
		else
			return NULL;
	}
	void SetSphereLightmap(unsigned p_id, char *p_name)
	{
		if (p_id < 10)
			sphereLightmaps[p_id] = GetTexture(p_name);
	}
	CAPITexture * GetSphereLightmap(unsigned p_id)
	{
		if (p_id < 10)
			return sphereLightmaps[p_id];
		else
			return NULL;
	}

	// Call this whenever you need another texture and it isn't in the static
	//   array.  This ensures the texture will be destroyed and detached
	//   when you change API's and end the program
	CAPITexture * CreateNewTexture(void);

	CAPITexture * GetDynamicLightmap(void) {return &dynamicLightmap;}

	void DeleteTexture(CAPITexture *p_texture);
};

#endif // !defined(AFX_APITEXTUREREGISTRY_H__C265D2B3_60A6_4F34_A667_5F0A93CAA07E__INCLUDED_)
