// IRender.h: interface for the IRender class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_IRENDER_H__61160028_325F_4A6F_A6FC_5FF247B16308__INCLUDED_)
#define AFX_IRENDER_H__61160028_325F_4A6F_A6FC_5FF247B16308__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "..\math\Orient3D.h"
#include "Model3D.h"
#include "Surface3D.h"
#include "ModelVertex3D.h"
#include "enums.h"
#include "..\structs\LinkedList.h"
#include "SurfaceDecal.h"
#include "LightSource.h"
#include "APIFont.h"
#include "Terrain.h"
#include "StarField.h"
#include "ParticleEmitter.h"

class IRender  
{
public:
	IRender();
	virtual ~IRender();

	virtual defRenderAPIType GetAPIType(void);

	virtual void Initialize(HWND p_hWnd);
	virtual void SetViewport(float p_angle, float p_nearZ, float p_farZ);
	virtual void SetViewport(float p_angle);
	virtual void SetViewportOrtho(void);
	virtual void Clear(long p_color);
	virtual void PushMatrix(void);
	virtual void PopMatrix(void);
	virtual void LoadIdentity(void);
	virtual void Translatef(float p_x, float p_y, float p_z);
	virtual void Scalef(float p_x, float p_y, float p_z);
	virtual void Rotate(float p_angle, float p_x, float p_y, float p_z);
	virtual COrient3D GetViewpoint(void);
	virtual void BeginScene(void);
	virtual void RenderTestTriangle(void);
	virtual void EndScene(void);
	virtual void Flip(void);
	virtual void Destroy(void);

	virtual void TransformWorld(COrient3D &p_orient);
	virtual void TransformViewpoint(COrient3D &p_orient);

	virtual bool ConvertEyeVectorToViewportXY(CVector3D &p_vector, int &p_xOut, int &p_yOut);
	virtual CVector3D ConvertViewportXYToWorldVector3D(int p_x, int p_y);

	inline void RenderTerrain(CTerrain &p_terrain, bool p_useTestColor, bool p_useTexture)
	{
		RenderTerrain(p_terrain, p_useTestColor, p_useTexture, NULL, false, false, false, false);
	}
	virtual void RenderTerrain(CTerrain &p_terrain, bool p_useTestColor, bool p_useTexture, CTerrain *upperLod, bool p_loX, bool p_loZ, bool p_hiX, bool p_hiZ);

	inline void RenderMultiQuad(CColorRGB p_color, BYTE p_alpha)
	{
		RenderMultiQuad(p_color, p_alpha, NULL, NULL, APIMULTIQUADTEXTUREOP_MODULATE);
	}
	inline void RenderMultiQuad(CColorRGB p_color, BYTE p_alpha, CAPITexture *p_texture1)
	{
		RenderMultiQuad(p_color, p_alpha, p_texture1, NULL, APIMULTIQUADTEXTUREOP_MODULATE);
	}
	inline void RenderMultiQuad(CColorRGB p_color, BYTE p_alpha, CAPITexture *p_texture1, CAPITexture *p_texture2)
	{
		RenderMultiQuad(p_color, p_alpha, p_texture1, p_texture2, APIMULTIQUADTEXTUREOP_MODULATE);
	}
	virtual void RenderMultiQuad(CColorRGB p_color, BYTE p_alpha, CAPITexture *p_texture1, CAPITexture *p_texture2, defMultiQuadTextureOp p_textureOp);
	inline void RenderQuad(CColorRGB p_color, BYTE p_alpha)
	{
		RenderQuad(p_color, p_alpha, NULL);
	}
	virtual void RenderQuad(CColorRGB p_color, BYTE p_alpha, CAPITexture *p_texture);

	inline void RenderModel(CModel3D *p_model)
	{
		RenderModelFinal(p_model, NULL, 255);
	}
	inline void RenderModelColor(CModel3D *p_model, CColorRGB *p_modColor)
	{
		RenderModelFinal(p_model, p_modColor, 255);
	}
	virtual void RenderModelFinal(CModel3D *p_model, CColorRGB *p_modColor, BYTE p_modAlpha);
	virtual void RenderModelFinalLM(CModel3D *p_model, CColorRGB *p_modColor, BYTE p_modAlpha, CAPITexture **p_lightmaps, CSurfaceDecalList *p_decalLists, CLightSourceRefList *p_lightSourceList, COrient3D *p_orient);
	inline void RenderSurfaces(CSurface3D *p_surfaces, unsigned p_surfaceQty, CModelVertex3D *p_vertices, CColorRGB *p_modColor, BYTE p_modAlpha)
	{
		RenderSurfaces(p_surfaces, p_surfaceQty, p_vertices, p_modColor, p_modAlpha, NULL, NULL, true, NULL, NULL, NULL);
	}
	inline void RenderSurfaces(CSurface3D *p_surfaces, unsigned p_surfaceQty, CModelVertex3D *p_vertices, CColorRGB *p_modColor, BYTE p_modAlpha, CAPITexture **p_lightMaps, CSurfaceDecalList *p_decalLists, bool p_ignoreMirrors, CLightSourceRefList *p_lightSourceList)
	{
		RenderSurfaces(p_surfaces, p_surfaceQty, p_vertices, p_modColor, p_modAlpha, p_lightMaps, p_decalLists, p_ignoreMirrors, p_lightSourceList, NULL, NULL);
	}
	virtual void RenderSurfaces(CSurface3D *p_surfaces, unsigned p_surfaceQty, CModelVertex3D *p_vertices, CColorRGB *p_modColor, BYTE p_modAlpha, CAPITexture **p_lightmaps, CSurfaceDecalList *p_decalLists, bool p_ignoreMirrors, CLightSourceRefList *p_lightSourceList, CAPITexture *p_useThisLightmap, COrient3D *p_orient);

	inline void RenderText(unsigned p_x, unsigned p_y, char *p_text, CAPIFont &p_font, CColorRGB &p_color)
	{
		RenderText(p_x, p_y, p_text, p_font, p_color, 255, NULL);
	}
	inline void RenderText(unsigned p_x, unsigned p_y, char *p_text, CAPIFont &p_font, CColorRGB &p_color, BYTE p_alpha)
	{
		RenderText(p_x, p_y, p_text, p_font, p_color, p_alpha, NULL);
	}
	virtual void RenderText(int p_x, int p_y, char *p_text, CAPIFont &p_font, CColorRGB &p_color, BYTE p_alpha, RECT *p_clipRect);
	inline void RenderStarField(CStarField &p_starField)
	{
		RenderStarField(p_starField, NULL, NULL, 0);
	}
	virtual void RenderStarField(CStarField &p_starField, COrient3D *p_currentOrient, COrient3D *p_oldOrient, DWORD p_backColor);
	inline void RenderParticleSet(CParticleSet &p_particleSet, CColorRGB *p_ambientColor, unsigned int p_renderID, CLightSourceRefList *p_lightSourceList)
	{
		//RenderParticles(p_particleEmitter.GetParticles(), p_particleEmitter.GetParticleQty(), NULL);
		RenderParticles(p_particleSet.GetParticles(), p_particleSet.GetParticleQty(), p_renderID, p_ambientColor, p_lightSourceList);
	}
	inline void RenderParticleSet(CParticleSet &p_particleSet, unsigned int p_renderID)
	{
		//RenderParticles(p_particleEmitter.GetParticles(), p_particleEmitter.GetParticleQty(), NULL);
		RenderParticles(p_particleSet.GetParticles(), p_particleSet.GetParticleQty(), p_renderID);
	}
/*	inline void RenderParticleEmitter(CParticleEmitter &p_particleEmitter, CLightSourceList *p_lightSources)
	{
		//RenderParticles(p_particleEmitter.GetParticles(), p_particleEmitter.GetParticleQty(), p_lightSources);
		RenderParticles(p_particleEmitter.GetParticles(), p_particleEmitter.GetParticleQty(), p_lightSources);
	}*/
	inline void RenderParticles(CParticle *p_particles, unsigned int p_particleQty, unsigned int p_renderID)
	{
		RenderParticles(p_particles, p_particleQty, p_renderID, NULL, NULL);
	}
	virtual void RenderParticles(CParticle *p_particles, unsigned int p_particleQty, unsigned int p_renderID, CColorRGB *p_ambientColor, CLightSourceRefList *p_lightSources);

	virtual void RenderWireframeCube(float p_minX, float p_minY, float p_minZ, float p_maxX, float p_maxY, float p_maxZ);

	///////////////////////////////////////
	// Ortho routines
	virtual void DrawBoxOrtho(int p_left, int p_top, int p_right, int p_bottom);

	///////////////////////////////////////
	// State routines
	virtual void EnableDepthTest(bool p_enable);
	virtual void EnableDepthWrite(bool p_enable);
	virtual void SetCullCCW(bool p_cullCCW);
	virtual bool GetCullCCW(void);
	virtual void SetStencilEnable(bool p_enable);
	virtual void SetStencilWrite(BYTE p_value);
	virtual void SetStencilTest(BYTE p_value);
	virtual void SetStencilIncrement(BYTE p_value, bool p_ignoreDepth);
	virtual void SetStencilDecrement(BYTE p_value, bool p_ignoreDepth);
	inline void SetStencilIncrement(BYTE p_value)
	{
		SetStencilIncrement(p_value, false);
	}
	inline void SetStencilDecrement(BYTE p_value)
	{
		SetStencilDecrement(p_value, false);
	}
	virtual void EnableColorWrite(bool p_enable);
	virtual void EnableCulling(bool p_enable);
	virtual void SetClipPlane(CVector3D &p_planeNormal, CVector3D &p_planePoint);
	virtual void DisableClipPlane(void);
	virtual void SetBlendMode(defAPIBlendMode p_blendMode);
	virtual void SetDepthTestFunc(defAPIDepthTestFunc p_depthTestFunc);
	virtual void SetZBias(bool p_zBias);

	virtual float GetScaleY(void);
};

#endif // !defined(AFX_IRENDER_H__61160028_325F_4A6F_A6FC_5FF247B16308__INCLUDED_)
