// LightSource.cpp: implementation of the CLightSource class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "LightSource.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

CLinkedList<CLightSource> CLinkedList<CLightSource>::DeletedList;
CLinkedList<CLightSourceRef> CLinkedList<CLightSourceRef>::DeletedList;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

unsigned int CLightSource::currentRenderID = 0;

CLightSource::CLightSource()
{
	staticLS = false;
	currentFlareSize = 0;
	targetFlareSize = 0;
	renderID = 0;
}

CLightSource::~CLightSource()
{

}

CColorRGB CLightSourceRefList::CalculateRenderColor(CVector3D &p_position, CColorRGB p_ambientColor)
{
	CColorRGB result;
	result.SetColor(p_ambientColor);

	float distance, ratio;
	CLinkedListNode<CLightSourceRef> *node;
	node = GetFirst();
	if (node)
    while (node->next != NULL)
	{
        if ((node->ref->visible) && (node->ref->intensity > 0))
		{
           switch (node->ref->type)
		   {
		   case APILIGHTTYPE_SPHERICALLIGHTMAP:
                distance = (p_position - (node->ref->position)).MgnTud();
                if (distance < node->ref->radius)
				{
                    ratio = ((node->ref->radius - distance) / node->ref->radius) * node->ref->intensity;
                    result.SetColor(result.AddTend(node->ref->color.Scm(ratio)));
                }
				break;
		   }
        }

        node = node->next;
    }

	return result;
}

void CLightSource::DetermineCellsTouched(COctree &p_octree)
{
	CVector3D minPoint;
	CVector3D maxPoint;

	COctCellRefList cellList;

	minPoint.x = position.x - radius;
	minPoint.y = position.y - radius;
	minPoint.z = position.z - radius;
	maxPoint.x = position.x + radius;
	maxPoint.y = position.y + radius;
	maxPoint.z = position.z + radius;

	p_octree.CellsTouched(minPoint.x, minPoint.y,minPoint.z, maxPoint.x, maxPoint.y, maxPoint.z, cellList);

	// Update cell object lists
	CLinkedListNode<COctCellRef> *cnode1, *cnode2;
	CLinkedListNode<CLPVoid> *vnode1;
	bool found;

	// For every cell in the old list that is not in the new list,
	//   remove the object from those cells
	cnode1 = cellsTouched.GetFirst();
	if (cnode1)
	while (cnode1->next != NULL)
	{
		cnode2 = cellList.GetFirst();
		found = false;

		if (cnode2)
		while ((cnode2->next != NULL) && (!found))
		{
			if (cnode1->cellRef == cnode2->cellRef)
				found = true;
			else
				cnode2 = cnode2->next;
		}

		// Not there, remove this object from that cell's list
		if (!found)
		{
			vnode1 = cnode1->cellRef->lpVoidList2.GetFirst();
			found = false;
			if (vnode1)
			while ((vnode1->next != NULL) && (!found))
			{
				if (vnode1->lpVoid == this)
				{
					found = true;
					cnode1->cellRef->lpVoidList2.Delete(vnode1);
				}
				else
					vnode1 = vnode1->next;
			}
		}

		cnode1 = cnode1->next;
	}

	// For every cell in the new list that is not in the old list,
	//   add the object to those cells
	cnode1 = cellList.GetFirst();
	if (cnode1)
	while (cnode1->next != NULL)
	{
		cnode2 = cellsTouched.GetFirst();
		found = false;

		if (cnode2)
		while ((cnode2->next != NULL) && (!found))
		{
			if (cnode1->cellRef == cnode2->cellRef)
				found = true;
			else
				cnode2 = cnode2->next;
		}

		// Not there, add this object to that cell's list
		if (!found)
		{
			vnode1 = cnode1->cellRef->lpVoidList2.NewNode();
			vnode1->lpVoid = this;
		}

		cnode1 = cnode1->next;
	}

	// Finally, replace the object's old list with the new list
	cellsTouched.EmptyList();
	if (!cellList.IsEmpty())
	{
		cellsTouched.GetHeader()->next = cellList.GetHeader()->next;
		cellsTouched.GetFooter()->prev = cellList.GetFooter()->prev;
		cellList.GetFirst()->prev = cellsTouched.GetHeader();
		cellList.GetLast()->next = cellsTouched.GetFooter();
		cellList.GetHeader()->next = cellList.GetFooter();
		cellList.GetFooter()->prev = cellList.GetHeader();
	}
}

