// LightSource.h: interface for the CLightSource class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_LIGHTSOURCE_H__F50D1FCC_0607_461D_9773_E2AE1DA27C1A__INCLUDED_)
#define AFX_LIGHTSOURCE_H__F50D1FCC_0607_461D_9773_E2AE1DA27C1A__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "..\structs\LinkedList.h"
#include "..\math\Vector3D.h"
#include "ColorRGB.h"
#include "enums.h"
#include "..\structs\octree.h"

class CLightSource  
{
public:
	CLightSource();
	virtual ~CLightSource();

	float radius;			// How far does the light extend?
	CVector3D position;		// Position in world space
	CColorRGB color;		// Color of the light
	float intensity;		// How bright is it?  (mod with color)
	bool visible;			// Is it 'on'?
	defAPILightSourceType	type;	//  What kind of light source is it?
	bool staticLS;			// Is it a static (neverchanging) light source?

	// Size ranges from 0.0-1.0 and is only for making flare
	//   appear or disappear around corners
	float currentFlareSize;	// For rendering flares
	float targetFlareSize;	// Current assumes 1 flare per light source,
								//   this could change with mirrors, and
								//   viewing light sources in multiple windows
	float flareRadius;		// Radius of flare at full size

	unsigned int renderID;	// To keep lightsource from being referred to twice
							//   when ref lists are assembled
	static unsigned int currentRenderID;	// ID to increment and use

	COctCellRefList cellsTouched;	// Octree cells that this light source is
									//    touching

	void DetermineCellsTouched(COctree &p_octree);
};

class CLightSourceRef
{
public:
	CLightSource *ref;
};

class CLightSourceRefListNode : public CLinkedListNode<CLightSourceRef>
{
};

class CLightSourceListNode : public CLinkedListNode<CLightSource>
{
};

class CLightSourceRefList : public CLinkedList<CLightSourceRef>
{
public:
	CLightSourceRefListNode * NewNode(void)
	{
		return NewNode(true);
	}
	CLightSourceRefListNode * NewNode(bool p_addAppend)
	{
		return (CLightSourceRefListNode *)CLinkedList<CLightSourceRef>::NewNode(p_addAppend);
	}
	CLightSourceRefListNode * GetFirst(void)
	{
		return (CLightSourceRefListNode *)CLinkedList<CLightSourceRef>::GetFirst();
	}

	CColorRGB CalculateRenderColor(CVector3D &p_position, CColorRGB p_ambientColor);
};

class CLightSourceList : public CLinkedList<CLightSource>
{
	// No special operations needed, these
	//   class names are just for convenience
public:
	// Well, there is one...
	CLightSourceListNode * NewNode(void)
	{
		return NewNode(true);
	}
	CLightSourceListNode * NewNode(bool p_addAppend)
	{
		return (CLightSourceListNode *)CLinkedList<CLightSource>::NewNode(p_addAppend);
	}
	CLightSourceListNode * GetFirst(void)
	{
		return (CLightSourceListNode *)CLinkedList<CLightSource>::GetFirst();
	}

	// We'll need to destroy the cellsTouched lists when we delete them
	void Delete(CLinkedListNode<CLightSource> *node)
	{
		node->cellsTouched.EmptyList();
		CLinkedList<CLightSource>::Delete(node);
	}

	void EmptyList(void)
	{
		CLinkedListNode<CLightSource> *node = GetFirst();
		if (node)
		while (node->next != NULL)
		{
			node->cellsTouched.EmptyList();

			node = node->next;
		}
		CLinkedList<CLightSource>::EmptyList();
	}
};

#endif // !defined(AFX_LIGHTSOURCE_H__F50D1FCC_0607_461D_9773_E2AE1DA27C1A__INCLUDED_)
