// Model3D.cpp: implementation of the CModel3D class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "Model3D.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CModel3D::CModel3D()
{

}

CModel3D::~CModel3D()
{
	Destroy();
}

void CModel3D::Destroy(void)
{
	if (surfaces)
	{
		delete [] surfaces;
		surfaces = NULL;
		surfaceQty = 0;
	}
	if (vertices)
	{
		delete [] vertices;
		vertices = NULL;
		vertexQty = 0;
	}
}

void CModel3D::Initialize(unsigned p_vertexQty, unsigned p_surfaceQty, bool p_allSurfacesWhite)
{
	vertexQty = p_vertexQty;
	surfaceQty = p_surfaceQty;

	vertices = new CModelVertex3D[p_vertexQty];
	surfaces = new CSurface3D[p_surfaceQty];

	unsigned i;
	for (i=0; i<surfaceQty; i++)
	{
		surfaces[i].SetID(i);
		if (p_allSurfacesWhite)
			surfaces[i].SetColor(255,255,255);
	}
}

void CModel3D::MakeTestBlock(void)
{
	Initialize(8, 6, false);

	vertices[0].vertex.SetVec(1,1,1);
	vertices[1].vertex.SetVec(-1,1,1);
	vertices[2].vertex.SetVec(-1,1,-1);
	vertices[3].vertex.SetVec(1,1,-1);
	vertices[4].vertex.SetVec(1,-1,1);
	vertices[5].vertex.SetVec(-1,-1,1);
	vertices[6].vertex.SetVec(-1,-1,-1);
	vertices[7].vertex.SetVec(1,-1,-1);

	vertices[0].color.SetColor(0,0,0);
	vertices[1].color.SetColor(255,0,0);
	vertices[2].color.SetColor(0,255,0);
	vertices[3].color.SetColor(0,0,255);
	vertices[4].color.SetColor(255,255,0);
	vertices[5].color.SetColor(255,0,255);
	vertices[6].color.SetColor(0,255,255);
	vertices[7].color.SetColor(255,255,255);

	surfaces[0].Initialize(4);
	surfaces[0].SetVertexRef(0,0);
	surfaces[0].SetVertexRef(1,1);
	surfaces[0].SetVertexRef(2,2);
	surfaces[0].SetVertexRef(3,3);
	surfaces[1].Initialize(4);
	surfaces[1].SetVertexRef(0,0);
	surfaces[1].SetVertexRef(1,3);
	surfaces[1].SetVertexRef(2,7);
	surfaces[1].SetVertexRef(3,4);
	surfaces[2].Initialize(4);
	surfaces[2].SetVertexRef(0,1);
	surfaces[2].SetVertexRef(1,0);
	surfaces[2].SetVertexRef(2,4);
	surfaces[2].SetVertexRef(3,5);
	surfaces[3].Initialize(4);
	surfaces[3].SetVertexRef(0,2);
	surfaces[3].SetVertexRef(1,1);
	surfaces[3].SetVertexRef(2,5);
	surfaces[3].SetVertexRef(3,6);
	surfaces[4].Initialize(4);
	surfaces[4].SetVertexRef(0,3);
	surfaces[4].SetVertexRef(1,2);
	surfaces[4].SetVertexRef(2,6);
	surfaces[4].SetVertexRef(3,7);
	surfaces[5].Initialize(4);
	surfaces[5].SetVertexRef(0,4);
	surfaces[5].SetVertexRef(1,7);
	surfaces[5].SetVertexRef(2,6);
	surfaces[5].SetVertexRef(3,5);
}

void CModel3D::MakeBlock(float p_minX, float p_minY, float p_minZ, float p_maxX, float p_maxY, float p_maxZ, CAPITexture *p_texture)
{
    // 0 1 2 3   top, 0 at max x,y,z
    // 4 5 6 7   base
    // both CW from top
    
    Initialize(8, 6);
    vertices[0].vertex.SetVec(p_maxX, p_maxY, p_maxZ);
    vertices[1].vertex.SetVec(p_minX, p_maxY, p_maxZ);
    vertices[2].vertex.SetVec(p_minX, p_maxY, p_minZ);
    vertices[3].vertex.SetVec(p_maxX, p_maxY, p_minZ);
    vertices[4].vertex.SetVec(p_maxX, p_minY, p_maxZ);
    vertices[5].vertex.SetVec(p_minX, p_minY, p_maxZ);
    vertices[6].vertex.SetVec(p_minX, p_minY, p_minZ);
    vertices[7].vertex.SetVec(p_maxX, p_minY, p_minZ);
    
    if (!p_texture)
	{
        surfaces[0].Initialize(4);
        surfaces[0].SetVertexRef(0, 0);
        surfaces[0].SetVertexRef(1, 1);
        surfaces[0].SetVertexRef(2, 2);
        surfaces[0].SetVertexRef(3, 3);
        surfaces[1].Initialize(4);
        surfaces[1].SetVertexRef(0, 1);
        surfaces[1].SetVertexRef(1, 0);
        surfaces[1].SetVertexRef(2, 4);
        surfaces[1].SetVertexRef(3, 5);
        surfaces[2].Initialize(4);
        surfaces[2].SetVertexRef(0, 0);
        surfaces[2].SetVertexRef(1, 3);
        surfaces[2].SetVertexRef(2, 7);
        surfaces[2].SetVertexRef(3, 4);
        surfaces[3].Initialize(4);
        surfaces[3].SetVertexRef(0, 3);
        surfaces[3].SetVertexRef(1, 2);
        surfaces[3].SetVertexRef(2, 6);
        surfaces[3].SetVertexRef(3, 7);
        surfaces[4].Initialize(4);
        surfaces[4].SetVertexRef(0, 2);
        surfaces[4].SetVertexRef(1, 1);
        surfaces[4].SetVertexRef(2, 5);
        surfaces[4].SetVertexRef(3, 6);
        surfaces[5].Initialize(4);
        surfaces[5].SetVertexRef(0, 4);
        surfaces[5].SetVertexRef(1, 7);
        surfaces[5].SetVertexRef(2, 6);
        surfaces[5].SetVertexRef(3, 5);
	}
    else
	{
        surfaces[0].Initialize(4, true, p_texture);
        surfaces[0].SetVertexRefTexture(0, 0, 0, 0);
        surfaces[0].SetVertexRefTexture(1, 1, 1, 0);
        surfaces[0].SetVertexRefTexture(2, 2, 1, 1);
        surfaces[0].SetVertexRefTexture(3, 3, 0, 1);
        surfaces[1].Initialize(4, true, p_texture);
        surfaces[1].SetVertexRefTexture(0, 1, 0, 0);
        surfaces[1].SetVertexRefTexture(1, 0, 1, 0);
        surfaces[1].SetVertexRefTexture(2, 4, 1, 1);
        surfaces[1].SetVertexRefTexture(3, 5, 0, 1);
        surfaces[2].Initialize(4, true, p_texture);
        surfaces[2].SetVertexRefTexture(0, 0, 0, 0);
        surfaces[2].SetVertexRefTexture(1, 3, 1, 0);
        surfaces[2].SetVertexRefTexture(2, 7, 1, 1);
        surfaces[2].SetVertexRefTexture(3, 4, 0, 1);
        surfaces[3].Initialize(4, true, p_texture);
        surfaces[3].SetVertexRefTexture(0, 3, 0, 0);
        surfaces[3].SetVertexRefTexture(1, 2, 1, 0);
        surfaces[3].SetVertexRefTexture(2, 6, 1, 1);
        surfaces[3].SetVertexRefTexture(3, 7, 0, 1);
        surfaces[4].Initialize(4, true, p_texture);
        surfaces[4].SetVertexRefTexture(0, 2, 0, 0);
        surfaces[4].SetVertexRefTexture(1, 1, 1, 0);
        surfaces[4].SetVertexRefTexture(2, 5, 1, 1);
        surfaces[4].SetVertexRefTexture(3, 6, 0, 1);
        surfaces[5].Initialize(4, true, p_texture);
        surfaces[5].SetVertexRefTexture(0, 4, 0, 0);
        surfaces[5].SetVertexRefTexture(1, 7, 1, 0);
        surfaces[5].SetVertexRefTexture(2, 6, 1, 1);
        surfaces[5].SetVertexRefTexture(3, 5, 0, 1);
    }
    
    CalculateSurfaceData();
}

// This one is more uniform, using the
//   type resulting from Game Programming Gems
void CModel3D::GenerateSphere2(float p_radius)
{
	GenerateSphere2(p_radius, NULL);
}
void CModel3D::GenerateSphere2(float p_radius, CColorRGB *p_color)
{	Initialize(18, 32, false);

	vertices[0].vertex.SetVec(0, 1, 0);
	vertices[1].vertex.SetVec(0.0f, 0.5f, 0.5f);
	vertices[2].vertex.SetVec(0.5f, 0.5f, 0.0f);
	vertices[3].vertex.SetVec(0.0f, 0.5f, -0.5f);
	vertices[4].vertex.SetVec(-0.5f, 0.5f, 0.0f);
	vertices[5].vertex.SetVec(0, 0, 1);
	vertices[6].vertex.SetVec(0.5f, 0.0f, 0.5f);
	vertices[7].vertex.SetVec(1, 0, 0);
	vertices[8].vertex.SetVec(0.5f, 0.0f, -0.5f);
	vertices[9].vertex.SetVec(0, 0, -1);
	vertices[10].vertex.SetVec(-0.5f, 0.0f, -0.5f);
	vertices[11].vertex.SetVec(-1, 0, 0);
	vertices[12].vertex.SetVec(-0.5f, 0.0f, 0.5f);
	vertices[13].vertex.SetVec(0.0f, -0.5f, 0.5f);
	vertices[14].vertex.SetVec(0.5f, -0.5f, 0.0f);
	vertices[15].vertex.SetVec(0.0f, -0.5f, -0.5f);
	vertices[16].vertex.SetVec(-0.5f, -0.5f, 0.0f);
	vertices[17].vertex.SetVec(0, -1, 0);

	int i;
	for (i=0; i<=17; i++)
	{
		vertices[i].vertex.Unitize();
		vertices[i].vertex = vertices[i].vertex * p_radius;
	}

	for (i=0; i<=3; i++)
	{
		surfaces[i].Initialize(3);
    
		surfaces[i].SetVertexRef(0, 0);
		if (i < 3)
			surfaces[i].SetVertexRef(1, i + 2);
		else
			surfaces[i].SetVertexRef(1, 1);
		surfaces[i].SetVertexRef(2, i + 1);
    
		surfaces[31 - i].Initialize(3);
		surfaces[31 - i].SetVertexRef(0, 17);
		surfaces[31 - i].SetVertexRef(1, i + 13);
		if (i < 3)
			surfaces[31 - i].SetVertexRef(2, i + 14);
		else
			surfaces[31 - i].SetVertexRef(2, 13);
	}

	for (i=0; i<=3; i++)
	{
		surfaces[i + 4].Initialize(3);
		surfaces[i + 4].SetVertexRef(0, i + 1);
		if (i < 3)
			surfaces[i + 4].SetVertexRef(1, i + 2);
		else
			surfaces[i + 4].SetVertexRef(1, 1);
		surfaces[i + 4].SetVertexRef(2, 6 + i * 2);
    
		surfaces[i + 8].Initialize(3);
		surfaces[i + 8].SetVertexRef(0, 6 + i * 2);
		if (i < 3)
			surfaces[i + 8].SetVertexRef(1, i + 14);
		else
			surfaces[i + 8].SetVertexRef(1, 13);
		surfaces[i + 8].SetVertexRef(2, 13 + i);
	}

	for (i=0; i<=7; i++)
	{
		surfaces[i + 12].Initialize(3);
		surfaces[i + 12].SetVertexRef(0, int((i + 2) / 2));
		surfaces[i + 12].SetVertexRef(1, 5 + i);
		if (i > 0)
			surfaces[i + 12].SetVertexRef(2, 4 + i);
		else
			surfaces[i + 12].SetVertexRef(2, 12);
    
		surfaces[i + 20].Initialize(3);
		if (i > 0)
			surfaces[i + 20].SetVertexRef(0, 4 + i);
		else
			surfaces[i + 20].SetVertexRef(0, 12);
		surfaces[i + 20].SetVertexRef(1, 5 + i);
		surfaces[i + 20].SetVertexRef(2, 13 + int(i / 2));
	}

	for (i=0; i<=17; i++)
	{
		vertices[i].color.red = BYTE(255 * (vertices[i].vertex.z + p_radius) / 2);
		vertices[i].color.green = BYTE(255 * (vertices[i].vertex.x + p_radius) / 2);
		vertices[i].color.blue = BYTE(255 * (vertices[i].vertex.y + p_radius) / 2);
	}

	vertices[0].color.red = 0;
	vertices[0].color.green = 0;
	vertices[0].color.blue = 0;

	vertices[17].color.red = 0;
	vertices[17].color.green = 0;
	vertices[17].color.blue = 0;

	vertices[5].color.red = 0;
	vertices[5].color.green = 0;
	vertices[5].color.blue = 0;

	vertices[7].color.red = 0;
	vertices[7].color.green = 0;
	vertices[7].color.blue = 0;

	vertices[9].color.red = 0;
	vertices[9].color.green = 0;
	vertices[9].color.blue = 0;

	vertices[11].color.red = 0;
	vertices[11].color.green = 0;
	vertices[11].color.blue = 0;

	if (p_color)
		for (i=0; i<=31; i++)
			surfaces[i].SetColor(*p_color);

	CalculateSurfaceData();
}

void CModel3D::CalculateSurfaceData(void)
{
	unsigned int i, j;
	for(i=0; i<surfaceQty; i++)
		surfaces[i].CalculateSurfaceData(vertices);

	// Get model radius from used vertices
	float maxRadiusSqr = 0.0f;
	for (i=0; i<surfaceQty; i++)
		for (j=0; j<surfaces[i].GetVertexRefQty(); j++)
		{
			if (vertices[surfaces[i].GetVertexRef(j)].vertex.MgnTudSqr() > maxRadiusSqr)
				maxRadiusSqr = vertices[surfaces[i].GetVertexRef(j)].vertex.MgnTudSqr();
		}

	radius = float(sqrt(maxRadiusSqr));
}
