// OpenGLRender.cpp: implementation of the OpenGLRender class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "OpenGLRender.h"
#include <math.h>	// tan
#include "..\globals.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

OpenGLRender::OpenGLRender()
{
	m_hDC = NULL;
	m_hWnd = NULL;
	m_hRC = NULL;
}

OpenGLRender::~OpenGLRender()
{
	Destroy();
}

defRenderAPIType OpenGLRender::GetAPIType(void)
{
	return APITYPE_OPENGL;
}

void OpenGLRender::Initialize(HWND p_hWnd)
{
	m_hWnd = p_hWnd;

	// Get window dimensions
	RECT rect;
	GetWindowRect(m_hWnd, &rect);
	m_windowWidth = rect.right - rect.left;
	m_windowHeight = rect.bottom - rect.top;

	PIXELFORMATDESCRIPTOR pfd;
	long result;

    m_hDC = GetDC(m_hWnd);
    
	//Setup PDF n RC ...
	pfd.nSize = sizeof(pfd);
	pfd.nVersion = 1;
	pfd.dwFlags = PFD_SUPPORT_OPENGL | PFD_DRAW_TO_WINDOW | PFD_DOUBLEBUFFER | PFD_TYPE_RGBA;
	pfd.iPixelType = PFD_TYPE_RGBA;
	pfd.cColorBits = 32;
	pfd.cDepthBits = 16;
	pfd.cStencilBits = 8;  // Putting this in allows the stencil to work in 16-bit color mode
  
	result = ChoosePixelFormat(m_hDC, &pfd);

	SetPixelFormat (m_hDC, result, &pfd);
	m_hRC = wglCreateContext(m_hDC);

	//Init Quadric Object...
	//QObj = gluNewQuadric()
	//gluQuadricTexture QObj, GL_TRUE
	//gluQuadricNormals QObj, GL_SMOOTH
  
	// Viewport inits...
	wglMakeCurrent (m_hDC, m_hRC);
	glClearColor (0.0f, 0.0f, 0.0f, 0.0f);  // clear view with black color
	glShadeModel (GL_SMOOTH); //Interpolate colors
	cullCCW = true;
	glEnable(GL_CULL_FACE);
	glFrontFace (GL_CW);
	glEnable (GL_DEPTH_TEST);
	//glClearDepth 1
	glDepthFunc (GL_LEQUAL);
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glEnable (GL_BLEND);
	glDisable (GL_COLOR_MATERIAL);

	//glEnable GL_LIGHTING    'enable lighting

	// NOTE: wglGetProcAddress fails if GL isn't initialized!

	// Get EXT functions
	glActiveTextureARB = (PFNGLACTIVETEXTUREARBPROC)wglGetProcAddress("glActiveTextureARB");
	glMultiTexCoord2f = (PFNGLMULTITEXCOORD2FARBPROC)wglGetProcAddress("glMultiTexCoord2f");

	// Off-screen renderbuffer functions
	glGenFramebuffersEXT = (PFNGLGENFRAMEBUFFERSEXTPROC)wglGetProcAddress("glGenFramebuffersEXT");
	glBindFramebufferEXT = (PFNGLBINDFRAMEBUFFEREXTPROC)wglGetProcAddress("glBindFramebufferEXT");
	glGenRenderbuffersEXT = (PFNGLGENRENDERBUFFERSEXTPROC)wglGetProcAddress("glGenRenderbuffersEXT");
	glBindRenderbufferEXT = (PFNGLBINDRENDERBUFFEREXTPROC)wglGetProcAddress("glBindRenderbufferEXT");
	glRenderbufferStorageEXT = (PFNGLRENDERBUFFERSTORAGEEXTPROC)wglGetProcAddress("glRenderbufferStorageEXT");
	glFramebufferRenderbufferEXT = (PFNGLFRAMEBUFFERRENDERBUFFEREXTPROC)wglGetProcAddress("glFramebufferRenderbufferEXT");
	glFramebufferTexture2DEXT = (PFNGLFRAMEBUFFERTEXTURE2DEXTPROC)wglGetProcAddress("glFramebufferTexture2DEXT");
	glCheckFramebufferStatusEXT = (PFNGLCHECKFRAMEBUFFERSTATUSEXTPROC)wglGetProcAddress("glCheckFramebufferStatusEXT");
	glDeleteFramebuffersEXT = (PFNGLDELETEFRAMEBUFFERSEXTPROC)wglGetProcAddress("glDeleteFramebuffersEXT");
	glDeleteRenderbuffersEXT = (PFNGLDELETERENDERBUFFERSEXTPROC)wglGetProcAddress("glDeleteRenderbuffersEXT");

	// Now set up the dynamic off-screen lightmap texture
	APITextures.GetDynamicLightmap()->SetupFrameBuffer(APITYPE_OPENGL, DYNAMICLIGHTMAPDIMENSION, DYNAMICLIGHTMAPDIMENSION);
  
    glGenFramebuffersEXT (1, &dynLMFB);
    glGenRenderbuffersEXT (1, &dynLMDB);
    
    if ((dynLMFB == 0) || (dynLMDB == 0))
	{
        glDeleteFramebuffersEXT (1, &dynLMFB);
        glDeleteRenderbuffersEXT (1, &dynLMDB);
        dynLMFB = 0;
        dynLMDB = 0;
        APITextures.GetDynamicLightmap()->SetValid(false);

		TRACE("* WARNING: Unable to create offscreen frame buffer or depth buffer\n");
	}
	else
	{
        glBindRenderbufferEXT (GL_RENDERBUFFER_EXT, dynLMDB);
		long dlmWidth, dlmHeight;
		APITextures.GetDynamicLightmap()->GetDimensions(dlmWidth, dlmHeight);
        glRenderbufferStorageEXT (GL_RENDERBUFFER_EXT, GL_DEPTH_COMPONENT16, dlmWidth, dlmHeight);
        glBindRenderbufferEXT (GL_RENDERBUFFER_EXT, 0);
        
        glBindFramebufferEXT (GL_FRAMEBUFFER_EXT, dynLMFB);
        
        // Attach texture
        glFramebufferTexture2DEXT (GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, APITextures.GetDynamicLightmap()->GetGLid(), 0);
        
        // Attach depth buffer
        glFramebufferRenderbufferEXT (GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT, GL_RENDERBUFFER_EXT, dynLMDB);
        
        // NOTE: If you attach the texture AFTER the depth buffer on ATI cards,
        //   the app will bomb.  So make sure attaching the depth buffer comes last
    
		if (glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT) != GL_FRAMEBUFFER_COMPLETE_EXT)	
		{
			TRACE("* Warning: Something went wrong during the offscreen buffer setup\n");
			glDeleteFramebuffersEXT (1, &dynLMFB);
			glDeleteRenderbuffersEXT (1, &dynLMDB);
			dynLMFB = 0;
			dynLMDB = 0;
			APITextures.GetDynamicLightmap()->SetValid(false);
		}
		else
		{
			TRACE("* Debug: Offscreen setup complete\n");
		}

        // Make sure the main context has the render focus
        glBindFramebufferEXT (GL_FRAMEBUFFER_EXT, 0);

    }
    
    offScreenMode = false;

	//////////////////////////////////////////////////
}

void OpenGLRender::SetViewport(float p_angle)
{
	SetViewport(p_angle, 0.1f, 1000.0f);
}

void OpenGLRender::SetViewport(float p_angle, float p_nearZ, float p_farZ)
{
    glViewport(0, 0, m_windowWidth, m_windowHeight);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    // GL uses x/y and degrees, DX uses y/x and radians
    gluPerspective(p_angle, float(m_windowWidth) / float(m_windowHeight), p_nearZ, p_farZ);
    glMatrixMode(GL_MODELVIEW);
    
    m_scaleY = float((m_windowHeight / 2.0f) / tan(p_angle / 360.0f * 3.14159266f));
}

void OpenGLRender::Clear(long p_color)
{
	glClearColor(float((p_color) & 0xFF)/255.0f, float((p_color >> 8) & 0xFF)/255.0f, float((p_color >> 16) & 0xFF)/255.0f, 0.0f);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
}

void OpenGLRender::LoadIdentity(void)
{
	glLoadIdentity();
}

void OpenGLRender::PushMatrix(void)
{
	glPushMatrix();
}

void OpenGLRender::PopMatrix(void)
{
	glPopMatrix();
}

void OpenGLRender::Translatef(float p_x, float p_y, float p_z)
{
	glTranslatef(p_x, p_y, p_z);
}

void OpenGLRender::Scalef(float p_x, float p_y, float p_z)
{
	glScalef(p_x, p_y, p_z);
}

void OpenGLRender::Rotate(float p_angle, float p_x, float p_y, float p_z)
{
	glRotatef(p_angle, p_x, p_y, p_z);
}

COrient3D OpenGLRender::GetViewpoint(void)
{
	float matrix[16];
	COrient3D vpTemp;

	glGetFloatv(GL_MODELVIEW_MATRIX, matrix);

	vpTemp.l.SetVec(matrix[0], matrix[4], matrix[8]);
	vpTemp.u.SetVec(matrix[1], matrix[5], matrix[9]);
	vpTemp.f.SetVec(matrix[2], matrix[6], matrix[10]);
	vpTemp.p.SetVec(matrix[12], matrix[13], matrix[14]);

	vpTemp.p = (vpTemp.l * -vpTemp.p.x) + (vpTemp.u * -vpTemp.p.y) + (vpTemp.f * -vpTemp.p.z);
	vpTemp.f = vpTemp.f * -1.0f;
	vpTemp.l = vpTemp.l * -1.0f;

	return vpTemp;
}

void OpenGLRender::BeginScene(void){}

void OpenGLRender::RenderTestTriangle(void)
{
	glBegin(GL_TRIANGLES);
		glColor3ub(255,0,0);
		glVertex2f(1.0f, 1.0f);
		glColor3ub(0,255,0);
		glVertex2f(-1.0f, 1.0f);
		glColor3ub(0,0,255);
		glVertex2f(-1.0f, -1.0f);
	glEnd();
}

void OpenGLRender::EndScene(void) {}

void OpenGLRender::Flip(void)
{
	SwapBuffers(m_hDC);
}

void OpenGLRender::Destroy(void)
{
	APITextures.ReleaseAPIAll();

	if (dynLMFB)
	{
		glDeleteFramebuffersEXT(1, &dynLMFB);
		dynLMFB = 0;
	}
	if (dynLMDB)
	{
		glDeleteRenderbuffersEXT(1, &dynLMDB);
		dynLMDB = 0;
	}
	APITextures.GetDynamicLightmap()->SetValid(false);

	if (m_hRC != NULL)
	{
		wglMakeCurrent(m_hDC, NULL);
		wglDeleteContext(m_hRC);
		m_hRC = NULL;
	}

	if (m_hDC != NULL)
	{
		ReleaseDC(m_hWnd, m_hDC);
		m_hDC = NULL;
	}
}

void OpenGLRender::TransformWorld(COrient3D &p_orient)
{
	// Initialize a matrix
	float matrix[] = { 1.0f, 0.0f, 0.0f, 0.0f,
						0.0f, 1.0f, 0.0f, 0.0f,
						0.0f, 0.0f, 1.0f, 0.0f,
						0.0f, 0.0f, 0.0f, 1.0f};

	CopyMemory(&matrix[0], &p_orient.l.x, 12);
	CopyMemory(&matrix[4], &p_orient.u.x, 12);
	CopyMemory(&matrix[8], &p_orient.f.x, 12);
	CopyMemory(&matrix[12], &p_orient.p.x, 12);

	glMultMatrixf(matrix);
}

void OpenGLRender::TransformViewpoint(COrient3D &p_orient)
{
	// Initialize a matrix
	float matrix[] = { 1.0f, 0.0f, 0.0f, 0.0f,
						0.0f, 1.0f, 0.0f, 0.0f,
						0.0f, 0.0f, 1.0f, 0.0f,
						0.0f, 0.0f, 0.0f, 1.0f};

	matrix[0] = p_orient.l.x;
	matrix[1] = p_orient.u.x;
	matrix[2] = p_orient.f.x;
	matrix[4] = p_orient.l.y;
	matrix[5] = p_orient.u.y;
	matrix[6] = p_orient.f.y;
	matrix[8] = p_orient.l.z;
	matrix[9] = p_orient.u.z;
	matrix[10] = p_orient.f.z;

	glMultMatrixf(matrix);

    glTranslatef(-p_orient.p.x, -p_orient.p.y, -p_orient.p.z);
}

bool OpenGLRender::ConvertEyeVectorToViewportXY(CVector3D &p_vector, int &p_xOut, int &p_yOut)
{
	if (p_vector.z > 0.00001f)
	{
		p_xOut = (int)(m_windowWidth / 2.0f - m_scaleY * p_vector.x / (p_vector.z));
		p_yOut = (int)(m_windowHeight / 2.0f - m_scaleY * p_vector.y / (p_vector.z));

		if ((fabs(p_xOut) < 32000) && (fabs(p_yOut) < 32000))
			return true;
		else
			return false;
	}
	else
		return false;
}

CVector3D OpenGLRender::ConvertViewportXYToWorldVector3D(int p_x, int p_y)
{
	float matrix[16];
	COrient3D viewpoint;
	CVector3D result;

	float x, y;
	x = (float(m_windowWidth) / 2.0f - float(p_x)) / float(m_scaleY);
	y = (float(m_windowHeight) / 2.0f - float(p_y)) / float(m_scaleY);

	glGetFloatv (GL_MODELVIEW_MATRIX, matrix);

	viewpoint.l.SetVec(matrix[0], matrix[4], matrix[8]);
	viewpoint.u.SetVec(matrix[1], matrix[5], matrix[9]);
	viewpoint.f.SetVec(matrix[2], matrix[6], matrix[10]);
	// viewpoint.p.SetVec(matrix[12], matrix[13], matrix[14]);  // Don't need this part

	result.SetVec(-viewpoint.f.x - viewpoint.l.x * x + viewpoint.u.x * y,
					-viewpoint.f.y - viewpoint.l.y * x + viewpoint.u.y * y,
					-viewpoint.f.z - viewpoint.l.z * x + viewpoint.u.z * y);
	result.Unitize();

	return result;
}

///////////////////////////////////////////////////
// Structure render routines

// p_upperLod and the p_hi and lo X and Z variables are used to connect
//  a lower Lod terrain with higher Lod terrains.  p_upperLod is the higher detail terrain
void OpenGLRender::RenderTerrain(CTerrain &p_terrain, bool p_useTestColor, bool p_useTexture, CTerrain *p_upperLod, bool p_loX, bool p_loZ, bool p_hiX, bool p_hiZ)
{

	unsigned x, z;
	BYTE color;
	BYTE alpha;
	bool texEnabled0, texEnabled1;
	float s, t;
	float s2, t2;

	CTerrainPoint *points;

	points = p_terrain.GetPoints();

	SetCullCCW(!cullCCW);

	float shade = p_terrain.shade;
	float invShade = 255.0f-p_terrain.shade;

	// Get texturing ready, assume we come in with texturing off
	texEnabled0 = false;
	if ((p_terrain.textureOp == TERRAINRENDER_SINGLETEXTURE) || (p_terrain.textureOp == TERRAINRENDER_TWOTEXTUREBLENDALPHA))
	{
		if ((p_terrain.texture0) && (p_useTexture))
		{
			if (p_terrain.texture0->IsValid())
			{
				if (!p_terrain.texture0->IsLoaded())
					p_terrain.texture0->LoadTextureGL();
            
				if (p_terrain.texture0->IsValid())
				{
					texEnabled0 = true;
					glActiveTextureARB (GL_TEXTURE0_ARB);
					glEnable (GL_TEXTURE_2D);
					glBindTexture (GL_TEXTURE_2D, p_terrain.texture0->GetGLid());
                
					glActiveTextureARB (GL_TEXTURE1_ARB);
					glDisable (GL_TEXTURE_2D);
				}
			}
		}
	}

	texEnabled1 = false;
	if ((p_terrain.textureOp == TERRAINRENDER_TWOTEXTUREBLENDALPHA) && (texEnabled0))
	{
		if ((p_terrain.texture1) && (p_useTexture))
		{
			if (p_terrain.texture1->IsValid())
			{
				if (!p_terrain.texture1->IsLoaded())
					p_terrain.texture1->LoadTextureGL();
            
				if (p_terrain.texture1->IsValid())
				{
					texEnabled1 = true;
					glActiveTextureARB (GL_TEXTURE1_ARB);
					glEnable (GL_TEXTURE_2D);
					glBindTexture (GL_TEXTURE_2D, p_terrain.texture1->GetGLid());
                
					glActiveTextureARB (GL_TEXTURE2_ARB);
					glDisable (GL_TEXTURE_2D);
				}
			}
		}
	}

	// Set up texture operations
	switch (p_terrain.textureOp)
	{
	case TERRAINRENDER_SINGLETEXTURE:
		if (texEnabled0)
		{
			glActiveTextureARB (GL_TEXTURE0_ARB);
			glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
		}
		break;
	case TERRAINRENDER_TWOTEXTUREBLENDALPHA:
		if ((texEnabled0) && (texEnabled1))
		{
			glActiveTextureARB (GL_TEXTURE0_ARB);
			glEnable (GL_TEXTURE_2D);
			glBindTexture (GL_TEXTURE_2D, p_terrain.texture0->GetGLid());
			// Make texture partially T1 and partially T2 guided by vertex alpha
			glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE);
			glTexEnvi (GL_TEXTURE_ENV, GL_COMBINE_RGB_ARB, GL_INTERPOLATE_ARB);
			glTexEnvi (GL_TEXTURE_ENV, GL_SOURCE0_RGB_ARB, GL_TEXTURE1_ARB);
			glTexEnvi (GL_TEXTURE_ENV, GL_OPERAND0_RGB_ARB, GL_SRC_COLOR);
			glTexEnvi (GL_TEXTURE_ENV, GL_SOURCE1_RGB_ARB, GL_TEXTURE);
			glTexEnvi (GL_TEXTURE_ENV, GL_OPERAND1_RGB_ARB, GL_SRC_COLOR);
			glTexEnvi (GL_TEXTURE_ENV, GL_SOURCE2_RGB_ARB, GL_PRIMARY_COLOR);
			glTexEnvi (GL_TEXTURE_ENV, GL_OPERAND2_RGB_ARB, GL_SRC_ALPHA);
			glTexEnvi (GL_TEXTURE_ENV, GL_COMBINE_ALPHA_ARB, GL_REPLACE);
			glTexEnvi (GL_TEXTURE_ENV, GL_SOURCE0_ALPHA_ARB, GL_TEXTURE);
			glTexEnvi (GL_TEXTURE_ENV, GL_OPERAND0_ALPHA_ARB, GL_SRC_ALPHA);
        
			// Modulate with vertex color
			glActiveTextureARB (GL_TEXTURE1_ARB);
			glEnable (GL_TEXTURE_2D);
			glBindTexture (GL_TEXTURE_2D, p_terrain.texture1->GetGLid());
			glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE);
			glTexEnvi (GL_TEXTURE_ENV, GL_COMBINE_RGB_ARB, GL_MODULATE);
			glTexEnvi (GL_TEXTURE_ENV, GL_SOURCE0_RGB_ARB, GL_PREVIOUS);
			glTexEnvi (GL_TEXTURE_ENV, GL_OPERAND0_RGB_ARB, GL_SRC_COLOR);
			glTexEnvi (GL_TEXTURE_ENV, GL_SOURCE1_RGB_ARB, GL_PRIMARY_COLOR);
			glTexEnvi (GL_TEXTURE_ENV, GL_OPERAND1_RGB_ARB, GL_SRC_COLOR);
			glTexEnvi (GL_TEXTURE_ENV, GL_COMBINE_ALPHA_ARB, GL_REPLACE);
			glTexEnvi (GL_TEXTURE_ENV, GL_SOURCE0_ALPHA_ARB, GL_TEXTURE0_ARB);
			glTexEnvi (GL_TEXTURE_ENV, GL_OPERAND0_ALPHA_ARB, GL_SRC_ALPHA);
        
			glActiveTextureARB (GL_TEXTURE2_ARB);
			glDisable (GL_TEXTURE_2D);
//            glBindTexture GL_TEXTURE_2D, p_terrain.texture2.GLid
		}
	}

	/*
	'If p_terrain.GLDisplayList <> 0 Then
	'    glCallList p_terrain.GLDisplayList
	'    GoTo done
	'Else
	'    p_terrain.GLDisplayList = glGenLists(1)
	'    glNewList p_terrain.GLDisplayList, GL_COMPILE_AND_EXECUTE
	'End If*/

	CTerrainPoint *lpPoint0;
	CTerrainPoint *lpPoint1;

	//glDepthFunc(GL_ALWAYS);

	unsigned int minX, maxX;
	unsigned int minZ, maxZ;
	minZ = 0; maxZ = p_terrain.zQty-1;
	minX = 0; maxX = p_terrain.xQty-1;
	if (p_loX) minX++;
	if (p_loZ) minZ++;
	if (p_hiX) maxX--;
	if (p_hiZ) maxZ--;

	// Initialize texture coordinates (even if we aren't using it)
	t = p_terrain.startT0 + minZ*p_terrain.texCoordPerCell0;
	t2 = p_terrain.startT1 + minZ*p_terrain.texCoordPerCell1;

	for (z=minZ; z<maxZ; z++)
	{

		glBegin (GL_TRIANGLE_STRIP);
    
		color = 0;
		s = p_terrain.startS0 + minX*p_terrain.texCoordPerCell0;
		s2 = p_terrain.startS1 + minX*p_terrain.texCoordPerCell1;

		lpPoint0 = &points[minX + z*p_terrain.xQty];
		lpPoint1 = &points[minX + (z+1)*p_terrain.xQty];
    
		for (x=minX; x<=maxX; x++)
		{

			// Render (x, z), (x, z+1), (x+1, z)
			// Then (x+1, z+1)
			//??
			if (p_useTestColor)
			{
				color = color + 1;
				if (color > 3) color = 0;
				switch (color)
				{
				case 0:
					glColor3ub (255, 0, 0);
					break;
				case 1:
					glColor3ub (0, 255, 0);
					break;
				case 2:
					glColor3ub (0, 0, 255);
					break;
				case 3:
					glColor3ub (255, 255, 255);
					break;
				}
			}
			else
			{
				//color = lpPoint0->color;
				color = BYTE(shade + invShade*lpPoint0->illumination);
				if (texEnabled1)
				{
					alpha = lpPoint0->alpha1;
					glColor4ub (color, color, color, alpha);
				}
				else
					glColor3ub (color, color, color);
			}
			if (texEnabled0)
			{
				glMultiTexCoord2f(GL_TEXTURE0_ARB, s, t);
				if (texEnabled1)
					glMultiTexCoord2f(GL_TEXTURE1_ARB, s2, t2);
			}
			glVertex3f (x * p_terrain.cellSize, lpPoint0->y, (z) * p_terrain.cellSize);
    
			if (p_useTestColor)
			{
				color = color + 1;
				if (color > 3) color = 0;
				switch (color)
				{
				case 0:
					glColor3ub (255, 0, 0);
					break;
				case 1:
					glColor3ub (0, 255, 0);
					break;
				case 2:
					glColor3ub (0, 0, 255);
					break;
				case 3:
					glColor3ub (255, 255, 255);
					break;
				}
			}
			else
			{
				//color = lpPoint1->color;
				color = BYTE(shade + invShade*lpPoint1->illumination);
				if (texEnabled1)
				{
					alpha = lpPoint1->alpha1;
					glColor4ub (color, color, color, alpha);
				}
				else
					glColor3ub (color, color, color);
			}
			if (texEnabled0)
			{
				glMultiTexCoord2f(GL_TEXTURE0_ARB, s, t + p_terrain.texCoordPerCell0);
				if (texEnabled1)
					glMultiTexCoord2f(GL_TEXTURE1_ARB, s2, t2 + p_terrain.texCoordPerCell1);
			}
			glVertex3f (x * p_terrain.cellSize, lpPoint1->y, (z+1) * p_terrain.cellSize);
    
			if (texEnabled0)
			{
				s = s + p_terrain.texCoordPerCell0;
				if (texEnabled1)
					s2 = s2 + p_terrain.texCoordPerCell1;
			}

			lpPoint0++;
			lpPoint1++;

		}	// for x
    
		glEnd();
    
		if (texEnabled0)
		{
			t = t + p_terrain.texCoordPerCell0;
			if (texEnabled1)
				t2 = t2 + p_terrain.texCoordPerCell1;
		}
	}	// for z

	//If p_terrain.GLDisplayList <> 0 Then
	//    glEndList
	//End If

//	done:

	// Connect terrain with higher LOD's surrounding it, if necessary
	if (p_upperLod)
	{
		CTerrainPoint *upperPoints = p_upperLod->GetPoints();
		BYTE color;
		BYTE alpha;
		float s, t;
		float s2,t2;
		CTerrainPoint *lpLowPt, *lpHighPt;
		unsigned int i;

		// lower X strip  (right)
		// Make triangle fans
		// 4 ----- 3
		//   |  /|
		//   | /_| 2
		//   |/  |
		// 0 ----- 1
		// 0 and 4 are from lower lod, 1-3 from higher
		if (p_loX)
		{
			minZ = 0;
			maxZ = p_terrain.zQty-1;
			if (p_loZ) minZ++;
			if (p_hiZ) maxZ--;

			for (z=minZ; z<maxZ; z++)
			{
				lpLowPt = &points[1+z*p_terrain.xQty];
				lpHighPt = &upperPoints[0+(z*2*p_upperLod->xQty)];
				glBegin(GL_TRIANGLE_FAN);

				// 0
				color = BYTE(shade + invShade*lpLowPt->illumination);
				alpha = lpLowPt->alpha1;
				// tex coords ??
				glColor4ub(color, color, color, alpha);
				glVertex3f(p_terrain.cellSize, lpLowPt->y, z*p_terrain.cellSize);

				// 1, 2, 3
				for (i=0; i<3; i++)
				{
					color = BYTE(shade + invShade*lpHighPt->illumination);
					alpha = lpHighPt->alpha1;
					// tex coords ??
					glColor4ub(color, color, color, alpha);
					glVertex3f(0.0f, lpHighPt->y, z*p_terrain.cellSize+(p_terrain.cellSize*i/2));

					lpHighPt += p_upperLod->xQty;
				}

				// 4
				lpLowPt += p_terrain.xQty;
				color = BYTE(shade + invShade*lpLowPt->illumination);
				alpha = lpLowPt->alpha1;
				// tex coords ??
				glColor4ub(color, color, color, alpha);
				glVertex3f(p_terrain.cellSize, lpLowPt->y, (z+1)*p_terrain.cellSize);

				glEnd();

				Metrics.triangles += 5;
			}
		}

		// upper X strip  (left)
		// Make triangle fans
		// 1 ----- 0
		//   |_ /|
		// 2 | / | 
		//   |/  |
		// 3 ----- 4
		// 0 and 4 are from lower lod, 1-3 from higher
		if (p_hiX)
		{
			minZ = 0;
			maxZ = p_terrain.zQty-1;
			if (p_loZ) minZ++;
			if (p_hiZ) maxZ--;

			for (z=minZ; z<maxZ; z++)
			{
				lpLowPt = &points[(p_terrain.xQty-2)+(z+1)*p_terrain.xQty];
				lpHighPt = &upperPoints[(p_upperLod->xQty-1)+((z+1)*2*p_upperLod->xQty)];
				glBegin(GL_TRIANGLE_FAN);

				// 0
				color = BYTE(shade + invShade*lpLowPt->illumination);
				alpha = lpLowPt->alpha1;
				// tex coords ??
				glColor4ub(color, color, color, alpha);
				glVertex3f((p_terrain.xQty-2)*p_terrain.cellSize, lpLowPt->y, (z+1)*p_terrain.cellSize);

				// 1, 2, 3
				for (i=0; i<3; i++)
				{
					color = BYTE(shade + invShade*lpHighPt->illumination);
					alpha = lpHighPt->alpha1;
					// tex coords ??
					glColor4ub(color, color, color, alpha);
					glVertex3f((p_terrain.xQty-1)*p_terrain.cellSize, lpHighPt->y, (z+1)*p_terrain.cellSize-(p_terrain.cellSize*i/2));

					lpHighPt -= p_upperLod->xQty;
				}

				// 4
				lpLowPt -= p_terrain.xQty;
				color = BYTE(shade + invShade*lpLowPt->illumination);
				alpha = lpLowPt->alpha1;
				// tex coords ??
				glColor4ub(color, color, color, alpha);
				glVertex3f((p_terrain.xQty-2)*p_terrain.cellSize, lpLowPt->y, z*p_terrain.cellSize);

				glEnd();

				Metrics.triangles += 5;
			}
		}

		// lower Z strip
		// Make triangle fans
		// 1 ----- 0
		//   |  /|
		//   | / | 
		//   |/| |
		//   ----- 
		//  2  3  4
		// 0 and 1 are from lower lod, 2-4 from higher
		if (p_loZ)
		{
			minX = 0;
			maxX = p_terrain.xQty-1;
			if (p_loX) minX++;
			if (p_hiX) maxX--;

			for (x=minX; x<maxX; x++)
			{
				lpLowPt = &points[x+1*p_terrain.xQty];
				lpHighPt = &upperPoints[(x+1)*2];

				glBegin(GL_TRIANGLE_FAN);

				// 0
				color = BYTE(shade + invShade*lpLowPt->illumination);
				alpha = lpLowPt->alpha1;
				// tex coords ??
				glColor4ub(color, color, color, alpha);
				glVertex3f(x*p_terrain.cellSize, lpLowPt->y, p_terrain.cellSize);

				// 1
				lpLowPt++;
				color = BYTE(shade + invShade*lpLowPt->illumination);
				alpha = lpLowPt->alpha1;
				// tex coords ??
				glColor4ub(color, color, color, alpha);
				glVertex3f((x+1)*p_terrain.cellSize, lpLowPt->y, p_terrain.cellSize);

				// 2, 3, 4
				for (i=0; i<3; i++)
				{
					color = BYTE(shade + invShade*lpHighPt->illumination);
					alpha = lpHighPt->alpha1;
					// tex coords ??
					glColor4ub(color, color, color, alpha);
					glVertex3f((x+1)*p_terrain.cellSize-(i*p_terrain.cellSize/2), lpHighPt->y, 0.0f);

					lpHighPt--;
				}

				glEnd();

				Metrics.triangles += 5;
			}
		}

		// upper Z strip
		// Make triangle fans
		//  4  3  2
		//   -----
		//   | |/|
		//   | / | 
		//   |/  |
		// 0 ----- 1
		// 0 and 1 are from lower lod, 2-4 from higher
		if (p_hiZ)
		{
			minX = 0;
			maxX = p_terrain.xQty-1;
			if (p_loX) minX++;
			if (p_hiX) maxX--;

			for (x=minX; x<maxX; x++)
			{
				lpLowPt = &points[(x+1) + (p_terrain.zQty-2)*p_terrain.xQty];
				lpHighPt = &upperPoints[x*2+(p_upperLod->zQty-1)*p_upperLod->xQty];

				glBegin(GL_TRIANGLE_FAN);

				// 0
				color = BYTE(shade + invShade*lpLowPt->illumination);
				alpha = lpLowPt->alpha1;
				// tex coords ??
				glColor4ub(color, color, color, alpha);
				glVertex3f((x+1)*p_terrain.cellSize, lpLowPt->y, (p_terrain.zQty-2)*p_terrain.cellSize);

				// 1
				lpLowPt--;
				color = BYTE(shade + invShade*lpLowPt->illumination);
				alpha = lpLowPt->alpha1;
				// tex coords ??
				glColor4ub(color, color, color, alpha);
				glVertex3f(x*p_terrain.cellSize, lpLowPt->y, (p_terrain.zQty-2)*p_terrain.cellSize);

				// 2, 3, 4
				for (i=0; i<3; i++)
				{
					color = BYTE(shade + invShade*lpHighPt->illumination);
					alpha = lpHighPt->alpha1;
					// tex coords ??
					glColor4ub(color, color, color, alpha);
					glVertex3f(x*p_terrain.cellSize+(i*p_terrain.cellSize/2), lpHighPt->y, (p_terrain.zQty-1)*p_terrain.cellSize);

					lpHighPt++;
				}

				glEnd();

				Metrics.triangles += 5;
			}
		}

		// ?? loX loZ corner
		if (p_loX && p_loZ)
		{
		}

		// ?? hiX loZ corner
		if (p_hiX && p_loZ)
		{
		}

		// ?? loX hiZ corner
		if (p_loX && p_hiZ)
		{
		}

		// ?? hiX hiZ corner
		if (p_hiX && p_hiZ)
		{
		}
	}

	//glDepthFunc(GL_LEQUAL);
	SetCullCCW(!cullCCW);
	Metrics.triangles += (p_terrain.xQty * p_terrain.zQty * 2);

	if ((texEnabled0) || (texEnabled1))
	{
		glActiveTextureARB (GL_TEXTURE2_ARB);
		glDisable (GL_TEXTURE_2D);
		glActiveTextureARB (GL_TEXTURE1_ARB);
		glDisable (GL_TEXTURE_2D);
		glActiveTextureARB (GL_TEXTURE0_ARB);
		glDisable (GL_TEXTURE_2D);
		texEnabled0 = false;
		texEnabled1 = false;
	}
}

void OpenGLRender::RenderQuad(CColorRGB p_color, BYTE p_alpha, CAPITexture *p_texture)
{
	bool texEnabled;

    texEnabled = false;
    if (p_texture)
	{
        if (p_texture->IsValid())
		{
            if (!p_texture->IsLoaded())
                p_texture->LoadTextureGL();
            
            if (p_texture->IsValid())
			{
                texEnabled = true;
                
                glActiveTextureARB (GL_TEXTURE1_ARB);
                glBindTexture (GL_TEXTURE_2D, 0);
                glDisable (GL_TEXTURE_2D);
                
                glActiveTextureARB (GL_TEXTURE0_ARB);
                glEnable (GL_TEXTURE_2D);
                glBindTexture (GL_TEXTURE_2D, p_texture->GetGLid());
                //glTexParameterf GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP
                //glTexParameterf GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP
                glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
                if (p_texture->GetMipMap())
                    glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
                else
                    glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
                glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
                glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
                //glTexParameteri GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP
                //glTexParameteri GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP
                glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
                
                /*if (p_textureOrient)
                    glMatrixMode mmTexture
                    Call gluPerspective(90, 1, 0.1, 1000)
                    Call IRender_TransformWorld(p_textureOrient)
                    glTranslatef -5, -5, -1
                    'glTranslatef 2, 2, -1
                    'glTranslatef 0, 0, -1
                    glScalef 10, 10, 1
                    glMatrixMode mmModelView
                End If*/
            }
        }
    }

	Metrics.triangles += 2;
    glBegin (GL_QUADS);
		if (texEnabled)
		{
			if (p_alpha < 255)
				glColor4ub (p_color.red, p_color.green, p_color.blue, p_alpha);
			else
				glColor3ub (p_color.red, p_color.green, p_color.blue);
			glMultiTexCoord2f(GL_TEXTURE0_ARB, 0.0f, 1.0f);
			//glVertex3f (1.0f, 1.0f, 0.0f);
			glVertex3i(1,1,0);
			glMultiTexCoord2f(GL_TEXTURE0_ARB, 1.0f, 1.0f);
			//glVertex3f (-1.0f, 1.0f, 0.0f);
			glVertex3i(-1,1,0);
			glMultiTexCoord2f(GL_TEXTURE0_ARB, 1.0f, 0.0f);
			//glVertex3f (-1.0f, -1.0f, 0.0f);
			glVertex3i(-1,-1,0);
			glMultiTexCoord2f(GL_TEXTURE0_ARB, 0.0f, 0.0f);
			//glVertex3f (1.0f, -1.0f, 0.0f);
			glVertex3i(1,-1,0);
		}
		else
		{
			if (p_alpha < 255)
				glColor4ub (p_color.red, p_color.green, p_color.blue, p_alpha);
			else
				glColor3ub (p_color.red, p_color.green, p_color.blue);
			//glVertex3f (1.0f, 1.0f, 0.0f);
			glVertex3i(1,1,0);
			//glVertex3f (-1.0f, 1.0f, 0.0f);
			glVertex3i(-1,1,0);
			//glVertex3f (-1.0f, -1.0f, 0.0f);
			glVertex3i(-1,-1,0);
			//glVertex3f (1.0f, -1.0f, 0.0f);
			glVertex3i(1,-1,0);
		}
    glEnd();
    
    if (texEnabled)
	{
        //glExt.glActiveTextureARB GL_TEXTURE0_ARB
        //glDisable GL_TEXTURE_2D
        //texEnabled = false;
        
        glActiveTextureARB (GL_TEXTURE0_ARB);
        glBindTexture (GL_TEXTURE_2D, 0);
        glDisable (GL_TEXTURE_2D);
        
        /*if (p_textureOrient)
            glMatrixMode mmTexture
            glLoadIdentity
            glMatrixMode mmModelView
        End If*/
	}
}

void OpenGLRender::RenderMultiQuad(CColorRGB p_color, BYTE p_alpha, CAPITexture *p_texture1, CAPITexture *p_texture2, defMultiQuadTextureOp p_textureOp)
{
	bool texEnabled1;
	bool texEnabled2;

    // Check for texture validity
    texEnabled1 = false;
    if (p_texture1)
		if (p_texture1->IsValid())
		{
            if (!p_texture1->IsLoaded())
                p_texture1->LoadTextureGL();
            
            if (p_texture1->IsValid())
                texEnabled1 = true;
        }

    texEnabled2 = false;
    if (p_texture2)
		if (p_texture2->IsValid())
		{
            if (!p_texture2->IsLoaded())
                p_texture2->LoadTextureGL();
            
            if (p_texture2->IsValid())
                texEnabled2 = true;
        }
    
    // If tex1 isn't enabled, just render a color quad
    if (!texEnabled1)
	{
        RenderQuad(p_color, p_alpha);
		return;
	}
    
    // Let's render a quad
    // Can either render with just tex1 or tex1 and tex2
    
    // Set up the texture pipeline
    glActiveTextureARB (GL_TEXTURE0_ARB);
    glEnable(GL_TEXTURE_2D);
    glBindTexture (GL_TEXTURE_2D, p_texture1->GetGLid());
    glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    if (p_texture1->GetMipMap())
        glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
    else
        glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
    glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
    //glTexParameteri GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP
    //glTexParameteri GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP
    glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE);
    switch (p_textureOp)
	{
	case APIMULTIQUADTEXTUREOP_MODULATE:
        // Just multiply them, including alpha
        if (texEnabled2)
		{
            glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE);
            glTexEnvi (GL_TEXTURE_ENV, GL_COMBINE_RGB_ARB, GL_MODULATE);
            glTexEnvi (GL_TEXTURE_ENV, GL_SOURCE0_RGB_ARB, GL_TEXTURE1_ARB);
            glTexEnvi (GL_TEXTURE_ENV, GL_OPERAND0_RGB_ARB, GL_SRC_COLOR);
            glTexEnvi (GL_TEXTURE_ENV, GL_SOURCE1_RGB_ARB, GL_TEXTURE);
            glTexEnvi (GL_TEXTURE_ENV, GL_OPERAND1_RGB_ARB, GL_SRC_COLOR);
            glTexEnvi (GL_TEXTURE_ENV, GL_COMBINE_ALPHA_ARB, GL_MODULATE);
            glTexEnvi (GL_TEXTURE_ENV, GL_SOURCE0_ALPHA_ARB, GL_TEXTURE1_ARB);
            glTexEnvi (GL_TEXTURE_ENV, GL_OPERAND0_ALPHA_ARB, GL_SRC_ALPHA);
            glTexEnvi (GL_TEXTURE_ENV, GL_SOURCE1_ALPHA_ARB, GL_TEXTURE);
            glTexEnvi (GL_TEXTURE_ENV, GL_OPERAND1_ALPHA_ARB, GL_SRC_ALPHA);
		}
        else
		{
            // Only tex1, so modulate with vertex
            glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE);
            glTexEnvi (GL_TEXTURE_ENV, GL_COMBINE_RGB_ARB, GL_MODULATE);
            glTexEnvi (GL_TEXTURE_ENV, GL_SOURCE0_RGB_ARB, GL_TEXTURE);
            glTexEnvi (GL_TEXTURE_ENV, GL_OPERAND0_RGB_ARB, GL_SRC_COLOR);
            glTexEnvi (GL_TEXTURE_ENV, GL_SOURCE1_RGB_ARB, GL_PRIMARY_COLOR);
            glTexEnvi (GL_TEXTURE_ENV, GL_OPERAND1_RGB_ARB, GL_SRC_COLOR);
            glTexEnvi (GL_TEXTURE_ENV, GL_COMBINE_ALPHA_ARB, GL_MODULATE);
            glTexEnvi (GL_TEXTURE_ENV, GL_SOURCE0_ALPHA_ARB, GL_TEXTURE);
            glTexEnvi (GL_TEXTURE_ENV, GL_OPERAND0_ALPHA_ARB, GL_SRC_ALPHA);
            glTexEnvi (GL_TEXTURE_ENV, GL_SOURCE1_ALPHA_ARB, GL_PRIMARY_COLOR);
            glTexEnvi (GL_TEXTURE_ENV, GL_OPERAND1_ALPHA_ARB, GL_SRC_ALPHA);
		}
		break;
	case APIMULTIQUADTEXTUREOP_BLENDALPHAMAPS:
		if (texEnabled2)
		{
            // Blend the two textures together guided by alpha
            // Used to combine two approximate light maps together to
            //   get a more accurate result
            glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE);
            // not really using the RGB components of the textures
            glTexEnvi (GL_TEXTURE_ENV, GL_COMBINE_RGB_ARB, GL_REPLACE);
            glTexEnvi (GL_TEXTURE_ENV, GL_SOURCE0_RGB_ARB, GL_TEXTURE);
            glTexEnvi (GL_TEXTURE_ENV, GL_OPERAND0_RGB_ARB, GL_SRC_ALPHA);
            glTexEnvi (GL_TEXTURE_ENV, GL_COMBINE_ALPHA_ARB, GL_INTERPOLATE_ARB);
            glTexEnvi (GL_TEXTURE_ENV, GL_SOURCE0_ALPHA_ARB, GL_TEXTURE1_ARB);
            glTexEnvi (GL_TEXTURE_ENV, GL_OPERAND0_ALPHA_ARB, GL_SRC_ALPHA);
            glTexEnvi (GL_TEXTURE_ENV, GL_SOURCE1_ALPHA_ARB, GL_TEXTURE);
            glTexEnvi (GL_TEXTURE_ENV, GL_OPERAND1_ALPHA_ARB, GL_SRC_ALPHA);
            glTexEnvi (GL_TEXTURE_ENV, GL_SOURCE2_ALPHA_ARB, GL_PRIMARY_COLOR);
            glTexEnvi (GL_TEXTURE_ENV, GL_OPERAND2_ALPHA_ARB, GL_SRC_ALPHA);
		}
        else
		{
            // Only tex1, so make sure our alpha is modulated with the vertex color
            glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE);
            // not really using the RGB components of the textures
            glTexEnvi (GL_TEXTURE_ENV, GL_COMBINE_RGB_ARB, GL_MODULATE);
            glTexEnvi (GL_TEXTURE_ENV, GL_SOURCE0_RGB_ARB, GL_TEXTURE);
            glTexEnvi (GL_TEXTURE_ENV, GL_OPERAND0_RGB_ARB, GL_SRC_ALPHA);
            glTexEnvi (GL_TEXTURE_ENV, GL_SOURCE1_RGB_ARB, GL_PRIMARY_COLOR);
            glTexEnvi (GL_TEXTURE_ENV, GL_OPERAND1_RGB_ARB, GL_SRC_COLOR);
            // Doesn't really matter what alpha becomes at this point
            glTexEnvi (GL_TEXTURE_ENV, GL_COMBINE_ALPHA_ARB, GL_REPLACE);
            glTexEnvi (GL_TEXTURE_ENV, GL_SOURCE0_ALPHA_ARB, GL_PRIMARY_COLOR);
            glTexEnvi (GL_TEXTURE_ENV, GL_OPERAND0_ALPHA_ARB, GL_SRC_ALPHA);
		}
		break;
    }

    if (texEnabled2)
	{
        glActiveTextureARB (GL_TEXTURE1_ARB);
        glEnable (GL_TEXTURE_2D);
        glBindTexture (GL_TEXTURE_2D, p_texture2->GetGLid());
        glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
        if (p_texture2->GetMipMap())
            glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
        else
            glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
        glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
        glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
        //glTexParameteri GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP
        //glTexParameteri GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP
        glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE);
        switch (p_textureOp)
		{
		case APIMULTIQUADTEXTUREOP_MODULATE:
            // Just multiply the result from texture step 0 with the vertex
            glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE);
            glTexEnvi (GL_TEXTURE_ENV, GL_COMBINE_RGB_ARB, GL_MODULATE);
            glTexEnvi (GL_TEXTURE_ENV, GL_SOURCE0_RGB_ARB, GL_PREVIOUS);
            glTexEnvi (GL_TEXTURE_ENV, GL_OPERAND0_RGB_ARB, GL_SRC_COLOR);
            glTexEnvi (GL_TEXTURE_ENV, GL_SOURCE1_RGB_ARB, GL_PRIMARY_COLOR);
            glTexEnvi (GL_TEXTURE_ENV, GL_OPERAND1_RGB_ARB, GL_SRC_COLOR);
            glTexEnvi (GL_TEXTURE_ENV, GL_COMBINE_ALPHA_ARB, GL_MODULATE);
            glTexEnvi (GL_TEXTURE_ENV, GL_SOURCE0_ALPHA_ARB, GL_PREVIOUS);
            glTexEnvi (GL_TEXTURE_ENV, GL_OPERAND0_ALPHA_ARB, GL_SRC_ALPHA);
            glTexEnvi (GL_TEXTURE_ENV, GL_SOURCE1_ALPHA_ARB, GL_PRIMARY_COLOR);
            glTexEnvi (GL_TEXTURE_ENV, GL_OPERAND1_ALPHA_ARB, GL_SRC_ALPHA);
			break;
		case APIMULTIQUADTEXTUREOP_BLENDALPHAMAPS:
            // Just multiply the resulting alphas from texture step 0 with the vertex color
            glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE);
            // not really using the RGB components of the textures
            glTexEnvi (GL_TEXTURE_ENV, GL_COMBINE_RGB_ARB, GL_MODULATE);
            glTexEnvi (GL_TEXTURE_ENV, GL_SOURCE0_RGB_ARB, GL_PREVIOUS);
            glTexEnvi (GL_TEXTURE_ENV, GL_OPERAND0_RGB_ARB, GL_SRC_ALPHA);
            glTexEnvi (GL_TEXTURE_ENV, GL_SOURCE1_RGB_ARB, GL_PRIMARY_COLOR);
            glTexEnvi (GL_TEXTURE_ENV, GL_OPERAND1_RGB_ARB, GL_SRC_COLOR);
            // Doesn't really matter what alpha becomes at this point,
            //   since this method should be used with a non-alpha using srcblend mode
            //   like ONE (add) or ONE MINUS DEST COLOR (tend)
            glTexEnvi (GL_TEXTURE_ENV, GL_COMBINE_ALPHA_ARB, GL_REPLACE);
            glTexEnvi (GL_TEXTURE_ENV, GL_SOURCE0_ALPHA_ARB, GL_TEXTURE);
            glTexEnvi (GL_TEXTURE_ENV, GL_OPERAND0_ALPHA_ARB, GL_SRC_ALPHA);
			break;
        }
        
        glActiveTextureARB (GL_TEXTURE2_ARB);
        glDisable (GL_TEXTURE_2D);
	}
    else
	{
        glActiveTextureARB (GL_TEXTURE1_ARB);
        glDisable (GL_TEXTURE_2D);
    }
        
    // Draw it
	Metrics.triangles += 2;
    if (texEnabled2)
	{
        glBegin(GL_QUADS);
            glColor4ub (p_color.red, p_color.green, p_color.blue, p_alpha);
            glMultiTexCoord2f (GL_TEXTURE0_ARB, 0.0f, 1.0f);
            glMultiTexCoord2f (GL_TEXTURE1_ARB, 0.0f, 1.0f);
            glVertex3f (1.0f, 1.0f, 0.0f);
            glMultiTexCoord2f (GL_TEXTURE0_ARB, 1.0f, 1.0f);
            glMultiTexCoord2f (GL_TEXTURE1_ARB, 1.0f, 1.0f);
            glVertex3f (-1.0f, 1.0f, 0.0f);
            glMultiTexCoord2f (GL_TEXTURE0_ARB, 1.0f, 0.0f);
            glMultiTexCoord2f (GL_TEXTURE1_ARB, 1.0f, 0.0f);
            glVertex3f (-1.0f, -1.0f, 0.0f);
            glMultiTexCoord2f (GL_TEXTURE0_ARB, 0.0f, 0.0f);
            glMultiTexCoord2f (GL_TEXTURE1_ARB, 0.0f, 0.0f);
            glVertex3f (1.0f, -1.0f, 0.0f);
        glEnd();
	}
    else
	{
        glBegin (GL_QUADS);
            glColor4ub (p_color.red, p_color.green, p_color.blue, p_alpha);
            glMultiTexCoord2f (GL_TEXTURE0_ARB, 0.0f, 1.0f);
            glVertex3f (1.0f, 1.0f, 0.0f);
            glMultiTexCoord2f (GL_TEXTURE0_ARB, 1.0f, 1.0f);
            glVertex3f (-1.0f, 1.0f, 0.0f);
            glMultiTexCoord2f (GL_TEXTURE0_ARB, 1.0f, 0.0f);
            glVertex3f (-1.0f, -1.0f, 0.0f);
            glMultiTexCoord2f (GL_TEXTURE0_ARB, 0.0f, 0.0f);
            glVertex3f (1.0f, -1.0f, 0.0f);
        glEnd();
    }
    
    // Disable texturing
    glActiveTextureARB (GL_TEXTURE1_ARB);
    glBindTexture (GL_TEXTURE_2D, 0);
    glDisable (GL_TEXTURE_2D);
    glActiveTextureARB (GL_TEXTURE0_ARB);
    glBindTexture (GL_TEXTURE_2D, 0);
    glDisable (GL_TEXTURE_2D);
}

void OpenGLRender::RenderModelFinal(CModel3D *p_model, CColorRGB *p_modColor, BYTE p_modAlpha)
{
	if (p_model)
	RenderSurfaces(p_model->GetSurfaces(), p_model->GetSurfaceQty(), p_model->GetVertices(), p_modColor, p_modAlpha);
}

void OpenGLRender::RenderModelFinalLM(CModel3D *p_model, CColorRGB *p_modColor, BYTE p_modAlpha, CAPITexture **p_lightmaps, CSurfaceDecalList *p_decalLists, CLightSourceRefList *p_lightSourceList, COrient3D *p_orient)
{
	if (p_model)
	RenderSurfaces(p_model->GetSurfaces(), p_model->GetSurfaceQty(), p_model->GetVertices(), p_modColor, p_modAlpha, p_lightmaps, p_decalLists, true, p_lightSourceList, NULL, p_orient);
}

void OpenGLRender::RenderSurfaces(CSurface3D *p_surfaces, unsigned p_surfaceQty, CModelVertex3D *p_vertices, CColorRGB *p_modColor, BYTE p_modAlpha, CAPITexture **p_lightmaps, CSurfaceDecalList *p_decalLists, bool p_ignoreMirrors, CLightSourceRefList *p_lightSourceList, CAPITexture *p_useThisLightmap, COrient3D *p_orient)
{
	unsigned s, v;
	CSurface3D *lpSurface = p_surfaces;
	int *lpVertexRefs;
	CColorRGB color;
	BYTE alpha;
	CVector3D *lpVertex;
	CModelVertex3D *lpModelVertex;
	CTexCoord2f *lpTexCoords;
	CTexCoord2f *lpTexCoordsLM;
	CAPITexture	*texture0;
	CLinkedList<CSurfaceDecal> *decalList;

	GLuint tex0Target, texLMTarget;
	CAPITexture *LMTexture;

	bool texEnabled, texEnabledLM;

	bool restoreClipPlane;
	CVector3D apiViewpoint = GetViewpoint().p;
	
	for (s=0; s<p_surfaceQty; s++)
	{
		// It's done this way instead of lpSurface++ at the end to allow us
		//   to bump s inside the loop up if need be
		// We may inner loop through some surfaces if we prepare a large
		//   off-screen texture for multiple at once
		lpSurface = &p_surfaces[s];

		// If surface isn't visible (flag or orientation), might as well skip 
		//   down now and save all this trouble
		// (This wouldn't be faster with a high polygon count mesh, but that
		//    would be another routine anyway)
		// Actually, we can't do that because we might be rendering a re-scaled
		//    object (flat, for example), so checking for real-world
		//    visibility wouldn't be accurate.  It isn't often we scale,
		//    but it happens.
		// ??

		// In case we prepare the offscreen buffer for multiple surfaces...
		unsigned surfaceStart = s;
		unsigned surfaceStop = s;
		int currentDynLMID = -1;
		bool continuePreparingDynLM = true;
		unsigned currentSurface = s;

		// Check for lightmap and prepare it with lightsources if necessary
		// And make sure it's loaded and ready to use
		// ?? Here, we'll prepare the dynamic lightmap too
		texEnabledLM = false;
		LMTexture = NULL;
		if (lpSurface->IsVisible())
		if (p_useThisLightmap)
		{
			texEnabledLM = true;
			LMTexture = p_useThisLightmap;
		}
		else
			if (p_lightmaps)
			if (lpSurface->GetLightmapID() != -1)
			{
				while (continuePreparingDynLM)
				{

				if (lpSurface->GetLightmapID() != -1)
				{
					p_lightmaps[lpSurface->GetLightmapID()]->UpdateLightmapGL();
					if ((p_lightmaps[lpSurface->GetLightmapID()]->IsValid()) && (!offScreenMode))
					{
						texEnabledLM = true;
						if (!LMTexture)
							LMTexture = p_lightmaps[lpSurface->GetLightmapID()];
					}
				}

				// At this point, we may prepare a dynamic lightmap and use
				// that instead, by resetting the LMTexture pointer...
				if ((lpSurface->IsVisible()) && (lpSurface->GetLightmapID() != -1))
				if (p_lightSourceList)
				if (!p_lightSourceList->IsEmpty())
				if (APITextures.GetDynamicLightmap()->IsValid())
				if (1!=2)	// For debugging
				if (((apiViewpoint - p_vertices[lpSurface->GetVertexRef(0)].vertex) * lpSurface->GetNormal()) > 0)
				{
					// Check for light sources
                    CLightSourceRefListNode *lnode;
                    lnode = p_lightSourceList->GetFirst();
                    if (lnode)
					{
                        float x, y, r, alpha, distance;
                        int index1;
						float alpha1;
						CVector3D tmpPos;
						CColorRGB white(255,255,255), tmpColor;
                        while (lnode->next != NULL)
						{
                            // Add it
                            if ((lnode->ref->intensity > 0.0f) && (lnode->ref->visible) && (!(lnode->ref->staticLS) || (p_orient)))
							{
                                switch(lnode->ref->type)
								{
								case APILIGHTTYPE_SPHERICALLIGHTMAP:
                                    // See if the light is even visible
                                    // Has to be on, in front of surface, and not be
                                    //   farther away than its radius
                                    // If an orientation was sent, convert the light position
                                    //    to object space
                                    if (p_orient)
                                        tmpPos = p_orient->TransformVectorToObjectSpace(lnode->ref->position - p_orient->p);
                                    else
                                        tmpPos = lnode->ref->position;

                                    distance = (tmpPos - p_vertices[lpSurface->GetVertexRef(0)].vertex) * lpSurface->GetNormal();
                                    if ((distance > -(lnode->ref->radius/8.0f)) && (distance < lnode->ref->radius))
									{
										// Allow light to illuminate from behind just a little
										//   so it starts illuminating as it comes around
										//   a corner
										// Just an option, make sure >= k above equals / k below
										if (distance < 0.0f)
											distance = (lnode->ref->radius * -distance / (lnode->ref->radius/8.0f));

                                        // Set up offscreen rendering if it isn't already
                                        if (!offScreenMode)
										{
											// We're using it now
                                            LMTexture = APITextures.GetDynamicLightmap();
                                                            
                                            // Activate dynamic light map buffer and copy the static contents over
                                            glMatrixMode (GL_MODELVIEW);
                                            glPushMatrix();
                                            glMatrixMode (GL_PROJECTION);
                                            glPushMatrix();
                                            
                                            // Sadly, clip planes can't be saved, as far as I know
                                            // So we have to restore it another way
											// Our current values are already stored from SetClipPlane()
                                            restoreClipPlane = false;
                                            if (glIsEnabled(GL_CLIP_PLANE0))
                                                restoreClipPlane = true;
                                            
                                            //glPushAttrib amAllAttribBits
                                            glPushAttrib (GL_VIEWPORT_BIT);
                                            glPushAttrib (GL_STENCIL_BUFFER_BIT); // stencil
                                            glPushAttrib (GL_POLYGON_BIT);       // culling mode (doesn't save clip plane)
                                            glPushAttrib (GL_DEPTH_BUFFER_BIT);   // depth states
                                            // glPushAttrib amTextureBit       ' texture values (will be changed in RenderQuad)
                                            glDisable (GL_STENCIL_TEST);
                                            glDisable (GL_CLIP_PLANE0);
                                            glDisable (GL_CULL_FACE);   // Could be in a mirror, need to disable
                                            glDisable (GL_DEPTH_TEST);
                                            
											SwitchToDynamicBuffer();
											Metrics.bufferSwitches++;

											surfaceStart = s;
											currentDynLMID = lpSurface->GetDynLMID();
											continuePreparingDynLM = true;
											currentSurface = s;

                                            glViewport (lpSurface->GetDynLMX(), lpSurface->GetDynLMY(), p_lightmaps[lpSurface->GetLightmapID()]->GetWidth(), p_lightmaps[lpSurface->GetLightmapID()]->GetHeight());
                                            glMatrixMode (GL_PROJECTION);
                                            glLoadIdentity();
                                            gluPerspective (90.0f, 1, 0.1f, 1000.0f);
                                            glMatrixMode (GL_MODELVIEW);
                                            glLoadIdentity();
                                            glRotatef (180.0f, 0.0f, 1.0f, 0.0f);
                                            glTranslatef (0.0f, 0.0f, 1.0f);
                            
                                            // And render it
                                            // No need to scale it since the projection settings (90 degrees, aspect 1)
                                            //   guarantee that a square fills the entire viewport
                                            //Clear(0xFFFF000);	// testing, if it works we can get rid of this, it's slow
											SetBlendMode(APIBLENDMODE_DISABLED);
                                            RenderQuad(white, 255, p_lightmaps[lpSurface->GetLightmapID()]);
                                                                                        
                                            // Move to LM origin coordinate of surface
                                            // Clever :)
                                            glTranslatef (1.0f, -1.0f, 0.0f);
                                            
                                            // And square up the coordinates (or we could redo the projection.  Nah!)
											// This took a while to get to work right, don't ask
                                            glScalef (-2.0f * p_lightmaps[lpSurface->GetLightmapID()]->GetHeightf() / p_lightmaps[lpSurface->GetLightmapID()]->GetWidthf(), 2.0f, 1.0f);
                                                                            
                                            //IRender_SetBlendMode (APIBLENDMODE_ADDITIVEBLEND)
                                            SetBlendMode (APIBLENDMODE_TENDDESTINATIONTOFULL);
                                        }	// Set up offscreen
										else
										{
											// We are already in offscreen, are we working on a different surface?
											// We'll need to copy that static lightmap over
											if (currentSurface != s)
											{
												currentSurface = s;

												Metrics.bufferSwitchesPrevented ++;
												glViewport (lpSurface->GetDynLMX(), lpSurface->GetDynLMY(), p_lightmaps[lpSurface->GetLightmapID()]->GetWidth(), p_lightmaps[lpSurface->GetLightmapID()]->GetHeight());
												//glMatrixMode (GL_PROJECTION);
												//glLoadIdentity();
												//gluPerspective (90.0f, 1, 0.1f, 1000.0f);
												//glMatrixMode (GL_MODELVIEW);
												glLoadIdentity();
												glRotatef (180.0f, 0.0f, 1.0f, 0.0f);
												glTranslatef (0.0f, 0.0f, 1.0f);
                            
												// And render it
												// No need to scale it since the projection settings (90 degrees, aspect 1)
												//   guarantee that a square fills the entire viewport
												//Clear(0xFFFF000);	// testing, if it works we can get rid of this, it's slow
												SetBlendMode(APIBLENDMODE_DISABLED);
												RenderQuad(white, 255, p_lightmaps[lpSurface->GetLightmapID()]);
                                                                                        
												// Move to LM origin coordinate of surface
												// Clever :)
												glTranslatef (1.0f, -1.0f, 0.0f);
                                            
												// And square up the coordinates (or we could redo the projection.  Nah!)
												// This took a while to get to work right, don't ask
												glScalef (-2.0f * p_lightmaps[lpSurface->GetLightmapID()]->GetHeightf() / p_lightmaps[lpSurface->GetLightmapID()]->GetWidthf(), 2.0f, 1.0f);
                                                                            
												//IRender_SetBlendMode (APIBLENDMODE_ADDITIVEBLEND)
												SetBlendMode (APIBLENDMODE_TENDDESTINATIONTOFULL);
											}
										}
                                    
                                        // Determine correct radius and alpha based on distance from surface
                                        r = float(sqrt(pow(lnode->ref->radius,2) - pow(distance,2)) * p_lightmaps[lpSurface->GetLightmapID()]->GetTexelPerUnit() / p_lightmaps[lpSurface->GetLightmapID()]->GetHeightf());
                                        // Note: this alpha calculation assumes a linear reduction,
                                        //   which should be reflected in the style of the texture used as well
                                        //   Otherwise, a different formula is needed
                                        // Using 0.0-1.0 now since we are modulating the color with it
                                        // We'll be using vertex alpha to blend two splices
                                        alpha = float((lnode->ref->radius - distance) / lnode->ref->radius) * lnode->ref->intensity;
                                        // adjust white-based color
                                        tmpColor.SetColor(lnode->ref->color.Scm(alpha));
                                        
                                        lpSurface->CalculateTexLMCoords(tmpPos, x, y);
                                        glPushMatrix();
										glTranslatef (x * p_lightmaps[lpSurface->GetLightmapID()]->GetWidthf() / p_lightmaps[lpSurface->GetLightmapID()]->GetHeightf(), 
														y, 0);
                                        
                                        glScalef (r, r, 1.0f);
                                        
                                        // Get the two approximate LM textures and add them both
                                        index1 = int((distance / lnode->ref->radius) * 10.0f);
                                        alpha1 = ((distance / lnode->ref->radius) * 10.0f - float(index1)) * 255.0f;
                                        if (alpha1 > 255) alpha1 = 255;
                                        if (alpha1 < 0) alpha1 = 0;										
                                        if (index1 >= 9)
										{
                                        // ? This isn't right
                                            //Call IRender_RenderQuad(tmpColor, 255, APITextures.GetLMTexture(9))
											// This is
                                            RenderMultiQuad(tmpColor, BYTE(alpha1), APITextures.GetSphereLightmap(index1), NULL, APIMULTIQUADTEXTUREOP_BLENDALPHAMAPS);
                                            //Call IRender_RenderQuad(tmpColor, 255)
										}
                                        else
										{
                                            //alpha1 = (1# - (alpha / 255#) - CSng(index1) / 10#) * 10#
                                            //Call IRender_RenderQuad(lnode.color, alpha * CSng(alpha1), APITextures.GetLMTexture(index1 + 1))
                                            //Call IRender_RenderQuad(lnode.color, alpha * CSng(1# - alpha1), APITextures.GetLMTexture(index1))
                                            RenderMultiQuad(tmpColor, BYTE(alpha1), APITextures.GetSphereLightmap(index1), APITextures.GetSphereLightmap(index1 + 1), APIMULTIQUADTEXTUREOP_BLENDALPHAMAPS);
										}
                                        glPopMatrix();
                                    }	// if light source is in position to affect surface
									break;
                                } // switch
                            }
                            
                            lnode = (CLightSourceRefListNode *)(lnode->next);
                        }  // while there are lightsources
                    }	// If there are lightsources


				}	// if ok to attempt applying dynamic light sources

				// If we enter this, it means we passed a surface that
				//   will use the common dynamic lightmap but had no
				//   reason to copy over its static lightmap or apply
				//   any lights.  We fix that here, we need its static
				//   data.
				if ((currentSurface != s) && (offScreenMode) && (lpSurface->IsVisible()))
				if (((apiViewpoint - p_vertices[lpSurface->GetVertexRef(0)].vertex) * lpSurface->GetNormal()) > 0)
				{
					currentSurface = s;

					CColorRGB white(255,255,255);

					Metrics.bufferSwitchesPrevented++;
					glViewport (lpSurface->GetDynLMX(), lpSurface->GetDynLMY(), p_lightmaps[lpSurface->GetLightmapID()]->GetWidth(), p_lightmaps[lpSurface->GetLightmapID()]->GetHeight());
					//glMatrixMode (GL_PROJECTION);
					//glLoadIdentity();
					//gluPerspective (90.0f, 1, 0.1f, 1000.0f);
					//glMatrixMode (GL_MODELVIEW);
					glLoadIdentity();
					glRotatef (180.0f, 0.0f, 1.0f, 0.0f);
					glTranslatef (0.0f, 0.0f, 1.0f);

					// And render it
					// No need to scale it since the projection settings (90 degrees, aspect 1)
					//   guarantee that a square fills the entire viewport
					//Clear(0xFFFF000);	// testing, if it works we can get rid of this, it's slow
					SetBlendMode(APIBLENDMODE_DISABLED);
					RenderQuad(white, 255, p_lightmaps[lpSurface->GetLightmapID()]);
                                                            
					// Move to LM origin coordinate of surface
					// Clever :)
					glTranslatef (1.0f, -1.0f, 0.0f);
                
					// And square up the coordinates (or we could redo the projection.  Nah!)
					// This took a while to get to work right, don't ask
					glScalef (-2.0f * p_lightmaps[lpSurface->GetLightmapID()]->GetHeightf() / p_lightmaps[lpSurface->GetLightmapID()]->GetWidthf(), 2.0f, 1.0f);
                                                
					//IRender_SetBlendMode (APIBLENDMODE_ADDITIVEBLEND)
					SetBlendMode (APIBLENDMODE_TENDDESTINATIONTOFULL);
				}
					lpSurface++;
					s++;
					if ((offScreenMode) && (s < p_surfaceQty))
					{

						if (((lpSurface->IsVisible()) && (lpSurface->GetDynLMID() != currentDynLMID)) || (currentDynLMID == -1))
						{
							continuePreparingDynLM = false;
							s--;	// s will be incremented at the far end of the for loop
						}
					}
					else
					{
						continuePreparingDynLM = false;
						s--;
					}

					//if (continuePreparingDynLM)
					//{
					//	continuePreparingDynLM = false;
					//	s--;
					//}
				}// while continue preparing Dyn LM for multiple surfaces

				if (offScreenMode)
				{
					// We prepared a lightmap, texture is ready, so switch back
					SetBlendMode(APIBLENDMODE_ALPHABLEND);

					SwitchToMainBuffer();

                    //Call glPopAttrib();    // textures
                    glPopAttrib();   // depth buffer
                    glPopAttrib();   // culling mide, and clip plane (doesn't always work for clip plane, so see below)
                    glPopAttrib();   // stencil
                    glPopAttrib();   // Viewport
                    glMatrixMode(GL_PROJECTION);
                    glPopMatrix();
                    glMatrixMode(GL_MODELVIEW);
                    glPopMatrix();

                    // If clip plane is enabled, restore it to be sure,
                    //   instead of depending on the attribute stack
                    if (restoreClipPlane)
					{
						if (p_orient)
						{
							glPushMatrix();
							glLoadMatrixf(currentClipPlaneMatrix);
						}
                        glEnable (GL_CLIP_PLANE0);
                        glClipPlane (GL_CLIP_PLANE0, currentClipPlane);
						if (p_orient)
							glPopMatrix();
						//TRACE(">> Restoring clip plane (%g, %g, %g, %g)\n", currentClipPlane[0],currentClipPlane[1],currentClipPlane[2],currentClipPlane[3]);
					}

					surfaceStop = s;
				}	// if offscreenMode
			}  // lightmap ID != -1

		unsigned s1;
//		if (surfaceStart!=surfaceStop)
//			Metrics.bufferSwitchesPrevented += (surfaceStop-surfaceStart);

		// Now draw the surfaces for which we prepared one large dynamic lightmap
		//   (or draw a single one for which dynLMID = -1, or it didn't
		//    need a dynamic one)
		for (s1=surfaceStart; s1<=surfaceStop; s1++)
		{
		lpSurface = &p_surfaces[s1];

		// Get vertex refs
		lpVertexRefs = lpSurface->GetVertexRefs();

		// If this is an alpha surface and it has backside decals, draw them
		// now, back to front
		// Don't draw decals if we aren't ignoring mirrors
		if (lpSurface->GetOpposingSurface())
		if (lpSurface->IsVisible())
		if (p_ignoreMirrors)
		if ((p_decalLists) && (lpSurface->GetDecalListID() != -1))
		if (!p_decalLists[lpSurface->GetDecalListID()].IsEmpty())
		if (!lpSurface->IsMirror())
		if (((apiViewpoint - p_vertices[lpVertexRefs[0]].vertex) * lpSurface->GetNormal()) > 0.0f)
		{
			decalList = &p_decalLists[lpSurface->GetDecalListID()];
			CLinkedListNode<CSurfaceDecal> *node;
			node = decalList->GetLast();
			if (node)
			{
				// Set up GL states to prevent Z fighting
				// and culling (they will be backwards)
				glDepthMask(false);
				SetCullCCW(!cullCCW);

				while (node->prev != NULL)
				{
					if (node->backside)
					{
						// Render it!
						// NOTE: We still need the lightmaps for texture scaling with the dynamic buffer
						RenderSurfaces(&(node->surface), 1, node->vertices, p_modColor, p_modAlpha, p_lightmaps, NULL, true, NULL, LMTexture, NULL);
					}

					node = node->prev;
				}	// while

				SetCullCCW(!cullCCW);
				glDepthMask(true);
			}
		}

		// If surface is a mirror and we are not drawing mirrors, just do decals
		// Also, ignore if surface is invisible, of course
		if ((!p_ignoreMirrors) || (!lpSurface->IsMirror()) && (lpSurface->IsVisible()))
		{

		// Set up textures to use and combine parameters
		texEnabled = false;
		texture0 = lpSurface->GetTexture();
		if (texture0 != NULL)
		{
			if (texture0->IsValid())
			{
				if (!texture0->IsLoaded())
				{
					texture0->LoadTextureGL();
					if (texture0->IsValid())
					{
						texEnabled = true;
					}
				}
				else
					texEnabled = true;
			}
		}

		// Set up texture pipeline
		if ((texEnabled) || (texEnabledLM))
		{
			if (texEnabled)
				if (!texEnabledLM)
				{
					// Set up slot 0 with texture 1, modulate with vertices
					glActiveTextureARB(GL_TEXTURE0_ARB);
					glEnable(GL_TEXTURE_2D);
					glBindTexture(GL_TEXTURE_2D, texture0->GetGLid());

					tex0Target = GL_TEXTURE0_ARB;

					if (texture0->GetForceClamp())
					{
						glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
						glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
					}
					else
					{
						glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
						glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
					}
					// GL_NEAREST makes them look pixelated (good for icons, bad for surfaces)
					glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
					if (texture0->GetMipMap())
						glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
					else
						glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
					// Multiply by color (GL_DECAL ignores polygon colors, and would be unaffected by lighting)
					glTexEnvf (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

					glActiveTextureARB(GL_TEXTURE2_ARB);
					glDisable(GL_TEXTURE_2D);
					glActiveTextureARB(GL_TEXTURE1_ARB);
					glDisable(GL_TEXTURE_2D);
				}
				else
				{
					// Set up slot 0 with texture1 and slot 1 with lightmap,
					//   and modulate their colors, and use texture 1's alpha,
					//   and modulate the result with the vertices
					glActiveTextureARB(GL_TEXTURE0_ARB);
					glEnable(GL_TEXTURE_2D);
					glBindTexture(GL_TEXTURE_2D, texture0->GetGLid());

					tex0Target = GL_TEXTURE0_ARB;

					glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
					glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
					// GL_NEAREST makes them look pixelated (good for icons, bad for surfaces)
					glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
					if (texture0->GetMipMap())
						glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
					else
						glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
					// Multiply by color (GL_DECAL ignores polygon colors, and would be unaffected by lighting)
					glTexEnvf (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE);
					glTexEnvi (GL_TEXTURE_ENV, GL_COMBINE_RGB_ARB, GL_MODULATE);
					glTexEnvi (GL_TEXTURE_ENV, GL_SOURCE0_RGB_ARB, GL_PRIMARY_COLOR);
					glTexEnvi (GL_TEXTURE_ENV, GL_OPERAND0_RGB_ARB, GL_SRC_COLOR);
					glTexEnvi (GL_TEXTURE_ENV, GL_SOURCE1_RGB_ARB, GL_TEXTURE);
					glTexEnvi (GL_TEXTURE_ENV, GL_OPERAND1_RGB_ARB, GL_SRC_COLOR);
					if (lpSurface->GetTexture()->GetFormat() == APITEXTUREFORMAT_ALPHA8)
					{
						glTexEnvi (GL_TEXTURE_ENV, GL_COMBINE_RGB_ARB, GL_REPLACE);
						glTexEnvi (GL_TEXTURE_ENV, GL_SOURCE0_RGB_ARB, GL_PRIMARY_COLOR);
						glTexEnvi (GL_TEXTURE_ENV, GL_OPERAND0_RGB_ARB, GL_SRC_COLOR);
					}
					glTexEnvi (GL_TEXTURE_ENV, GL_COMBINE_ALPHA_ARB, GL_MODULATE);
					glTexEnvi (GL_TEXTURE_ENV, GL_SOURCE0_ALPHA_ARB, GL_TEXTURE);
					glTexEnvi (GL_TEXTURE_ENV, GL_OPERAND0_ALPHA_ARB, GL_SRC_ALPHA);
					glTexEnvi (GL_TEXTURE_ENV, GL_SOURCE1_ALPHA_ARB, GL_PRIMARY_COLOR);
					glTexEnvi (GL_TEXTURE_ENV, GL_OPERAND1_ALPHA_ARB, GL_SRC_ALPHA);


					// Slot 1 will be the lightmap
					glActiveTextureARB(GL_TEXTURE1_ARB);
					glEnable(GL_TEXTURE_2D);
					glBindTexture(GL_TEXTURE_2D, LMTexture->GetGLid());
                    if (LMTexture->IsRenderBuffer())
					{
                        // Apply texture matrix transform to make sure everything
                        //      lines up using surface's LM coords
                        glMatrixMode (GL_TEXTURE);
                        glLoadIdentity();
						glTranslatef(lpSurface->GetDynLMXf()/float(DYNAMICLIGHTMAPDIMENSION), lpSurface->GetDynLMYf()/float(DYNAMICLIGHTMAPDIMENSION),0);
                        glScalef ( p_lightmaps[lpSurface->GetLightmapID()]->GetWidthf() / LMTexture->GetWidthf(),
									p_lightmaps[lpSurface->GetLightmapID()]->GetHeightf() / LMTexture->GetHeightf(), 1.0f);
                        glMatrixMode (GL_MODELVIEW);
					}

					texLMTarget = GL_TEXTURE1_ARB;

					glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
					glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
					// GL_NEAREST makes them look pixelated (good for icons, bad for surfaces)
					glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
					if (LMTexture->GetMipMap())
						glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
					else
						glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
					// Multiply by color (GL_DECAL ignores polygon colors, and would be unaffected by lighting)
					// This stage has no alpha in it
					glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE);
					glTexEnvi (GL_TEXTURE_ENV, GL_COMBINE_RGB_ARB, GL_MODULATE);
					glTexEnvi (GL_TEXTURE_ENV, GL_SOURCE0_RGB_ARB, GL_PREVIOUS);
					glTexEnvi (GL_TEXTURE_ENV, GL_OPERAND0_RGB_ARB, GL_SRC_COLOR);
					glTexEnvi (GL_TEXTURE_ENV, GL_SOURCE1_RGB_ARB, GL_TEXTURE);
					glTexEnvi (GL_TEXTURE_ENV, GL_OPERAND1_RGB_ARB, GL_SRC_COLOR);
					glTexEnvi (GL_TEXTURE_ENV, GL_COMBINE_ALPHA_ARB, GL_REPLACE);
					glTexEnvi (GL_TEXTURE_ENV, GL_SOURCE0_ALPHA_ARB, GL_PREVIOUS);
					glTexEnvi (GL_TEXTURE_ENV, GL_OPERAND0_ALPHA_ARB, GL_SRC_ALPHA);

					glActiveTextureARB(GL_TEXTURE2_ARB);
					glDisable(GL_TEXTURE_2D);
				}
			else
			{
				if (texEnabledLM)
				{
					// Set up slot 0 with lightmap
					// Modulate with vertices
					glActiveTextureARB(GL_TEXTURE0_ARB);
					glEnable(GL_TEXTURE_2D);
					glBindTexture(GL_TEXTURE_2D, LMTexture->GetGLid());
                    if (LMTexture->IsRenderBuffer())
					{
                        // Apply texture matrix transform to make sure everything
                        //      lines up using surface's LM coords
                        glMatrixMode (GL_TEXTURE);
                        glLoadIdentity();
						glTranslatef(lpSurface->GetDynLMXf()/float(DYNAMICLIGHTMAPDIMENSION), lpSurface->GetDynLMYf()/float(DYNAMICLIGHTMAPDIMENSION),0);
                        glScalef ( p_lightmaps[lpSurface->GetLightmapID()]->GetWidthf() / LMTexture->GetWidthf(),
									p_lightmaps[lpSurface->GetLightmapID()]->GetHeightf() / LMTexture->GetHeightf(), 1.0f);
                        glMatrixMode (GL_MODELVIEW);
					}

					texLMTarget = GL_TEXTURE0_ARB;

					glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
					glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
					// GL_NEAREST makes them look pixelated (good for icons, bad for surfaces)
					glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
					if (LMTexture->GetMipMap())
						glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
					else
						glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
					// Multiply by color (GL_DECAL ignores polygon colors, and would be unaffected by lighting)
					glTexEnvf (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

					glActiveTextureARB(GL_TEXTURE2_ARB);
					glDisable(GL_TEXTURE_2D);
					glActiveTextureARB(GL_TEXTURE1_ARB);
					glDisable(GL_TEXTURE_2D);
				}
				else
				{
					// Disable all texturing (redundent)
					glActiveTextureARB(GL_TEXTURE2_ARB);
					glDisable(GL_TEXTURE_2D);
					glActiveTextureARB(GL_TEXTURE1_ARB);
					glDisable(GL_TEXTURE_2D);
					glActiveTextureARB(GL_TEXTURE0_ARB);
					glDisable(GL_TEXTURE_2D);
				}
			}
		}
		else
		{
			// Disable all texturing
			glActiveTextureARB(GL_TEXTURE2_ARB);
			glDisable(GL_TEXTURE_2D);
			glActiveTextureARB(GL_TEXTURE1_ARB);
			glDisable(GL_TEXTURE_2D);
			glActiveTextureARB(GL_TEXTURE0_ARB);
			glDisable(GL_TEXTURE_2D);
		}

		// Set rendering mode
		switch (lpSurface->GetVertexRefQty())
		{
		case 1:
			glBegin (GL_POINTS);
			break;
		case 2:
			glBegin (GL_LINES);
			break;
		case 3:
			glBegin (GL_TRIANGLES);
			Metrics.triangles++;
			break;
		default:
			glBegin (GL_POLYGON);
			Metrics.triangles += (lpSurface->GetVertexRefQty() - 2);
			break;
		}

		// Set up full surface color now if necessary, modulate with parameter
		bool useSurfaceColor = lpSurface->UseSurfaceColor();
		if (useSurfaceColor)
			if (p_modColor)
				if (p_modColor->IsWhite())
					color = lpSurface->GetColor();
				else
					color = lpSurface->GetColor().Modulate(*p_modColor);
			else
				color = lpSurface->GetColor();

		// Prepare alpha for surface, modulate with parameter
		if (p_modAlpha != 255)
			alpha = lpSurface->GetAlpha() * p_modAlpha / 255;
		else
			alpha = lpSurface->GetAlpha();
		
		// Get ready to loop through vertices for the surface
		if (texEnabled)
			lpTexCoords = lpSurface->GetTexCoords();
		if (texEnabledLM)
			lpTexCoordsLM = lpSurface->GetTexCoordsLM();
		unsigned vertexQty = lpSurface->GetVertexRefQty();
		for (v=0; v<vertexQty; v++)
		{
			// ?? I'll have to try this with various attempts at
			//   de-referencing to see what's faster
			lpModelVertex = &p_vertices[lpVertexRefs[v]];
			if (!useSurfaceColor)
			{
				// Modulate with parameter if necessary
				if (p_modColor)
					if (p_modColor->IsWhite())
						color = lpModelVertex->color;
					else
						color = lpModelVertex->color.Modulate(*p_modColor);
				else
					color = lpModelVertex->color;
			}

			lpVertex = &(lpModelVertex->vertex);

			// Submit the data:
			// Color
			if (alpha == 255)
				glColor3ubv(&color.red);
			else
				glColor4ub(color.red, color.green, color.blue, alpha);

			// Tex coords
			if (texEnabled)
			{
				glMultiTexCoord2f(tex0Target, lpTexCoords->GetS(), lpTexCoords->GetT());
				lpTexCoords++;
			}
			if (texEnabledLM)
			{
				glMultiTexCoord2f(texLMTarget, lpTexCoordsLM->GetS(), lpTexCoordsLM->GetT());
				lpTexCoordsLM++;
			}

			// Vertex coordinates
			glVertex3fv((float *)(&lpVertex->x));
		}

		glEnd();

		// Restore texture matrix if it was modified
		if (LMTexture)
		if (LMTexture->IsRenderBuffer())
		if (texEnabled)
		{
			glActiveTextureARB(GL_TEXTURE1_ARB);
			glMatrixMode(GL_TEXTURE);
			glLoadIdentity();
			glMatrixMode(GL_MODELVIEW);
		}
		else
		{
			glActiveTextureARB(GL_TEXTURE0_ARB);
			glMatrixMode(GL_TEXTURE);
			glLoadIdentity();
			glMatrixMode(GL_MODELVIEW);
		}

		}	// Ignoring mirrors according to parameter and surface, and invisible surfaces

		// If this surface has front-side decals, draw them now
		// Don't draw decals if we aren't ignoring mirrors
		if (lpSurface->IsVisible())
		if ((p_ignoreMirrors) || (!lpSurface->IsMirror()))
		if ((p_decalLists) && (lpSurface->GetDecalListID() != -1))
		if (!p_decalLists[lpSurface->GetDecalListID()].IsEmpty())
		if (((apiViewpoint - p_vertices[lpVertexRefs[0]].vertex) * lpSurface->GetNormal()) > 0.0f)
		{
			decalList = &p_decalLists[lpSurface->GetDecalListID()];
			CLinkedListNode<CSurfaceDecal> *node;
			node = decalList->GetFirst();
			if (node)
			{
				// Set up GL states to prevent Z fighting
				// and culling (they will be backwards)
				SetZBias(true);
				glDepthMask(false);

				while (node->next != NULL)
				{
					if (!(node->backside))
					{
						// Render it!
						// NOTE: We still need the lightmaps for texture scaling with the dynamic buffer
						RenderSurfaces(&(node->surface), 1, node->vertices, p_modColor, p_modAlpha, p_lightmaps, NULL, true, NULL, LMTexture, NULL);
					}

					node = node->next;
				}
				glDepthMask(true);
				SetZBias(false);
			}
		} // Draw frontside decals

		} // for s1

	}	// for s

	if (texEnabled || texEnabledLM)
	{
		glActiveTextureARB(GL_TEXTURE1_ARB);
		glDisable(GL_TEXTURE_2D);
		glActiveTextureARB(GL_TEXTURE0_ARB);
		glDisable(GL_TEXTURE_2D);
	}

}

void OpenGLRender::RenderStarField(CStarField &p_starField, COrient3D *p_currentOrient, COrient3D *p_oldOrient, DWORD p_backColor)
{
	if (!p_starField.stars) return;

	// The right balance of point size, line width and alpha adjustment for drawn lines
	//   has to be right so that the sky doesn't brighten at the horizon when
	//   spinning, and stars near the points at which they would be rendered as points don't
	//   darken.  What is here is close.

	glActiveTextureARB(GL_TEXTURE0_ARB);
	glDisable(GL_TEXTURE_2D);

	glPushMatrix();

	bool renderPoints = true;
	if ((p_currentOrient) && (p_oldOrient))
	{
		// For this to be accurate, it's important for the orients to
		//   be normalized
		if ((p_currentOrient->f * p_oldOrient->f) < 0.99999f)
			renderPoints = false;
		if ((p_currentOrient->l * p_oldOrient->l) < 0.99999f)
			renderPoints = false;
		if ((p_currentOrient->u * p_oldOrient->u) < 0.99999f)
			renderPoints = false;
	}

	if (renderPoints)
	{
		glEnable(GL_POINT_SMOOTH);

		// Can't put this in the glBegin/glEnd block
		glPointSize(1.5f);

		CVector3D vp = GetViewpoint().p;
		glTranslatef(vp.x, vp.y, vp.z);

		// Just render the stars without any stretching
		unsigned int i;
		strStar *lpStar = &p_starField.stars[0];

		glBegin(GL_POINTS);
		for (i=0; i<p_starField.starQty; i++)
		{
			glColor4ub(lpStar->color.red, lpStar->color.green, lpStar->color.blue, lpStar->intensity);
			glVertex3fv(&(lpStar->pos.x));

			lpStar++;
		}
		glEnd();

		glDisable(GL_POINT_SMOOTH);
	}
	else
	{
		// Render lines to express the change in orientation
		// (Best to do this only in games that have smooth orientation changes,
		//    like space simulations)
		// This slows down the frame rate a litte, but it doesn't kill the app
		//   Optionally, we could store the old position value with the assumption that
		//   they willb e valid on the next call
		glLoadIdentity();
		glRotatef(180.0f, 0.0f, 1.0f, 0.0f);

		unsigned int i;
		strStar *lpStar = &p_starField.stars[0];

		glEnable(GL_POINT_SMOOTH);
		glEnable(GL_LINE_SMOOTH);

		// See what the largest color component is for the background color.
		// If it's > 128, use line width 1.0
		// If it's between 0 and 128, adjust the line width between 1.0 and 2.0,
		//    2.0 at 0.
		BYTE value;
		float fValue;
		value = BYTE(p_backColor & 0xFF);
		if (value < ((p_backColor >> 8) & 0xFF))
			value = BYTE((p_backColor >> 8) & 0xFF);
		if (value < ((p_backColor >> 16) & 0xFF))
			value = BYTE((p_backColor >> 16) & 0xFF);
		if (value > 128)
		{
			glLineWidth(1.0f);
			fValue = 0.0f;
		}
		else
		{
			glLineWidth(2.0f - float(value)/128.0f);
			fValue = float(sqrt(1.0f - float(value)/128.0f));
		}

		// Can't put this in the glBegin/glEnd block
//		glLineWidth(1.0f);	// 1.0f for light background, 2.0 for dark
		glPointSize(1.5f);

		CVector3D p1, p2;
		CVector3D pm;
		float mgn;
		float factor;

		GLuint renderMode = GL_LINES;
		glBegin(GL_LINES);
		for (i=0; i<p_starField.starQty; i++)
		{
			p1 = p_currentOrient->TransformVectorToObjectSpace(lpStar->pos);
			if (p1.z > 0.7f)
			{
				p2 = p_oldOrient->TransformVectorToObjectSpace(lpStar->pos);

				// If point didn't move very much, render it as a point
				mgn = (p1-p2).MgnTudSqr();
				if (mgn < 0.000055f)
				{
					if (renderMode == GL_LINES)
					{
						glEnd();
						glBegin(GL_POINTS);
						renderMode = GL_POINTS;
					}
					glColor4ub(lpStar->color.red, lpStar->color.green, lpStar->color.blue, lpStar->intensity);
					glVertex3fv(&(p1.x));
				}
				else
				// Render it as a line
				{
					// We want to calm down the brightness and size of the line a little sooner
					//   when it's darker, but not shorten it so much at the extreme
					// When the sky is lighter, we can adjust it later but will
					//   need to shorten it more

					pm = (p1+p2) * 0.5f;
					// If line is really long, shorten it
					//    (helps with low frame rates, keeps lines from
					//     saturating the view)
					float val2 = 0.0001f + 0.0003f * (1.0f-fValue);
					if (mgn > val2)
					{
						p1 = pm + ((p1-pm) * (0.5f + 0.5f * fValue * float(sqrt(val2/mgn))));
						p2 = pm + ((p2-pm) * (0.5f + 0.5f * fValue * float(sqrt(val2/mgn))));
						factor = 0.5f + 0.2f * fValue * val2/mgn;
					}
					else
						factor = 0.7f;

					if (renderMode == GL_POINTS)
					{
						glEnd();
						glBegin(GL_LINES);
						renderMode = GL_LINES;
					}

					// The 0.7 should be higher when it's pitch black, lower when it's bright out
					// Because the lines are much more pronounced against a light background
					// So this routine shoudl probably receive the color the stars are against
					// 0.7f in pitch black, 0.4f at dusk
					// ??
					glColor4ub(lpStar->color.red, lpStar->color.green, lpStar->color.blue, BYTE(lpStar->intensity*factor));
					glVertex3fv(&(pm.x));
					glColor4ub(lpStar->color.red, lpStar->color.green, lpStar->color.blue, 0);
					glVertex3fv(&(p1.x));
					glColor4ub(lpStar->color.red, lpStar->color.green, lpStar->color.blue, BYTE(lpStar->intensity*factor));
					glVertex3fv(&(pm.x));
					glColor4ub(lpStar->color.red, lpStar->color.green, lpStar->color.blue, 0);
					glVertex3fv(&(p2.x));
				}
			}

			lpStar++;
		}
		glEnd();

		glDisable(GL_POINT_SMOOTH);
		glDisable(GL_LINE_SMOOTH);
	}

	glPopMatrix();
}

// p_renderID should be a single full scene render, mirrors and all
// It keeps the lit particle colors from being calcualted multiple times,
//   which can really slow things down
void OpenGLRender::RenderParticles(CParticle *p_particles, unsigned int p_particleQty, unsigned int p_renderID, CColorRGB *p_ambientColor, CLightSourceRefList *p_lightSources)
{
	if (!p_particles) return;

	CParticle *lpParticle;
	unsigned int i;
	CAPITexture *lastTexture = NULL;

	CColorRGB color;
	BYTE alpha;

	COrient3D vp = GetViewpoint();
	CVector3D offset;
	bool texEnabled;

	// Set up for z sprite-style rendering
	glPushMatrix();
	glLoadIdentity();
	glRotatef(180,0,1,0);

	glActiveTextureARB(GL_TEXTURE1_ARB);
	glDisable(GL_TEXTURE_2D);
	glActiveTextureARB(GL_TEXTURE0_ARB);
	glDisable(GL_TEXTURE_2D);
	texEnabled = false;

	lpParticle = &(p_particles[0]);
	for (i=0; i<p_particleQty; i++)
	{
		// Minimize texture switching (can raise 420 fps up to 800 fps)
		if (lpParticle->texture != lastTexture)
		{
			lastTexture = lpParticle->texture;

			if (lpParticle->texture == NULL)
			{
				if (texEnabled)
				{
					texEnabled = false;
					glDisable(GL_TEXTURE_2D);
				}
			}	// texture null
			else
			{
				if (lpParticle->texture->IsValid())
				{
					if (!lpParticle->texture->IsLoaded())
						lpParticle->texture->LoadTextureGL();

					if (lpParticle->texture->IsValid())
					{

						if (!texEnabled)
						{
							texEnabled = true;

							glEnable(GL_TEXTURE_2D);
						}

						glBindTexture(GL_TEXTURE_2D, lpParticle->texture->GetGLid());

						// Set up parameters
						glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
						glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
						glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
						if (lpParticle->texture->GetMipMap())
							glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
						else
							glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
						glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

					}	// texture valid
					else
					{
						if (texEnabled)
						{
							texEnabled = false;

							glDisable(GL_TEXTURE_2D);
						}
					}	// texture not valid after load
				}	// texture initially valid
				else
				{
					if (texEnabled)
					{
						texEnabled = false;

						glDisable(GL_TEXTURE_2D);
					}
				} // texture not valid on entry
				
			} // texture not null
		} // Set up texturing - not same texture

		// Render only if in front
		offset = lpParticle->position - vp.p;
		if (offset * vp.f > 0)
		{
			glPushMatrix();
			glTranslatef(offset * vp.l, offset * vp.u, offset * vp.f);
			glScalef(lpParticle->size * lpParticle->radius, lpParticle->size * lpParticle->radius, 1.0f);

			// Determine color (lighting)
			if (lpParticle->emitting)
				color = lpParticle->color;
			else
			{
				if (lpParticle->renderID != p_renderID)
				{
					lpParticle->renderID = p_renderID;

					if ((p_ambientColor) && (p_lightSources))
					{
						color = p_lightSources->CalculateRenderColor(lpParticle->position, *p_ambientColor);
						color = color.Modulate(lpParticle->color);
					}
					else
						if (p_ambientColor)
							color = lpParticle->color.Modulate(*p_ambientColor);

					lpParticle->renderColor = color;
				}
				else
					color = lpParticle->renderColor;
			}

			alpha = BYTE(float(lpParticle->alpha) * lpParticle->alphaAmount);

			// Render it
			glBegin(GL_QUADS);
			if (texEnabled)
			{
				// ?? with texture
				glColor4ub(color.red, color.green, color.blue, alpha);
				glMultiTexCoord2f(GL_TEXTURE0_ARB, 0.0f, 1.0f);
				glVertex3i(1,1,0);
				glMultiTexCoord2f(GL_TEXTURE0_ARB, 1.0f, 1.0f);
				glVertex3i(-1,1,0);
				glMultiTexCoord2f(GL_TEXTURE0_ARB, 1.0f, 0.0f);
				glVertex3i(-1,-1,0);
				glMultiTexCoord2f(GL_TEXTURE0_ARB, 0.0f, 0.0f);
				glVertex3i(1,-1,0);
			}
			else
			{
				// ?? without
				glColor4ub(color.red, color.green, color.blue, alpha);
				glVertex3i(1,1,0);
				glVertex3i(-1,1,0);
				glVertex3i(-1,-1,0);
				glVertex3i(1,-1,0);
			}
			glEnd();
			Metrics.triangles += 2;
			Metrics.particles ++;

			glPopMatrix();
		}

		lpParticle++;
	}

	glPopMatrix();
	if (texEnabled)
	{
		glDisable(GL_TEXTURE_2D);
	}
}

void OpenGLRender::RenderWireframeCube(float p_minX, float p_minY, float p_minZ, float p_maxX, float p_maxY, float p_maxZ)
{
	glBegin(GL_LINE_STRIP);
		glColor3ub(0,255,0);	// green
		glVertex3f(p_minX, p_minY, p_minZ);
		glVertex3f(p_maxX, p_minY, p_minZ);
		glVertex3f(p_maxX, p_minY, p_maxZ);
		glVertex3f(p_minX, p_minY, p_maxZ);
		glVertex3f(p_minX, p_minY, p_minZ);
	glEnd();

	glBegin(GL_LINES);
		glColor3ub(0,255,0);	// green
		glVertex3f(p_minX, p_minY, p_minZ);
		glVertex3f(p_minX, p_maxY, p_minZ);

		glVertex3f(p_maxX, p_minY, p_minZ);
		glVertex3f(p_maxX, p_maxY, p_minZ);

		glVertex3f(p_minX, p_minY, p_maxZ);
		glVertex3f(p_minX, p_maxY, p_maxZ);

		glVertex3f(p_maxX, p_minY, p_maxZ);
		glVertex3f(p_maxX, p_maxY, p_maxZ);
	glEnd();

	glBegin(GL_LINE_STRIP);
		glColor3ub(0,255,0);	// green
		glVertex3f(p_minX, p_maxY, p_minZ);
		glVertex3f(p_maxX, p_maxY, p_minZ);
		glVertex3f(p_maxX, p_maxY, p_maxZ);
		glVertex3f(p_minX, p_maxY, p_maxZ);
		glVertex3f(p_minX, p_maxY, p_minZ);
	glEnd();
}

///////////////////////////////////////
// Ortho routines (viewpoint set up for x,y coordinates, like a Windows object

// NOTE: This routine doesn't require rotating on 180 degrees on 0,1,0
void OpenGLRender::SetViewportOrtho(void)
{
    glViewport(0, 0, m_windowWidth, m_windowHeight);
    m_scaleY = 1.0f;
    glMatrixMode (GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(90.0f, float(m_windowWidth) / float(m_windowHeight), 0.1f, 1000.0f);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glScalef (2.0f / float(m_windowHeight), -2.0f / float(m_windowHeight), 1.0f);	// assuming 90 degree aspect
    glTranslatef (-float(m_windowWidth) / 2.0f, -float(m_windowHeight) / 2.0f, -1.0f);
}


void OpenGLRender::DrawBoxOrtho(int p_left, int p_top, int p_right, int p_bottom)
{
    glBegin (GL_QUADS);
        glColor3ub (0, 0, 0);
        
        glVertex2i (p_left, p_top);
        glVertex2i (p_right, p_top);
        glVertex2i (p_right, p_bottom);
        glVertex2i (p_left, p_bottom);
    glEnd();

	Metrics.triangles += 2;
}

///////////////////////////////////////
// Font functions

void OpenGLRender::RenderText(int p_x, int p_y, char *p_text, CAPIFont &p_font, CColorRGB &p_color, BYTE p_alpha, RECT *p_clipRect)
{
    int startX, startY;
    
	BYTE chr;
    
    bool *charValid;
    BYTE *charWidth;
    float *charStartS;
    float *charStartT;
    
    bool clippingY;
    BYTE clipStartY;
    BYTE clipEndY;

	if (p_font.IsValid())
	{
        clippingY = false;
        clipStartY = 0;
        clipEndY = p_font.GetHeight();
        if (p_clipRect)
		{
			// See if it's impossible to see the text based on coords
			//    and clip rect
            // Check X's
            if (p_clipRect->right <= p_clipRect->left)
				return;
            if (p_clipRect->right <= p_x)
                return;
        
            // Clip Y
            if (p_clipRect->top >= p_clipRect->bottom)
				return;
            if (p_clipRect->top > (p_y + int(p_font.GetHeight())))
				return;
            if (p_clipRect->bottom <= p_y)
				return;

			// Looks like we can see the text, set up for rendering
            if (p_clipRect->top > p_y)
			{
                clippingY = true;
                clipStartY = p_clipRect->top - p_y;
			}
            if (p_clipRect->bottom < int((p_y + p_font.GetHeight())))
			{
                clippingY = true;
                clipEndY = p_clipRect->bottom - p_y;
			}
        }
        
        if (p_font.GetTexture())
		{
        
            if ((p_font.GetTexture()->IsValid()) && ((!p_font.GetTexture()->IsLoaded())))
                p_font.GetTexture()->LoadTextureGL();

            if (p_font.GetTexture()->IsValid())
			{
            
                //  Bind the texture and set up texturing
                //  Clamp it!
                glActiveTextureARB (GL_TEXTURE0_ARB);
                glEnable (GL_TEXTURE_2D);
                glBindTexture (GL_TEXTURE_2D, p_font.GetTexture()->GetGLid());
                glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
                glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
				//glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
				//glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
                glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
                
                glMatrixMode (GL_TEXTURE);
                glLoadIdentity();
                // For some reason, DX divides tex coords by 255 to
                //   work, but GL divides by 256.  Strange.
                glScalef (1.0f / 256.0f, 1.0f / 256.0f, 1.0f);
                glMatrixMode (GL_MODELVIEW);
                
                glActiveTextureARB (GL_TEXTURE1_ARB);
                glDisable (GL_TEXTURE_2D);
            
                charValid = p_font.GetCharValid();
                charWidth = p_font.GetCharWidth();
                charStartS = p_font.GetCharStartS();
                charStartT = p_font.GetCharStartT();
                
                startX = p_x;
                startY = p_y;
                
                glBegin (GL_QUADS);
                if (p_alpha == 255)
                    glColor3ubv (&p_color.red);
                else
                    glColor4ub (p_color.red, p_color.green, p_color.blue, p_alpha);

                unsigned i, j;
                
                if (p_clipRect)
				{
                    j = 1;
                    while (startX < p_clipRect->left)
					{
                        chr = p_text[j-1];
                        if (charValid[chr])
						{
                            if (startX + charWidth[chr] > p_clipRect->left)
							{
                                // Draw character clipped
                                if (startX + charWidth[chr] > p_clipRect->right)
								{
                                    // Draw clipped on both ends and stop
                                    glMultiTexCoord2f (GL_TEXTURE0_ARB, charStartS[chr] + p_clipRect->left - startX, charStartT[chr] + clipStartY);
                                    glVertex2i (p_clipRect->left, p_y + clipStartY);
                                    glMultiTexCoord2f (GL_TEXTURE0_ARB, charStartS[chr] + p_clipRect->right - startX, charStartT[chr] + clipStartY);
                                    glVertex2i (p_clipRect->right, p_y + clipStartY);
                                    glMultiTexCoord2f (GL_TEXTURE0_ARB, charStartS[chr] + p_clipRect->right - startX, charStartT[chr] + clipEndY);
                                    glVertex2i (p_clipRect->right, p_y + clipEndY);
                                    glMultiTexCoord2f (GL_TEXTURE0_ARB, charStartS[chr] + p_clipRect->left - startX, charStartT[chr] + clipEndY);
                                    glVertex2i (p_clipRect->left, p_y + clipEndY);
                                    
                                    startX = startX + charWidth[chr];
                                    j = strlen(p_text) + 1;

									Metrics.triangles += 2;
								}
                                else
								{
                                    // Draw it clipped on the left and proceed to the next loop
                                    glMultiTexCoord2f (GL_TEXTURE0_ARB, charStartS[chr] + p_clipRect->left - startX, charStartT[chr] + clipStartY);
                                    glVertex2i (p_clipRect->left, p_y + clipStartY);
                                    glMultiTexCoord2f (GL_TEXTURE0_ARB, charStartS[chr] + charWidth[chr], charStartT[chr] + clipStartY);
                                    glVertex2i (startX + charWidth[chr], p_y + clipStartY);
                                    glMultiTexCoord2f (GL_TEXTURE0_ARB, charStartS[chr] + charWidth[chr], charStartT[chr] + clipEndY);
                                    glVertex2i (startX + charWidth[chr], p_y + clipEndY);
                                    glMultiTexCoord2f (GL_TEXTURE0_ARB, charStartS[chr] + p_clipRect->left - startX, charStartT[chr] + clipEndY);
                                    glVertex2i (p_clipRect->left, p_y + clipEndY);
                                    
                                    startX = startX + charWidth[chr];
                                    j = j + 1;

									Metrics.triangles += 2;
                                }
							}
                            else
							{
                                // try next character, not into clip yet
                                startX = startX + charWidth[chr];
                                j = j + 1;
                            }
                        }
                    }
                    
                    for (i=j; i<=strlen(p_text); i++)
					{
                        chr = p_text[i-1];
                        if (charValid[chr])
						{
                            if (startX + charWidth[chr] <= p_clipRect->right)
							{
                                glMultiTexCoord2f (GL_TEXTURE0_ARB, charStartS[chr], charStartT[chr] + clipStartY);
                                glVertex2i (startX, p_y + clipStartY);
                                glMultiTexCoord2f (GL_TEXTURE0_ARB, charStartS[chr] + charWidth[chr], charStartT[chr] + clipStartY);
                                glVertex2i (startX + charWidth[chr], p_y + clipStartY);
                                glMultiTexCoord2f (GL_TEXTURE0_ARB, charStartS[chr] + charWidth[chr], charStartT[chr] + clipEndY);
                                glVertex2i (startX + charWidth[chr], p_y + clipEndY);
                                glMultiTexCoord2f (GL_TEXTURE0_ARB, charStartS[chr], charStartT[chr] + clipEndY);
                                glVertex2i (startX, p_y + clipEndY);
                                
                                startX = startX + charWidth[chr];

								Metrics.triangles += 2;
							}
                            else
							{
                                // Clip character on the right and leave
                                glMultiTexCoord2f (GL_TEXTURE0_ARB, charStartS[chr], charStartT[chr] + clipStartY);
                                glVertex2i (startX, p_y + clipStartY);
                                glMultiTexCoord2f (GL_TEXTURE0_ARB, charStartS[chr] + p_clipRect->right - startX, charStartT[chr] + clipStartY);
                                glVertex2i (p_clipRect->right, p_y + clipStartY);
                                glMultiTexCoord2f (GL_TEXTURE0_ARB, charStartS[chr] + p_clipRect->right - startX, charStartT[chr] + clipEndY);
                                glVertex2i (p_clipRect->right, p_y + clipEndY);
                                glMultiTexCoord2f (GL_TEXTURE0_ARB, charStartS[chr], charStartT[chr] + clipEndY);
                                glVertex2i (startX, p_y + clipEndY);
                                
                                i = strlen(p_text);

								Metrics.triangles += 2;
                            }
                        }
                    }
				}
                else
				{
                    // Render without any clipping
                    for (i=1; i<=strlen(p_text); i++)
					{
                        chr = p_text[i-1];;
                        if (charValid[chr])
						{   
                            glMultiTexCoord2f (GL_TEXTURE0_ARB, charStartS[chr], charStartT[chr]);
                            glVertex2i (startX, p_y);
                            glMultiTexCoord2f (GL_TEXTURE0_ARB, charStartS[chr] + charWidth[chr], charStartT[chr]);
                            glVertex2i (startX + charWidth[chr], p_y);
                            glMultiTexCoord2f (GL_TEXTURE0_ARB, charStartS[chr] + charWidth[chr], charStartT[chr] + p_font.GetHeight());
                            glVertex2i (startX + charWidth[chr], p_y + p_font.GetHeight());
                            glMultiTexCoord2f (GL_TEXTURE0_ARB, charStartS[chr], charStartT[chr] + p_font.GetHeight());
                            glVertex2i (startX, p_y + p_font.GetHeight());
                            
                            startX = startX + charWidth[chr];

							Metrics.triangles += 2;
                        }
                    }
                }
                glEnd();
                
                // Restore matrix transform and states
                glActiveTextureARB (GL_TEXTURE0_ARB);
                glMatrixMode (GL_TEXTURE);
                glLoadIdentity();
                glMatrixMode (GL_MODELVIEW);
                glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
                glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
                glDisable (GL_TEXTURE_2D);
            } // texture valid
        }	// texture exists
    }	// font valid
    
}

///////////////////////////////////////
// State routines

void OpenGLRender::EnableDepthTest(bool p_enable) 
{
    if (p_enable)
        glEnable (GL_DEPTH_TEST);
    else
        glDisable (GL_DEPTH_TEST);
}

void OpenGLRender::EnableDepthWrite(bool p_enable)
{
	glDepthMask(p_enable);
}

void OpenGLRender::SetCullCCW(bool p_cullCCW)
{
	cullCCW = p_cullCCW;
	if (cullCCW)
		glFrontFace(GL_CW);
	else
		glFrontFace(GL_CCW);
}

bool OpenGLRender::GetCullCCW(void)
{	
	return cullCCW;
}

void OpenGLRender::SetStencilEnable(bool p_enable) 
{
    if (p_enable)
        glEnable (GL_STENCIL_TEST);
    else
        glDisable (GL_STENCIL_TEST);
}

void OpenGLRender::SetStencilWrite(BYTE p_value) 
{
    glStencilFunc(GL_ALWAYS, p_value, 0xFF);
    glStencilOp(GL_KEEP, GL_KEEP, GL_REPLACE);
}

void OpenGLRender::SetStencilTest(BYTE p_value) 
{
    glStencilFunc(GL_EQUAL, p_value, 0xFF);
    glStencilOp(GL_KEEP, GL_KEEP, GL_KEEP);
}

void OpenGLRender::SetStencilIncrement(BYTE p_value, bool p_ignoreDepth) 
{
    glStencilFunc(GL_EQUAL, p_value, 0xFF);
    if (p_ignoreDepth)
        glStencilOp(GL_KEEP, GL_INCR, GL_INCR);
    else
        glStencilOp(GL_KEEP, GL_KEEP, GL_INCR);
}

void OpenGLRender::SetStencilDecrement(BYTE p_value, bool p_ignoreDepth) 
{
    glStencilFunc(GL_EQUAL, p_value, 0xFF);
    if (p_ignoreDepth)
        glStencilOp(GL_KEEP, GL_DECR, GL_DECR);
    else
        glStencilOp(GL_KEEP, GL_KEEP, GL_DECR);
}

void OpenGLRender::EnableColorWrite(bool p_enable)
{
    glColorMask(p_enable, p_enable, p_enable, p_enable);
}

void OpenGLRender::EnableCulling(bool p_enable)
{
	if (p_enable)
	{
		glEnable(GL_CULL_FACE);
		if (cullCCW)
			glFrontFace(GL_CW);
		else
			glFrontFace(GL_CCW);
	}
	else
		glDisable(GL_CULL_FACE);
}

void OpenGLRender::SetClipPlane(CVector3D &p_planeNormal, CVector3D &p_planePoint)
{
    currentClipPlane[0] = p_planeNormal.x;
    currentClipPlane[1] = p_planeNormal.y;
    currentClipPlane[2] = p_planeNormal.z;
    currentClipPlane[3] = -(p_planeNormal * p_planePoint);

	glGetFloatv(GL_MODELVIEW_MATRIX, currentClipPlaneMatrix);

    glEnable (GL_CLIP_PLANE0);
    glClipPlane (GL_CLIP_PLANE0, currentClipPlane);

	//TRACE("Storing clip plane (%g, %g, %g, %g)\n", currentClipPlane[0],currentClipPlane[1],currentClipPlane[2],currentClipPlane[3]);
}

void OpenGLRender::DisableClipPlane(void)
{
	glDisable(GL_CLIP_PLANE0);
}

void OpenGLRender::SetBlendMode(defAPIBlendMode p_blendMode)
{
    switch (p_blendMode)
	{
    case APIBLENDMODE_DISABLED:
        glDisable (GL_BLEND);
		break;
    case APIBLENDMODE_ALPHABLEND:
        glEnable (GL_BLEND);
        glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		break;
    case APIBLENDMODE_ADDITIVEBLEND:
        glEnable (GL_BLEND);
        glBlendFunc (GL_SRC_ALPHA, GL_ONE);
		break;
    case APIBLENDMODE_DESTINATIONREDUCE:
        glEnable (GL_BLEND);
        glBlendFunc (GL_ZERO, GL_ONE_MINUS_SRC_ALPHA);
		break;
    case APIBLENDMODE_TENDDESTINATIONTOFULL:
        glEnable (GL_BLEND);
        glBlendFunc (GL_ONE_MINUS_DST_COLOR, GL_ONE);
		break;
	}
}

void OpenGLRender::SetDepthTestFunc(defAPIDepthTestFunc p_depthTestFunc)
{
    switch (p_depthTestFunc)
	{
    case APIDEPTHTESTFUNC_LESSEQUAL:
        glDepthFunc(GL_LEQUAL);
		break;
	case APIDEPTHTESTFUNC_ALWAYS:
        glDepthFunc(GL_ALWAYS);
		break;
	}
}

void OpenGLRender::SetZBias(bool p_zBias)
{
	if (p_zBias)
	{
		glEnable(GL_POLYGON_OFFSET_FILL);
		glPolygonOffset (-1.0f, -2.0f);
	}
	else
	{
		glDisable(GL_POLYGON_OFFSET_FILL);
	}
}



