// OpenGLRender.h: interface for the OpenGLRender class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_OPENGLRENDER_H__9A9E9883_6FB0_4915_87AB_BF25B19AEFCE__INCLUDED_)
#define AFX_OPENGLRENDER_H__9A9E9883_6FB0_4915_87AB_BF25B19AEFCE__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "IRender.h"

#include <gl\gl.h>
#include <gl\glu.h>
#include <gl\ext\glext.h>

class OpenGLRender : public IRender
{
public:
	OpenGLRender();
	virtual ~OpenGLRender();

	defRenderAPIType GetAPIType(void);

	void Initialize(HWND p_hWnd);
	void SetViewport(float p_angle, float p_nearZ, float p_farZ);
	void SetViewport(float p_angle);
	void SetViewportOrtho(void);
	void Clear(long p_color);
	void LoadIdentity(void);
	void PushMatrix(void);
	void PopMatrix(void);
	void Translatef(float p_x, float p_y, float p_z);
	void Scalef(float p_x, float p_y, float p_z);
	void Rotate(float p_angle, float p_x, float p_y, float p_z);
	COrient3D GetViewpoint(void);
	void BeginScene(void);
	void RenderTestTriangle(void);
	void EndScene(void);
	void Flip(void);
	void Destroy(void);

	void TransformWorld(COrient3D &p_orient);
	void TransformViewpoint(COrient3D &p_orient);

	bool ConvertEyeVectorToViewportXY(CVector3D &p_vector, int &p_xOut, int &p_yOut);
	CVector3D ConvertViewportXYToWorldVector3D(int p_x, int p_y);

	////////////////////////////////////////
	// Rendering routines

	void RenderTerrain(CTerrain &p_terrain, bool p_useTestColor, bool p_useTexture, CTerrain *p_upperLod, bool p_loX, bool p_loZ, bool p_hiX, bool p_hiZ);

	inline void RenderQuad(CColorRGB p_color, BYTE p_alpha)
	{
		RenderQuad(p_color, p_alpha, NULL);
	}
	void RenderQuad(CColorRGB p_color, BYTE p_alpha, CAPITexture *p_texture);

	inline void RenderMultiQuad(CColorRGB p_color, BYTE p_alpha)
	{
		RenderMultiQuad(p_color, p_alpha, NULL, NULL, APIMULTIQUADTEXTUREOP_MODULATE);
	}
	inline void RenderMultiQuad(CColorRGB p_color, BYTE p_alpha, CAPITexture *p_texture1)
	{
		RenderMultiQuad(p_color, p_alpha, p_texture1, NULL, APIMULTIQUADTEXTUREOP_MODULATE);
	}
	inline void RenderMultiQuad(CColorRGB p_color, BYTE p_alpha, CAPITexture *p_texture1, CAPITexture *p_texture2)
	{
		RenderMultiQuad(p_color, p_alpha, p_texture1, p_texture2, APIMULTIQUADTEXTUREOP_MODULATE);
	}
	void RenderMultiQuad(CColorRGB p_color, BYTE p_alpha, CAPITexture *p_texture1, CAPITexture *p_texture2, defMultiQuadTextureOp p_textureOp);

	inline void RenderModel(CModel3D *p_model)
	{
		RenderModelFinal(p_model, NULL, 255);
	}
	inline void RenderModelColor(CModel3D *p_model, CColorRGB *p_modColor)
	{
		RenderModelFinal(p_model, p_modColor, 255);
	}
	void RenderModelFinal(CModel3D *p_model, CColorRGB *p_modColor, BYTE p_modAlpha);
	void RenderModelFinalLM(CModel3D *p_model, CColorRGB *p_modColor, BYTE p_modAlpha, CAPITexture **p_lightmaps, CSurfaceDecalList *p_decalLists, CLightSourceRefList *p_lightSourceList, COrient3D *p_orient);

	inline void RenderSurfaces(CSurface3D *p_surfaces, unsigned p_surfaceQty, CModelVertex3D *p_vertices, CColorRGB *p_modColor, BYTE p_modAlpha)
	{
		RenderSurfaces(p_surfaces, p_surfaceQty, p_vertices, p_modColor, p_modAlpha, NULL, NULL, true, NULL, NULL, NULL);
	}
	inline void RenderSurfaces(CSurface3D *p_surfaces, unsigned p_surfaceQty, CModelVertex3D *p_vertices, CColorRGB *p_modColor, BYTE p_modAlpha, CAPITexture **p_lightMaps, CSurfaceDecalList *p_decalLists, bool p_ignoreMirrors, CLightSourceRefList *p_lightSourceList)
	{
		RenderSurfaces(p_surfaces, p_surfaceQty, p_vertices, p_modColor, p_modAlpha, p_lightMaps, p_decalLists, p_ignoreMirrors, p_lightSourceList, NULL, NULL);
	}
	void RenderSurfaces(CSurface3D *p_surfaces, unsigned p_surfaceQty, CModelVertex3D *p_vertices, CColorRGB *p_modColor, BYTE p_modAlpha, CAPITexture **p_lightmaps, CSurfaceDecalList *p_decalLists, bool p_ignoreMirrors, CLightSourceRefList *p_lightSourceList, CAPITexture *p_useThisLightmap, COrient3D *p_orient);
	void RenderStarField(CStarField &p_starField, COrient3D *p_currentOrient, COrient3D *p_oldOrient, DWORD p_backColor);
	void RenderParticles(CParticle *p_particles, unsigned int p_particleQty, unsigned int p_renderID, CColorRGB *p_ambientColor, CLightSourceRefList *p_lightSources);

	void RenderWireframeCube(float p_minX, float p_minY, float p_minZ, float p_maxX, float p_maxY, float p_maxZ);

	///////////////////////////////////////
	// Ortho routines
	void RenderText(int p_x, int p_y, char *p_text, CAPIFont &p_font, CColorRGB &p_color, BYTE p_alpha, RECT *p_clipRect);
	void DrawBoxOrtho(int p_left, int p_top, int p_right, int p_bottom);

	///////////////////////////////////////
	// State routines
	void EnableDepthTest(bool p_enable);
	void EnableDepthWrite(bool p_enable);
	void SetCullCCW(bool p_cullCCW);
	bool GetCullCCW(void);
	void SetStencilEnable(bool p_enable);
	void SetStencilWrite(BYTE p_value);
	void SetStencilTest(BYTE p_value);
	void SetStencilIncrement(BYTE p_value, bool p_ignoreDepth);
	void SetStencilDecrement(BYTE p_value, bool p_ignoreDepth);
	void EnableColorWrite(bool p_enable);
	void EnableCulling(bool p_enable);
	void SetClipPlane(CVector3D &p_planeNormal, CVector3D &p_planePoint);
	void DisableClipPlane(void);
	void SetBlendMode(defAPIBlendMode p_blendMode);
	void SetDepthTestFunc(defAPIDepthTestFunc p_depthTestFunc);
	void SetZBias(bool p_zBias);

//	float GetScaleY(void) {return m_scaleY;}

	// Render buffer routines
	inline void SwitchToDynamicBuffer(void)
	{
        glBindFramebufferEXT (GL_FRAMEBUFFER_EXT, dynLMFB);
        offScreenMode = true;
	}
	inline void SwitchToMainBuffer(void)
	{
		glBindFramebufferEXT (GL_FRAMEBUFFER_EXT, 0);
		offScreenMode = false;
	}

private:
	HWND m_hWnd;		// Window handle for rendering buffer
	HDC m_hDC;			// Device context for rendering buffer
	HGLRC m_hRC;		// Rendering context for OpenGL instance

	long m_windowWidth;	// Render window width
	long m_windowHeight;// Render window height
	float m_scaleY;		// Number of vertical (and horizonal) pixels to
						//    equal 45 degrees viewed from 1 unit away
	bool cullCCW;		// culling mode

	bool offScreenMode;	// Are we rendering to the off-screen buffer?
	GLuint dynLMFB;		// Frame buffer and depth buffer references
	GLuint dynLMDB;		//   for offscreen rendering

	GLdouble currentClipPlane[4];	// latest clip plane applied
									//   used to restore clip plane in
									//   RenderSurfaces
	GLfloat currentClipPlaneMatrix[16];	// Matrix at the time the
										//   clip plane was set (same purpose)

	// Multitexturing functions
	PFNGLMULTITEXCOORD2FARBPROC glMultiTexCoord2f;
	PFNGLACTIVETEXTUREARBPROC	glActiveTextureARB;

	// Off-screen renderbuffer functions
	PFNGLGENFRAMEBUFFERSEXTPROC glGenFramebuffersEXT;
	PFNGLBINDFRAMEBUFFEREXTPROC glBindFramebufferEXT;
	PFNGLGENRENDERBUFFERSEXTPROC glGenRenderbuffersEXT;
	PFNGLBINDRENDERBUFFEREXTPROC glBindRenderbufferEXT;
	PFNGLRENDERBUFFERSTORAGEEXTPROC glRenderbufferStorageEXT;
	PFNGLFRAMEBUFFERRENDERBUFFEREXTPROC glFramebufferRenderbufferEXT;
	PFNGLFRAMEBUFFERTEXTURE2DEXTPROC glFramebufferTexture2DEXT;
	PFNGLCHECKFRAMEBUFFERSTATUSEXTPROC glCheckFramebufferStatusEXT;
	PFNGLDELETEFRAMEBUFFERSEXTPROC glDeleteFramebuffersEXT;
	PFNGLDELETERENDERBUFFERSEXTPROC glDeleteRenderbuffersEXT;
};

#endif // !defined(AFX_OPENGLRENDER_H__9A9E9883_6FB0_4915_87AB_BF25B19AEFCE__INCLUDED_)
