// ParticleEmitter.cpp: implementation of the CParticleEmitter class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "ParticleEmitter.h"
#include "..\globals.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CLinkedList<CParticleEmitter> CLinkedList<CParticleEmitter>::DeletedList;

// If return false, particle should be removed
bool CParticle::Animate(DWORD p_elapsedTimeMS)
{
	float elapsedTimef = float(p_elapsedTimeMS);

	// Store old position in case it moves
	oldPosition = position;

	if (lifeMS > 0)
	{
		if (lifeMS <= p_elapsedTimeMS)
			return false;
		else
			lifeMS -= p_elapsedTimeMS;
	}
	if (sizei)
	{
		size += (sizei * elapsedTimef);
		if (size <= 0)
			return false;
	}
	if (alphaAmounti)
	{
		alphaAmount += (alphaAmounti * elapsedTimef);
		if (alphaAmount <= 0)
			return false;
	}

	if (gravity.x)
	{
		position.x += (velocity.x * elapsedTimef + 0.5f*gravity.x*elapsedTimef*elapsedTimef);
		velocity.x += (gravity.x * elapsedTimef);
	}
	else
		if (velocity.x)
			position.x += (velocity.x * elapsedTimef);

	if (gravity.y)
	{
		position.y += (velocity.y * elapsedTimef + 0.5f*gravity.y*elapsedTimef*elapsedTimef);
		velocity.y += (gravity.y * elapsedTimef);
	}
	else
		if (velocity.y)
			position.y += (velocity.y * elapsedTimef);

	if (gravity.z)
	{
		position.z += (velocity.z * elapsedTimef + 0.5f*gravity.z*elapsedTimef*elapsedTimef);
		velocity.z += (gravity.z * elapsedTimef);
	}
	else
		if (velocity.z)
			position.z += (velocity.z * elapsedTimef);

	return true;
}

CParticleSet::CParticleSet()
{
	particles = NULL;
	maxParticleQty = 0;
	currentParticleQty = 0;
}

CParticleSet::~CParticleSet()
{
	Destroy();
}

void CParticleSet::Destroy(void)
{
	if (particles)
		delete [] particles;
	particles = NULL;

	maxParticleQty = 0;
	currentParticleQty = 0;
}

void CParticleSet::Initialize(unsigned int p_maxParticleQty)
{
	maxParticleQty = p_maxParticleQty;
	if (maxParticleQty > 0)
		particles = new CParticle[maxParticleQty];
	currentParticleQty = 0;
	lastTimeAnimated = 0;
}

CParticle * CParticleSet::AddParticle(void)		// returns pointer to particle to initialize
{
	if (currentParticleQty < maxParticleQty)
	{
		currentParticleQty++;
		//particles[currentParticleQty-1].animated = false;
		return &particles[currentParticleQty-1];
	}
	else
	{
		// Find the particle with the lowest lifeMS remaining and make room
		//unsigned int i;
		//for (i=0; i<currentParticleQty; i++)
		return NULL;
	}
}

void CParticleSet::AnimateParticles(DWORD p_elapsedTimeMS, DWORD p_currentTime)
{
	DWORD elapsedTimeMS;

	if (lastTimeAnimated == p_currentTime)
		return;
	else
	{
		// Base animation on last time the particle set was aniimated,
		//   not the last game tick (p_elapsedTimeMS isn't really necessary)
		elapsedTimeMS = p_currentTime - lastTimeAnimated;
		lastTimeAnimated = p_currentTime;
	}

	if (particles)
	{
		CParticle *lpParticle = &particles[0];
		unsigned int i;
		for (i=0; i<currentParticleQty; i++)
		{
			if (!lpParticle->animated)
			{
				if (!lpParticle->Animate(elapsedTimeMS))
				{
					// Get rid of particle, keep loop on same element
					*lpParticle = particles[currentParticleQty-1];
					currentParticleQty--;
					i--;
				}
				else
					lpParticle++;
			}
			else
			{
				// We skipped this particle, lastTimeAnimated
				//   will prevent it from being animated again this tick
				lpParticle->animated = false;
				lpParticle++;
			}
		}
	}

}

////////////////////////////////////////

CParticleEmitter::CParticleEmitter()
{
	position.SetVec(0,0,0);

	behavior = APIPARTICLEBEHAVIOR_NONE;
	frequency = 0;
	timerMS = 0;

	particleSetRef = NULL;

	intRef = -1;
}

CParticleEmitter::~CParticleEmitter()
{
	Destroy();
}

void CParticleEmitter::Destroy(void)
{
	particleSetRef = NULL;
}

void CParticleEmitter::Animate(DWORD p_elapsedTimeMS, DWORD p_currentTime)
{
	if (lastTimeAnimated == p_currentTime)
		return;
	else
		lastTimeAnimated = p_currentTime;

	if (p_elapsedTimeMS == 0)
	{
		//oldPosition = position;
		return;
	}

	DWORD timeSpent = 0;			// How much time of p_elapsedTimeMS have we consumed so far?
	DWORD remainingTimeMS = p_elapsedTimeMS;	// How much time left of p_elaspedTimeMS to consume?

	// Add particles according to timing and behavior
	CParticle *lpParticle;
	if (particleSetRef)
	if (frequency > 0)
	{
		CVector3D particlePosition = position;	// new particle position
		
		while (remainingTimeMS >= timerMS)
		{
			// NOTE: particleSetRef could be the wrong particle set,
			//   since we are placing the particle between
			//   oldPosition and position.  That location could be
			//   where a different particle set handles the
			//   particles.  It's up to the rest of the engine
			//   to make sure particles are places in the correct
			//   particle sets as they move
			lpParticle = particleSetRef->AddParticle();

			timeSpent += timerMS;

			if (oldPosition != position)
			{
				// Use old and current position, and amount of time
				//   to accurately place the particle along the path of
				//   the emitter
				particlePosition = oldPosition + ((position - oldPosition) * (float(timeSpent)/float(p_elapsedTimeMS)));
			}

			if (lpParticle)
			switch (behavior)
			{
				case APIPARTICLEBEHAVIOR_TORCHFLAME:
					lpParticle->alpha = 255;
					//lpParticle->position = CVector3D::Vector3D(Random() * 1.0f-0.5f, 0.0f, Random() * 1.0f-0.5f);
					lpParticle->position = particlePosition + CVector3D::Vector3D(Random() * 0.25f - 0.125f, 0.0f, Random() * 0.25f-0.125f);
					lpParticle->color.SetColor(255,BYTE(Random() * 30 + 205),0);
					lpParticle->size = 1.0f;
					//lpParticle->radius = Random() * 0.5f + 0.5f;
					lpParticle->radius = Random() * 0.15f + 0.15f;
					lpParticle->alphaAmount = 1.0f;
					lpParticle->texture = APITextures.GetTexture("lightmap9");

					lpParticle->alphaAmounti = -0.0005f;
					//lpParticle->sizei = -0.0015f * Random() - 0.0015f;
					lpParticle->sizei = -0.0020f * Random() - 0.0020f;
					lpParticle->lifeMS = 1000;
					lpParticle->gravity = CVector3D::Vector3D(0,0.0000025f,0);
					lpParticle->velocity = CVector3D::Vector3D(0,0.0025f,0);

					lpParticle->emitting = true;

					// Animate the particle forward to current time
					lpParticle->Animate(remainingTimeMS);
					// And make sure the particle loop doesn't animate it
					lpParticle->animated = true;

					break;
				case APIPARTICLEBEHAVIOR_STREAMTEST:
					lpParticle->alpha = 255;
					//lpParticle->position = CVector3D::Vector3D(Random() * 1.0f-0.5f, 0.0f, Random() * 1.0f-0.5f);
					lpParticle->position = particlePosition;
					lpParticle->color.SetColor(255,255,255);
					lpParticle->size = 1.0f;
					//lpParticle->radius = Random() * 0.5f + 0.5f;
					lpParticle->radius = 0.30f;
					lpParticle->alphaAmount = 1.0f;
					lpParticle->texture = APITextures.GetTexture("lightmap9");

					lpParticle->alphaAmounti = 0;
					//lpParticle->sizei = -0.0015f * Random() - 0.0015f;
					lpParticle->sizei = 0;
					lpParticle->lifeMS = 20000;
					lpParticle->gravity = CVector3D::ZeroVector();
					lpParticle->velocity = direction * 0.01f;

					//lpParticle->emitting = false;
					lpParticle->emitting = true;

					// Animate the particle forward to current time
					lpParticle->Animate(remainingTimeMS);
					// And make sure the particle loop doesn't animate it
					lpParticle->animated = true;

					break;
			}

			remainingTimeMS -= timerMS;
			timerMS = frequency;
		}

		timerMS -= remainingTimeMS;
	}

	// update old position
	oldPosition = position;
}

