// ParticleEmitter.h: interface for the CParticleEmitter class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_PARTICLEEMITTER_H__3D6B686A_094B_45EA_BD17_AD1730008F6F__INCLUDED_)
#define AFX_PARTICLEEMITTER_H__3D6B686A_094B_45EA_BD17_AD1730008F6F__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "..\math\Vector3D.h"
#include "ColorRGB.h"
#include "APITexture.h"
#include "..\structs\LinkedList.h"

enum defParticleBehavior
{
	APIPARTICLEBEHAVIOR_NONE = 0,
	APIPARTICLEBEHAVIOR_STREAMTEST,
	APIPARTICLEBEHAVIOR_TORCHFLAME,
	APIPARTICLEBEHAVIOR_FADEWITHLIFE,
};

class CParticle
{
public:
	unsigned int renderID;		// tag to prevent multiple calculation
								//   of lit particle color (mirrors)
	CColorRGB	renderColor;	// Calculated render color (lit only)

	CVector3D position;		// current position
	CVector3D oldPosition;	// old position (for detecting movement through portals only)
	CVector3D velocity;		// movement offset per MS
	CVector3D gravity;		// which way will the particle try to move?
	CColorRGB color;		// color to mod rendered quad
	BYTE alpha;				// base alpha
	float radius;			// base size of quad

	bool emitting;			// Unaffected by light sources?

	// These range from 0.0-1.0 to affect alpha and radius (modulate)
	float alphaAmount;
	float size;
	float alphaAmounti;
	float sizei;

	CAPITexture *texture;	// which texture to use to render
	unsigned int lifeMS;	// life in MS, if 0 lasts forever

	bool Animate(DWORD p_elapsedTimeMS);

	DWORD lastTimeAnimated;
	bool animated;			// Was it already aniamted this tick?
};

// A single set of particles
class CParticleSet
{
public:
	CParticleSet();
	virtual ~CParticleSet();
	void Destroy(void);

	void Initialize(unsigned int p_maxParticleQty);
	CParticle * GetParticles(void) {return particles;}
	unsigned int GetParticleQty(void) {return currentParticleQty;}

	inline bool AlreadyAnimated (DWORD p_currentTime)
		{return (lastTimeAnimated == p_currentTime);}

	CParticle *particles;		// Array of particles, always left-justified
								// Whenever one is removed, the last one is
								//   copied to that location and currentQty
								//   reduced
	unsigned int maxParticleQty;		// How many can we create?
	unsigned int currentParticleQty;	// How many are set up now?
	CParticle * AddParticle(void);		// returns pointer to particle to initialize

	void AnimateParticles(DWORD p_elapsedTimeMS, DWORD p_currentTime);

	inline void RemoveParticle(unsigned int p_id)
	{
		particles[p_id] = particles[currentParticleQty-1];
		currentParticleQty--;
	}

	inline bool SetTimeRoomsCalculated(DWORD p_currentTime)
	{
		if (lastTimeRoomsCalculated == p_currentTime)
			return false;
		else
		{
			lastTimeRoomsCalculated = p_currentTime;
			return true;
		}
	}

private:
	DWORD lastTimeAnimated;
	DWORD lastTimeRoomsCalculated;
};

class CParticleEmitter  
{
public:
	CParticleEmitter();
	virtual ~CParticleEmitter();
	void Destroy(void);

	CParticleSet * particleSetRef;

	void SetParticleSet(CParticleSet *p_particleSet) {particleSetRef = p_particleSet;};

	CVector3D position;			// position of the emitter, where applicable
	CVector3D direction;		// parameter for particle creation

	CVector3D oldPosition;		// stored in case emitter moves so we can
								//    calculate which room it's in
								// NOTE: It is vital that this only be set to
								//     position AFTER all particles have been
								//     generated for the given game tick!!!

	//void RemoveParticle(unsigned int p_id);
	void Animate(DWORD p_elapsedTimeMS, DWORD p_currentTime);

	DWORD frequency;			// How often do we make them? (0 means never,
								//   otherwise it's one per particleFrequency MS,
								//   (1000 is 1/second, 500 is 2/second, etc.)
	defParticleBehavior	behavior;	// What kind of particles should we make?

	int intRef;			// General purpose reference, for portal map,
						//   it's the room the emitter is in the list of

	// We'll need a definition of behavior and various parameters (floats,
	//   vectors, textures, etc.) so we can create and animate particles
	//   properly.  (Jet engine, torch, etc.) and support user-defined ones as
	//   well, I guess, somehow.
private:
	DWORD timerMS;
	DWORD lastTimeAnimated;
};

class CParticleEmitterList : public CLinkedList<CParticleEmitter>
{
public:
	void EmptyList(void)
	{
		CLinkedListNode<CParticleEmitter> *node = GetFirst();
		if (node)
		while (node->next != NULL)
		{
			node->Destroy();
			node = node->next;
		}

		CLinkedList<CParticleEmitter>::EmptyList();
	}

	void Delete(CLinkedListNode<CParticleEmitter> *p_node)
	{
		p_node->Destroy();
		CLinkedList<CParticleEmitter>::Delete(p_node);
	}

	CLinkedListNode<CParticleEmitter> * NewNode(void)
	{
		CLinkedListNode<CParticleEmitter> *node;
		node = CLinkedList<CParticleEmitter>::NewNode();
		node->intRef = -1;
		return node;
	}
};

#endif // !defined(AFX_PARTICLEEMITTER_H__3D6B686A_094B_45EA_BD17_AD1730008F6F__INCLUDED_)
