// RenderObject.cpp: implementation of the CRenderObject class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "RenderObject.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CLinkedList<CRenderObject> CLinkedList<CRenderObject>::DeletedList;

CRenderObject::CRenderObject()
{
	scalef.SetVec(1,1,1);
	modelRef = NULL;
	renderID = 0;
	controlObject = false;
}

CRenderObject::~CRenderObject()
{
	// No need to destroy modelRef, since it's only a reference
}

void CRenderObject::DetermineCellsTouched(COctree &p_octree)
{
	if (!modelRef) return;

	CVector3D minPoint;
	CVector3D maxPoint;

	COctCellRefList cellList;

	float radius = modelRef->GetRadius() * scalef.x;
	if ((modelRef->GetRadius() * scalef.y) > radius)
		radius = modelRef->GetRadius() * scalef.y;
	if ((modelRef->GetRadius() * scalef.z) > radius)
		radius = modelRef->GetRadius() * scalef.z;

	minPoint.x = orient.p.x - radius;
	minPoint.y = orient.p.y - radius;
	minPoint.z = orient.p.z - radius;
	maxPoint.x = orient.p.x + radius;
	maxPoint.y = orient.p.y + radius;
	maxPoint.z = orient.p.z + radius;

	p_octree.CellsTouched(minPoint.x, minPoint.y,minPoint.z, maxPoint.x, maxPoint.y, maxPoint.z, cellList);

	// Update cell object lists
	CLinkedListNode<COctCellRef> *cnode1, *cnode2;
	CLinkedListNode<CLPVoid> *vnode1;
	bool found;

	// For every cell in the old list that is not in the new list,
	//   remove the object from those cells
	cnode1 = cellsTouched.GetFirst();
	if (cnode1)
	while (cnode1->next != NULL)
	{
		cnode2 = cellList.GetFirst();
		found = false;

		if (cnode2)
		while ((cnode2->next != NULL) && (!found))
		{
			if (cnode1->cellRef == cnode2->cellRef)
				found = true;
			else
				cnode2 = cnode2->next;
		}

		// Not there, remove this object from that cell's list
		if (!found)
		{
			vnode1 = cnode1->cellRef->lpVoidList.GetFirst();
			found = false;
			if (vnode1)
			while ((vnode1->next != NULL) && (!found))
			{
				if (vnode1->lpVoid == this)
				{
					found = true;
					cnode1->cellRef->lpVoidList.Delete(vnode1);
				}
				else
					vnode1 = vnode1->next;
			}
		}

		cnode1 = cnode1->next;
	}

	// For every cell in the new list that is not in the old list,
	//   add the object to those cells
	cnode1 = cellList.GetFirst();
	if (cnode1)
	while (cnode1->next != NULL)
	{
		cnode2 = cellsTouched.GetFirst();
		found = false;

		if (cnode2)
		while ((cnode2->next != NULL) && (!found))
		{
			if (cnode1->cellRef == cnode2->cellRef)
				found = true;
			else
				cnode2 = cnode2->next;
		}

		// Not there, add this object to that cell's list
		if (!found)
		{
			vnode1 = cnode1->cellRef->lpVoidList.NewNode();
			vnode1->lpVoid = this;
		}

		cnode1 = cnode1->next;
	}

	// Finally, replace the object's old list with the new list
	cellsTouched.EmptyList();
	if (!cellList.IsEmpty())
	{
		cellsTouched.GetHeader()->next = cellList.GetHeader()->next;
		cellsTouched.GetFooter()->prev = cellList.GetFooter()->prev;
		cellList.GetFirst()->prev = cellsTouched.GetHeader();
		cellList.GetLast()->next = cellsTouched.GetFooter();
		cellList.GetHeader()->next = cellList.GetFooter();
		cellList.GetFooter()->prev = cellList.GetHeader();
	}
}
