// Terrain.cpp: implementation of the CTerrain class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "Terrain.h"
#include "IRender.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CTerrainPoint::CTerrainPoint()
{
	color = 255;
}

CTerrainPoint::~CTerrainPoint()
{

}

//////////////////////////////////////////////

CTerrain::CTerrain()
{
	xQty = 0;
	zQty = 0;
	points = NULL;
	startS0 = 0.0f;
	startT0 = 0.0f;
	startS1 = 0.0f;
	startT1 = 0.0f;
}

CTerrain::~CTerrain()
{
	Destroy();
}

void CTerrain::Destroy(void)
{
	xQty = 0;
	zQty = 0;
	if (points)
		delete [] points;
	points = NULL;
}

void CTerrain::Initialize(unsigned p_xQty, unsigned p_zQty, float p_cellSize)
{   
    xQty = p_xQty;
    zQty = p_zQty;
    cellSize = p_cellSize;

	if ((xQty) && (zQty))
		points = new CTerrainPoint[xQty*zQty];
}

void CTerrain::CalculateSurfaceData(bool p_calcAlpha)
{
	unsigned x, z;

	CVector3D v1, v2;

	for (x=0; x<=xQty-2; x++)
	{
		for (z=0; z<=zQty-2; z++)
		{
    
			// Calculate primary
			v1.y = points[x + (z + 1)*xQty].y - points[x + z*xQty].y;
				v2.y = points[x + 1 + (z*xQty)].y - points[x + z*xQty].y;
        
			v1.x = 0;
			v1.z = cellSize;
			v2.x = cellSize;
			v2.z = 0;
        
			points[x + z*xQty].primaryNormal = v1.CrsProd(v2);
			points[x + z*xQty].primaryNormal.Unitize();
        
			// Calculate secondary
			v1.y = points[x + (z + 1)*xQty].y - points[x + 1 + (z + 1)*xQty].y;
			v2.y = points[x + 1 + z*xQty].y - points[x + 1 + (z + 1)*xQty].y;
        
			v1.x = -cellSize;
			v1.z = 0;
			v2.x = 0;
			v2.z = -cellSize;
        
			points[x + z*xQty].secondaryNormal = v2.CrsProd(v1); // Might need to reverse these
			points[x + z*xQty].secondaryNormal.Unitize();
        
			// Cheating, make sure normals face upwards
			// Could just make sure vertices used are in clockwise direction, but bleh
			if (points[x + z*xQty].primaryNormal.y < 0.0f)
				points[x + z*xQty].primaryNormal = points[x + z*xQty].primaryNormal * (-1.0f);

			if (points[x + z*xQty].secondaryNormal.y < 0.0f)
				points[x + z*xQty].secondaryNormal = points[x + z*xQty].secondaryNormal * (-1.0f);
        
			// Calculate K's so CalculateYatXZ works (n dot v = k)
			points[x + z*xQty].pK = points[x + z*xQty].y * points[x + z*xQty].primaryNormal.y;
			points[x + z*xQty].sK = points[x + z*xQty].secondaryNormal.x * cellSize + points[x + 1 + (z + 1)*xQty].y * points[x + z*xQty].secondaryNormal.y + points[x + z*xQty].secondaryNormal.z * cellSize;
		}
	}

	// Now calculate highest and lowest points
	highPoint.y = points[0].y;
	lowPoint.y = points[0].y;

	for (x=0; x<=xQty - 1; x++)
	{
		for (z=0; z<=zQty - 1; z++)
		{
			if (points[x + z*xQty].y > highPoint.y) highPoint.y = points[x + z*xQty].y;
			if (points[x + z*xQty].y < lowPoint.y) lowPoint.y = points[x + z*xQty].y;
		}
	}

	highPoint.x = cellSize * (xQty - 1);
	highPoint.z = cellSize * (zQty - 1);
	lowPoint.x = 0;
	lowPoint.z = 0;

	// Now, calculate point normals from surrounding surfaces and
	//   initialize point colors as if illuminated from straight above
	for (x=0; x<=xQty - 1; x++)
	for (z=0; z<=zQty - 1; z++)
	{
		v1.SetVec(0, 0, 0);
    
		if ((x < xQty - 1) && (z < zQty - 1))
			v1 = points[x + z*xQty].primaryNormal;

		if ((x > 0) && (z < zQty - 1))
		{
			v1 = v1 + points[x - 1 + z*xQty].secondaryNormal;
			v1 = v1 + points[x - 1 + z*xQty].primaryNormal;
		}

		if ((x > 0) && (z > 0))
			v1 = v1 + points[x - 1 + (z - 1)*xQty].secondaryNormal;

		if ((z > 0) & (x < xQty - 1))
		{
			v1 = v1 + points[x + (z - 1)*xQty].secondaryNormal;
			v1 = v1 + points[x + (z - 1)*xQty].primaryNormal;
		}
    
		v1.Unitize();
		points[x + z*xQty].pointNormal = v1;
    
		// Initialize a light value as if illuminated from straight above
		points[x + z*xQty].color = BYTE(255 * points[x + z*xQty].pointNormal.y);
		points[x + z*xQty].color = points[x + z*xQty].color/2 + 128;
		//points[x + z*xQty].color = 255;
    
		// And initialize blending alpha
		if (p_calcAlpha)
			points[x + z*xQty].alpha1 = BYTE((sqrt(1.0f - points[x + z*xQty].pointNormal.y)) * 255.0f);
		//points(x, z).alpha1 = 128
	}
}

CVector3D CTerrain::GetNormalAtXZ(float p_x, float p_z)
{
	unsigned x, z;
	float xi, zi;

	x = int(p_x / cellSize);
	z = int(p_z / cellSize);
	if (x < 0) x = 0;
	if (x > xQty - 2) x = xQty - 2;
	if (z < 0) z = 0;
	if (z > zQty - 2) z = zQty - 2;

	xi = p_x - x * cellSize;
	zi = p_z - z * cellSize;

	if ((cellSize - xi) > zi)
		// Use primary
		return points[x + z*xQty].primaryNormal;
	else
		// Use secondary
		return points[x + z*xQty].secondaryNormal;
}

float CTerrain::CalculateYAtXZ(float p_x, float p_z)
{
	unsigned x, z;
	float xi, zi;
	CTerrainPoint *tmp;

	x = int(p_x / cellSize);
	z = int(p_z / cellSize);
	if (x < 0) x = 0;
	if (x > xQty - 2) x = xQty - 2;
	if (z < 0) z = 0;
	if (z > zQty - 2) z = zQty - 2;

	xi = p_x - x * cellSize;
	zi = p_z - z * cellSize;

	// Ah, the magic of the planar equation n dot v = k
	tmp = &points[x + z*xQty];
	if ((cellSize - xi) > zi)
	{
		// Use primary
		// Can base off 0,0
		return ((tmp->pK - tmp->primaryNormal.x * xi - tmp->primaryNormal.z * zi) / tmp->primaryNormal.y);
	}
	else
	{
		// Use secondary
		// Will need a constant (remember the plane formulas?)
		return ((tmp->sK - tmp->secondaryNormal.x * xi - tmp->secondaryNormal.z * zi) / tmp->secondaryNormal.y);
	}
}

void CTerrain::LoadFromBMP(char *p_BMPfileName, float p_cellSize, float p_minY, float p_maxY)
{
	FILE *fp;
	BYTE row[4096];

	unsigned width, height;
	unsigned start;

	fp = fopen(p_BMPfileName, "rb");
	if (fp)
	{
		// BMP files:
		// 18 - width
		// 22 - height
		// 10 - start of data
		fseek(fp, 18, SEEK_SET);
		fread(&width, 4, 1, fp);
		fseek(fp, 22, SEEK_SET);
		fread(&height, 4, 1, fp);

		Initialize(width, height, p_cellSize);

		fseek(fp, 10, SEEK_SET);
		fread(&start, 4, 1, fp);

		fseek(fp, start, SEEK_SET);

		unsigned x, z;
		BYTE color1;
		for (z=0; z<height; z++)
		{
			fread(row, width*3, 1, fp);

			for (x=0; x<width; x++)
			{
				color1 = row[x*3];
				// BMP is upside down in file
				points[x+(height-z-1)*xQty].y = p_minY + (float(color1) / 255.0f) * (p_maxY - p_minY);
			}
		}

		CalculateSurfaceData();

		fclose (fp);
	}
	else
	{
		TRACE("* Warning - Unable to open terrain file %s\n", p_BMPfileName);
	}
}

bool CTerrain::RayCollision(CVector3D &p_start, CVector3D &p_directionUnit, float &p_collisionDistanceOut, CVector3D &p_collisionPointOut)
{
	// Partition volumes would also be very helpful
	// ?? !!!!

	// ?? Also, the tolerance values should probably be a ratio of
	//   cellSize

	bool result = false;

	float tolerance;
	tolerance = cellSize / 4000.0f;

	int x, z;
	float nDotD;
	float d;
	CVector3D pp;	// point on both primary and secondary planes
	CVector3D po;	// offset from point on planes to ray start

	float xo, zo;	// x,z offset of direction
	float xp, zp;	// current position of incremented x,z in terrain
	float xt, zt;	// next x,z targets for xp, zp to reach using xo, zo

	CVector3D point;	// test collision point

	// Find the first cell that would could possibly collide with and
	//   work across the terrain, checking ONLY those cells we could collide with

	// ?? This could be further sped up with colision volumes encompassing
	//   many cells, and checking lower and upper y bounds

	if ((fabs(p_directionUnit.z) < 0.00001f) && (fabs(p_directionUnit.x) < 0.00001f))
	{
		xo = 0;
		zo = 0;
	}
	else
	{
		xo = p_directionUnit.x;
		zo = p_directionUnit.z;
	}

	// Calculate starting x and z
	x = int(p_start.x / cellSize);
	z = int(p_start.z / cellSize);
	xp = p_start.x;
	zp = p_start.z;

	// ?? Fix x,z if going negative and are on a border
	if (((p_start.x / cellSize) == x) && (xo < 0))
		x = x - 1;

	if (((p_start.z / cellSize) == z) && (zo < 0))
		z = z - 1;

	// If the starting cell is invalid, calculate the first
	//   valid cell we reach
	if (((x < 0) || (x > (int(xQty) - 2))) && (xo != 0))
	{
		// Get a better x starting point
		if ((xo < 0) && (p_start.x > (xQty - 1) * cellSize))
		{
			xp = (xQty - 1) * cellSize;
			zp = p_start.z + zo * -((p_start.x - (xQty - 1) * cellSize) / xo);
			x = xQty - 2;
			z = int(zp / cellSize);
		}
		else
		{
			if ((xo > 0) && (p_start.x < 0))
			{
				xp = 0;
				zp = p_start.z + zo * -(p_start.x / xo);
				x = 0;
				z = int(zp / cellSize);
			}
		}
	}
	if (((z < 0) || (z > (int(zQty) - 2))) && (zo != 0))
	{
		// Get a better x starting point
		if ((zo < 0) && (p_start.z > (zQty - 1) * cellSize))
		{
			xp = p_start.x + xo * -((p_start.z - (zQty - 1) * cellSize) / zo);
			zp = (zQty - 1) * cellSize;
			x = int(xp / cellSize);
			z = zQty - 2;
		}
		else
		{
			if ((zo > 0) && (p_start.z < 0))
			{
				xp = p_start.x + xo * (-p_start.z / zo);
				zp = 0;
				x = int(xp / cellSize);
				z = 0;
			}
		}
	}

	// Initialize xt and zt targets
	if (xo > 0)
		xt = (x + 1) * cellSize;
	else
		xt = x * cellSize;

	if (zo > 0)
		zt = (z + 1) * cellSize;
	else
		zt = z * cellSize;

	// Do tests until we collide or leave the terrain square
	while ((x >= 0) && (z >= 0) && (x <= (int(xQty) - 2)) && (z <= (int(zQty) - 2)))
	{
		// Test current x,z square
		pp.SetVec(x * cellSize, points[x +(z + 1)*xQty].y, (z + 1) * cellSize);
		po = p_start - pp;

		// Check primary normal
		nDotD = points[x + z*xQty].primaryNormal * p_directionUnit;
		if ((fabs(nDotD) > 0.000001f) && (nDotD < 0))
		{
			d = -(points[x + z*xQty].primaryNormal * po) / nDotD;
        
			if ((d >= -0.001f) && (d < p_collisionDistanceOut))
			{
				// Get the collision point
				point = p_start + (p_directionUnit * d);
				// Make sure it falls within the cell
				if ((point.x >= x * cellSize - tolerance) && (point.x <= ((x + 1) * cellSize + tolerance)) && (point.z >= (z * cellSize - tolerance)) && (point.z <= ((z + 1) * cellSize + tolerance)))
				{
					// Make sure it's in the primary triangle
					if ((point.x - x * cellSize) <= (cellSize - (point.z - z * cellSize) + tolerance))
					{
						result = true;
						p_collisionDistanceOut = d;
						p_collisionPointOut = point;
					}
				}
			}
		}
    
		// Check secondary normal, even if we found a collision on the primary
		//   triangle
		nDotD = points[x + z*xQty].secondaryNormal * p_directionUnit;
		if ((fabs(nDotD) > 0.000001f) && (nDotD < 0))
		{
			d = -(points[x + z*xQty].secondaryNormal * po) / nDotD;
        
			if ((d >= -0.001f) && (d < p_collisionDistanceOut))
			{
				// Get the collision point
				point = p_start + (p_directionUnit * d);
				// Make sure point falls within the cell
				if (((point.x >= (x * cellSize - tolerance))) && (point.x <= ((x + 1) * cellSize + tolerance)) && (point.z >= (z * cellSize - tolerance)) && (point.z <= ((z + 1) * cellSize + tolerance)))
				{
					// Make sure it's in the secondary triangle
					if ((point.x - x * cellSize) >= (cellSize - (point.z - z * cellSize) - tolerance))
					{
						result = true;
						p_collisionDistanceOut = d;
						p_collisionPointOut = point;
					}
				}
			}
		}
    
		// We have our collision, no need to check more
		// We don't need to check anymore since we are checking in the
		//   direction of the ray.  First collision = result.
		if (result) return result;
    
		// Get next x,z
		// (Advance xp and zp, increment x or z or both, choose new xt, zt)
		if ((xo == 0) && (zo == 0))
		{
			// We are done, only checked the one cell
			return result;
		}
		else
		{
			if (xo == 0)
			{
				if (zo > 0)
				{
					zp = zp + cellSize;
					z = z + 1;
				}
				else
				{
					zp = zp - cellSize;
					z = z - 1;
				}
			}
			else
			{
				if (zo == 0)
				{
					if (xo > 0)
					{
						xp = xp + cellSize;
						x = x + 1;
					}
					else
					{
						xp = xp - cellSize;
						x = x - 1;
					}
				}
				else
				{
					// Sloped line, do the complicated stuff
					// Find first xt or zt reached and advance
					//   x, z, xp, and zp accordingly
					if (((xt - xp) / xo) < ((zt - zp) / zo))
					{
						// x gets there first
						zp = zp + zo * (xt - xp) / xo;
						xp = xt;
						if (xo > 0)
						{
							xt = xt + cellSize;
							x = x + 1;
						}
						else
						{
							xt = xt - cellSize;
							x = x - 1;
						}
					}
					else
					{
						// z gets there first
						xp = xp + xo * (zt - zp) / zo;
						zp = zt;
						if (zo > 0)
						{
							zt = zt + cellSize;
							z = z + 1;
						}
						else
						{
							zt = zt - cellSize;
							z = z - 1;
						}
					}
				}
			}
		}
	}
    
	// No collision

	return result;
}

///////////////////////////////////////////////////////////////////
// Special rendering

void CTerrain::RenderProjectedTexture(void *p_render, float p_x, float p_z, float p_radius, float p_rotation, CAPITexture *p_texture, CColorRGB *p_modColor, BYTE p_alpha)
{
	// Render a texture projected straight down on the terrain
	// Center at p_x, p_z, of size p_radius.
	// p_rotation is in degrees and spins on a vertical axis

	// This routine calculates a surface along with texture axes,
	//   then figures out which triangles on the terrain touch the sized,
	//   rotated quad.  It then assembles a surface out of each triangle
	//   and submits each to RenderSurfaces

	// We'll need to z bias away from the terrain surface too
	//  ?? er... if decals are already there, we'll need to raise it
	//     even more...  so we'll need numerical equivalents.  1 and 2?

	// ??
	IRender *render = (IRender *)p_render;

	int xMin, xMax;
	int zMin, zMax;

	// Based on radius, figure out which cells we are going to draw
	float actualRadius = float(p_radius * sqrt(2.0f));
	// ?? This radius would be p_radius for angle = 0, 90, 180, 270 and max for 45, 135, 225 and 315
	//   always multiplying by 1.414 ultimately slows things down

	xMin = int((p_x - actualRadius) / cellSize);
	if (xMin > int(xQty-1)) return;
	xMax = int((p_x + actualRadius) / cellSize);
	if (xMax < 0) return;
	zMin = int((p_z - actualRadius) / cellSize);
	if (zMin > int(zQty-1)) return;
	zMax = int((p_z + actualRadius) / cellSize);
	if (zMax < 0) return;

	if (xMin < 0) xMin = 0;
	if (xMax > int(xQty-2)) xMax = xQty-2;
	if (zMin < 0) zMin = 0;
	if (zMax > int(zQty-2)) zMax = zQty-2;

	static CSurface3D surface;
	//CModelVertex3D vertices[3];
	static CModelVertex3D vertices[4];
	//surface.Initialize(3, p_texture);

	static bool init=false;
	if (!init)
	{
		init = true;
		surface.Initialize(4);
	}
	surface.SetTexture(p_texture);

	render->SetZBias(true);
	render->EnableDepthWrite(false);

	CVector3D offset;
	offset.z = zMin * cellSize - p_z;
	
	CVector3D axis1, axis2;
	axis1.x = float(cos(p_rotation * 3.14159266f/180.0f) / p_radius);
	axis1.y = 0;
	axis1.z = float(sin(p_rotation * 3.14159266f/180.0f) / p_radius);
	axis2.x = -axis1.z;
	axis2.y = 0;
	axis2.z = axis1.x;
	float a1, a2;
	a1 = float(cos(-p_rotation * 3.14159266f/180.0f));
	a2 = float(sin(-p_rotation * 3.14159266f/180.0f));

	int x, z;
	float radius = p_radius * 2.0f;
	float a1x0, a1x1, a2x0, a2x1;
	float a1z0, a1z1, a2z0, a2z1;
	for (z=zMin; z<=zMax; z++)
	{
		offset.x = xMin * cellSize - p_x;

		a1z0 = a1*(z*cellSize-p_z)/radius;
		a2z0 = a2*(z*cellSize-p_z)/radius;
		a1z1 = a1*((z+1)*cellSize-p_z)/radius;
		a2z1 = a2*((z+1)*cellSize-p_z)/radius;

		for (x=xMin; x<=xMax; x++)
		{
			a1x0 = a1*(x*cellSize-p_x)/radius;
			a2x0 = a2*(x*cellSize-p_x)/radius;
			a1x1 = a1*((x+1)*cellSize-p_x)/radius;
			a2x1 = a2*((x+1)*cellSize-p_x)/radius;

			// This can most certainly be sped up be setting up constants and offsets
			// But as long as you are only rendering one projected texture each
			//   game tick, it shouldn't affect frame rate too much
			// ??

			// Sending as a single 4-vertex surface (even though it's not planar)
			//   is much faster and still works since it isn't lit

			vertices[0].vertex.SetVec((x+1)*cellSize, points[(x+1)+z*xQty].y, z*cellSize);
			vertices[1].vertex.SetVec((x+1)*cellSize, points[(x+1)+(z+1)*xQty].y, (z+1)*cellSize);
			vertices[2].vertex.SetVec(x*cellSize, points[x+(z+1)*xQty].y, (z+1)*cellSize);
			vertices[3].vertex.SetVec(x*cellSize, points[x+z*xQty].y, z*cellSize);

//			surface.SetVertexRef(0,0,0.5f-a1*((x+1)*cellSize-p_x)/radius+a2*(z*cellSize-p_z)/radius,0.5f+a2*((x+1)*cellSize-p_x)/radius+a1*(z*cellSize-p_z)/radius);
//			surface.SetVertexRef(1,1,0.5f-a1*((x+1)*cellSize-p_x)/radius+a2*((z+1)*cellSize-p_z)/radius,0.5f+a2*((x+1)*cellSize-p_x)/radius+a1*((z+1)*cellSize-p_z)/radius);
//			surface.SetVertexRef(2,2,0.5f-a1*(x*cellSize-p_x)/radius+a2*((z+1)*cellSize-p_z)/radius,0.5f+a2*(x*cellSize-p_x)/radius+a1*((z+1)*cellSize-p_z)/radius);
//			surface.SetVertexRef(3,3,0.5f-a1*(x*cellSize-p_x)/radius+a2*(z*cellSize-p_z)/radius,0.5f+a2*(x*cellSize-p_x)/radius+a1*(z*cellSize-p_z)/radius);
			surface.SetVertexRef(0,0,0.5f-a1x1+a2z0,0.5f+a2x1+a1z0);
			surface.SetVertexRef(1,1,0.5f-a1x1+a2z1,0.5f+a2x1+a1z1);
			surface.SetVertexRef(2,2,0.5f-a1x0+a2z1,0.5f+a2x0+a1z1);
			surface.SetVertexRef(3,3,0.5f-a1x0+a2z0,0.5f+a2x0+a1z0);

			render->RenderSurfaces(&surface, 1, vertices, p_modColor, p_alpha, NULL, NULL, true, NULL, NULL, NULL);

			offset.x += cellSize;
		}
	}

	render->EnableDepthWrite(true);
	render->SetZBias(false);

}

