// TerrainRegion.cpp: implementation of the CTerrainRegion class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "TerrainRegion.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

// Chunks
CTerrainChunk::CTerrainChunk()
{
	lods = NULL;
}

CTerrainChunk::~CTerrainChunk()
{
	Destroy();
}

void CTerrainChunk::Destroy(void)
{
	if (lods) delete [] lods;
	lods = NULL;
}

// Region

CTerrainRegion::CTerrainRegion()
{
	lodQty = 0;
	lodDistances = NULL;
	chunks = NULL;
	chunkXQty = 0;
	chunkZQty = 0;
}

CTerrainRegion::~CTerrainRegion()
{
	Destroy();
}

void CTerrainRegion::Destroy(void)
{
	if (lodDistances) delete [] lodDistances;
	if (chunks) delete [] chunks;

	lodDistances = NULL;
	lodQty = 0;
	chunks = NULL;
	chunkXQty = 0;
	chunkZQty = 0;
}

void CTerrainRegion::LoadFromBMP(char *p_filename, unsigned int p_initialDimensionP2, float p_cellSize, float p_minY, float p_maxY, CAPITexture *p_texture0, float p_texCoordPerCell0, CAPITexture *p_texture1, float p_texCoordPerCell1, defTerrainRenderType p_textureOp)
{
// It's assumed that p_initialDimension is an even multiple of the
// totalWidth and totalHeight of the region file.
// lodQty also must be set first

	FILE *fp;
	fp = fopen(p_filename, "rb");
	if (fp)
	{
		// Get terrain dimensions, figure out how many terrain chunks we need

		// BMP files:
		// using 0-based indices
		// 18-21 - width
		// 22-25 - height
		// 10-13 - start of pixel data
		fseek(fp, 18, SEEK_SET);
		fread(&totalWidth, 4, 1, fp);
		fseek(fp, 22, SEEK_SET);
		fread(&totalHeight, 4, 1, fp);

		long start;
		fseek(fp, 10, SEEK_SET);
		fread(&start, 4, 1, fp);

		// Prepare and load up all the terrains making up the region
		// Remember, BMP is upside down in file

		chunkXQty = totalWidth / p_initialDimensionP2;
		chunkZQty = totalHeight / p_initialDimensionP2;

		lod0Size = float(p_initialDimensionP2 * p_cellSize);

		chunks = new CTerrainChunk[chunkXQty * chunkZQty];
		unsigned int cx, cz;
		BYTE row[3072];	// Supports up to 1024 p_initialDimensionP2
		// Load up LOD 0 for each terrain chunk directly from the file, then
		//   make the lower LOD's later
		for (cx=0; cx<chunkXQty; cx++)
		for (cz=0; cz<chunkZQty; cz++)
		{
			chunks[cx + cz*chunkXQty].lods = new CTerrain[lodQty];

			unsigned int x, z;
			unsigned int adjustX;
			unsigned int adjustZ;
			if (cx < chunkXQty - 1)
				adjustX = 1;
			else
				adjustX = 0;
			if (cz < chunkZQty - 1)
				adjustZ = 1;
			else
				adjustZ = 0;
			CTerrainChunk *lpChunk = &(chunks[cx + cz*chunkXQty]);
			lpChunk->lods[0].Initialize(p_initialDimensionP2+adjustX, p_initialDimensionP2+adjustZ, p_cellSize);
			CTerrainPoint *points = lpChunk->lods[0].GetPoints();
			CTerrainPoint *lpPoint;
			for (z=0; z<p_initialDimensionP2+adjustZ; z++)
			{
				// Get a row from the file
				// Start from the bottom and work up
				fseek(fp, start + cx * p_initialDimensionP2 * 3 + (totalHeight-(cz*p_initialDimensionP2+z)-1) * totalWidth * 3, SEEK_SET);
				fread(row, (p_initialDimensionP2+adjustX) * 3, 1, fp);

				lpPoint = &(points[z*(p_initialDimensionP2+adjustX)]);
				for (x=0; x<p_initialDimensionP2+adjustX; x++)
				{
					lpPoint->y = p_minY + (float(row[x*3]) / 255.0f) * (p_maxY - p_minY);
					lpPoint->illumination = 1.0f;
					lpPoint++;
				}
			}
			lpChunk->lods[0].CalculateSurfaceData();
			lpChunk->lods[0].SetTexture0(p_texture0, p_texCoordPerCell0);
			lpChunk->lods[0].SetTexture1(p_texture1, p_texCoordPerCell1);
			lpChunk->lods[0].SetTextureOp(p_textureOp);
			lpChunk->lods[0].shade = 32;
			lpChunk->lods[0].startS0 = cx * p_initialDimensionP2 * p_texCoordPerCell0;
			lpChunk->lods[0].startT0 = cz * p_initialDimensionP2 * p_texCoordPerCell0;
			lpChunk->lods[0].startS1 = cx * p_initialDimensionP2 * p_texCoordPerCell1;
			lpChunk->lods[0].startT1 = cz * p_initialDimensionP2 * p_texCoordPerCell1;
		}

		fclose(fp);

		// NOTE: point normals at the edges are incorrect at this point
		//   because they don't incorporate adjacent chunk normals!!!
		// ??

		// Prepare the LOD's based on how many LOD's we were told to have
		// Halve each LOD's dimension and pass down the y's at the edges
		//   to the midpoints at the edges of the lower LOD's
		unsigned int i;
		float totalY;
		long totalAlpha;
		unsigned int x, z;
		unsigned int adjustX, adjustZ;
		CTerrainChunk *lpChunk, *lpChunk1;
		unsigned int xt, zt;
		for (i=1; i<lodQty; i++)
		{
			for (cz=0; cz<chunkZQty; cz++)
			{
				lpChunk = &chunks[0+cz*chunkXQty];
				if (cz < chunkZQty-1)
					adjustZ = 1;
				else
					adjustZ = 0;

				for (cx=0; cx<chunkXQty; cx++)
				{
					if (cx < chunkXQty - 1)
						adjustX = 1;
					else
						adjustX = 0;
					lpChunk->lods[i].Initialize(lpChunk->lods[i-1].xQty/2 + adjustX, lpChunk->lods[i-1].zQty/2 + adjustZ, lpChunk->lods[i-1].cellSize * 2.0f);

					for (x=0; x<lpChunk->lods[i].xQty; x++)
					for (z=0; z<lpChunk->lods[i].zQty; z++)
					{
						lpChunk1 = NULL;

						if ((adjustX == 0) && (adjustZ == 0))
						{
							lpChunk1 = lpChunk;
							xt=x; zt=z;
						}
						else
						{
							if ((x<lpChunk->lods[i].xQty-1) || (adjustX == 0))
							{
								if ((z<lpChunk->lods[i].zQty-1) || (adjustZ == 0))
								{
									lpChunk1 = lpChunk;
									xt=x; zt=z;
								}
								else
								{
									lpChunk1 = lpChunk+chunkXQty;
									xt=x; zt=0;
								}
							}
							else
							{
								if ((z<lpChunk->lods[i].zQty-1) || (adjustZ == 0))
								{
									lpChunk1 = lpChunk+1;
									xt=0; zt=z;
								}
								else
								{
									lpChunk1 = lpChunk+chunkXQty+1;
									xt=0; zt=0;
								}
							}
						}
						totalY = 0.0f;
						totalAlpha = 0;
						if (lpChunk1)
						{
							totalY += lpChunk1->lods[i-1].GetY(xt*2, zt*2);
							totalY += lpChunk1->lods[i-1].GetY(xt*2+1, zt*2);
							totalY += lpChunk1->lods[i-1].GetY(xt*2, zt*2+1);
							totalY += lpChunk1->lods[i-1].GetY(xt*2+1, zt*2+1);

							totalAlpha += lpChunk->lods[i-1].GetAlpha1(xt*2, zt*2);
							totalAlpha += lpChunk->lods[i-1].GetAlpha1(xt*2+1, zt*2);
							totalAlpha += lpChunk->lods[i-1].GetAlpha1(xt*2, zt*2+1);
							totalAlpha += lpChunk->lods[i-1].GetAlpha1(xt*2+1, zt*2+1);
						}
						lpChunk->lods[i].SetY(x, z, totalY / 4.0f);
						lpChunk->lods[i].SetAlpha1(x, z, totalAlpha/4);
						lpChunk->lods[i].SetIllumination(x, z, 1.0f);
					}

					lpChunk->lods[i].CalculateSurfaceData(false);
					lpChunk->lods[i].SetTexture0(p_texture0, p_texCoordPerCell0 * lpChunk->lods[i-1].cellSize/lpChunk->lods[0].cellSize*2.0f);
					lpChunk->lods[i].SetTexture1(p_texture1, p_texCoordPerCell1 * lpChunk->lods[i-1].cellSize/lpChunk->lods[0].cellSize*2.0f);
					lpChunk->lods[i].SetTextureOp(p_textureOp);
					lpChunk->lods[i].shade = 32;
					lpChunk->lods[i].startS0 = lpChunk->lods[i-1].startS0;
					lpChunk->lods[i].startT0 = lpChunk->lods[i-1].startT0;
					lpChunk->lods[i].startS1 = lpChunk->lods[i-1].startS1;
					lpChunk->lods[i].startT1 = lpChunk->lods[i-1].startT1;

					lpChunk++;
				}
			}
		}	// for i
	}
	else
	{
		TRACE("* Debug: Unable to open terrain region file %s\n", p_filename);
	}
}

void CTerrainRegion::Render(IRender *p_render, COrient3D &p_viewpoint)
{
	// for now, render them all at LOD 0

	unsigned int cx, cz;


	// Prepare junks for rendering
	// Figure out which LOD's we are using
	for (cx=0; cx<chunkXQty; cx++)
	for (cz=0; cz<chunkZQty; cz++)
	{
		//  Figure out if this chunk is visible to the viewpoint
		CVector3D v1 (lod0Size * float(cx), 0.0f, lod0Size * float(cz));
		CVector3D vx (lod0Size, 0.0f, 0.0f);
		CVector3D vz (0.0f, 0.0f, lod0Size);
		bool visible = false;
		visible = visible || ((v1-p_viewpoint.p) * p_viewpoint.f > 0.0f);
		visible = visible || (((v1+vx)-p_viewpoint.p) * p_viewpoint.f > 0.0f);
		visible = visible || (((v1+vz)-p_viewpoint.p) * p_viewpoint.f > 0.0f);
		visible = visible || (((v1+vx+vz)-p_viewpoint.p) * p_viewpoint.f > 0.0f);

		if (visible)
		{
			// Figure out which LOD to render
			v1 = v1 + (vx * 0.5f) + (vz * 0.5f) - p_viewpoint.p;
			float distance = v1.MgnTud();

			unsigned int i;
			chunks[cx+cz*chunkXQty].renderLod = lodQty-1;
			for (i=0; i<lodQty-1; i++)
			{
				if (distance <= lodDistances[i])
				{
					chunks[cx+cz*chunkXQty].renderLod = i;
					i = lodQty;
				}
			}
		}
		else
			chunks[cx+cz*chunkXQty].renderLod = -1;

	}

	for (cx=0; cx<chunkXQty; cx++)
	for (cz=0; cz<chunkZQty; cz++)
	{
		if (chunks[cx+cz*chunkXQty].renderLod != -1)
		{
			// Decide if we can do a straight render or a special LOD border
			//   join
			bool loX, loZ, hiX, hiZ, borderRender;
			loX = false;
			loZ = false;
			hiX = false;
			hiZ = false;
			borderRender = false;

			if (chunks[cx+cz*chunkXQty].renderLod > 0)
			{
				if (cx>0)
				if (chunks[cx-1 + cz*chunkXQty].renderLod < chunks[cx+cz*chunkXQty].renderLod)
				{
					borderRender = true;
					loX = true;
				}

				if (cx<chunkXQty-1)
				if (chunks[cx+1 + cz*chunkXQty].renderLod < chunks[cx+cz*chunkXQty].renderLod)
				{
					borderRender = true;
					hiX = true;
				}

				if (cz>0)
				if (chunks[cx + (cz-1)*chunkXQty].renderLod < chunks[cx+cz*chunkXQty].renderLod)
				{
					borderRender = true;
					loZ = true;
				}

				if (cz<chunkZQty-1)
				if (chunks[cx + (cz+1)*chunkXQty].renderLod < chunks[cx+cz*chunkXQty].renderLod)
				{
					borderRender = true;
					hiZ = true;
				}
			}

			p_render->PushMatrix();
			p_render->Translatef(lod0Size * float(cx), 0.0f, lod0Size * float(cz));
			if (borderRender)
				p_render->RenderTerrain(chunks[cx+cz*chunkXQty].lods[chunks[cx+cz*chunkXQty].renderLod], false, true,
						&chunks[cx+cz*chunkXQty].lods[chunks[cx+cz*chunkXQty].renderLod-1], loX, loZ, hiX, hiZ);
			else
				p_render->RenderTerrain(chunks[cx+cz*chunkXQty].lods[chunks[cx+cz*chunkXQty].renderLod], false, true);
			p_render->PopMatrix();
		}
	}
}