// TerrainRegion.h: interface for the CTerrainRegion class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_TERRAINREGION_H__58F4173F_AEF7_4773_9A4C_512BD23E5D40__INCLUDED_)
#define AFX_TERRAINREGION_H__58F4173F_AEF7_4773_9A4C_512BD23E5D40__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "terrain.h"
#include "iRender.h"

// A single chunk of a larger terrain, including the deeper LOD's
class CTerrainChunk
{
public:
	CTerrainChunk();
	virtual ~CTerrainChunk();
	void Destroy(void);

	CTerrain *lods;		// array of chunks, element 0 is lod 0, 1 lod 1, etc.
						// same qty as parent CTerrainRegion lodQty

	int renderLod;		// Which LOD is this chunk going to be rendered as?
						// This value is prepared before the actual rendering
						//   and is used to accurately join LOD chunks together
						//   and assist with projected textures
						// If -1, this chunk isn't rendered this tick
};

// This class encompasses the entire set of visible terrain at any given time
// Calling the render class causes an entire surrounding world render, minus
//   specific objects on the terrain
class CTerrainRegion  
{
public:
	CTerrainRegion();
	virtual ~CTerrainRegion();
	void Destroy(void);

private:
	float *lodDistances;	// Distances at which to use LODs
	unsigned int lodQty;	// How many LOD's do we have?

	float lod0Size;			// World dimension of a lod 0 chunk

	CTerrainChunk	*chunks;	// populated by Load routine

	unsigned int totalWidth;	// Dimensions of region in cells (pixels
	unsigned int totalHeight;	//   of heightmap file)

	// Number of terrain chunks in regions in x and z direction
	unsigned int chunkXQty;
	unsigned int chunkZQty;

public:
	void SetLODQty(unsigned int p_lodQty)
	{
		lodQty = p_lodQty;
		if (lodQty>1)
			lodDistances = new float[lodQty-1];
	}
	void SetLODDistance(unsigned int p_id, float p_distance)
	{
		lodDistances[p_id] = p_distance;
	}

	// Make sure you set LODqty before loading
	// Each LOD will always be half of the next size up
	// NOTE: this is a full region file load, a better version will be
	//   needed for multiple large files that make up an entire
	//   world
	void LoadFromBMP(char *p_filename, unsigned int p_initialDimensionP2, float p_cellSize, float p_minY, float p_maxY, CAPITexture *p_texture0, float p_texCoordPerCell0, CAPITexture *p_texture1, float p_texCoordPerCell1, defTerrainRenderType p_textureOp);

	// This function is used to render it, it decides which LODs to
	//   use and intersects them and occludes according to view frustum
	//   and terrain
	void Render(IRender *p_render, COrient3D &p_viewpoint);
};

#endif // !defined(AFX_TERRAINREGION_H__58F4173F_AEF7_4773_9A4C_512BD23E5D40__INCLUDED_)
