#include "stdafx.h"
#include "globals.h"
#include "api3d\ApiTextureRegistry.h"
#include "api3d\IRender.h"
#include "api3d\OpenGLRender.h"

IRender				*Render;	
CAPITextureRegistry	APITextures;

OpenGLRender		OpenGL;

CMetrics			Metrics;

bool SphereCollision(CSurface3D *p_surfaces, unsigned p_surfaceQty, CModelVertex3D *p_vertices, unsigned p_vertexQty, CVector3D &p_pos, CVector3D &p_offset, float p_radius, float &p_timeOut, CVector3D &p_normalOut, CSurface3D **p_collisionSurfaceOut, CLinkedList<CInteger> *p_vertexList)
// This function assumes that the model has had all its
//    surface normals calculated, and all the surfaces are convex!
// ?? It does NOT account for a segment that is not part of
//    a surface (not sure what would happen in the definition,
//    it might actually work, I have to confirm)
// It will account for orphaned points (points in space will cause a collision)
// Careful with scaled models, distance may not evaluate
//   correctly and hence cause comparisons to be inaccurate
// p_timeOut is the offset (0.0-1.0) along the sphere's path below which
//    the collision must occur to be valid.  It's usually initialized
//    to 100, and wil be set by the routine if a collision occurs
{
	bool result = false;

	CVector3D normal;

	unsigned i;

	// Check surfaces and segments
	for (i=0; i<p_surfaceQty; i++)
	{
		if (p_surfaces[i].IsVisible())
			if (p_surfaces[i].SphereCollision(p_pos, p_offset, p_radius, p_timeOut, p_vertices, normal))
			{
				result = true;
				p_normalOut = normal;
				*p_collisionSurfaceOut = &p_surfaces[i];
			}
	}

	// If we have a collision, leave
	if (result)
		return result;

	// Check points
	if (p_vertexList == NULL)
	{
		for (i=0; i<p_vertexQty; i++)
		{
			if (p_vertices[i].vertex.SphereCollision(p_pos, p_offset, p_radius, p_timeOut, normal))
			{
				result = true;
				p_normalOut = normal;
				//Debug.Print "Point collision"
				*p_collisionSurfaceOut = NULL;
			}
		}
	}
	else
	{
		CLinkedListNode<CInteger> *node = p_vertexList->GetFirst();
		if (node)
		while (node->next != NULL)
		{
			if (p_vertices[node->value].vertex.SphereCollision(p_pos, p_offset, p_radius, p_timeOut, p_normalOut))
			{
				result = true;
				*p_collisionSurfaceOut = NULL;
			}
			node = node->next;
		}
	}

	return result;
}

bool RayCollision(CSurface3D *p_surfaces, unsigned p_surfaceQty, CModelVertex3D *p_vertices, CVector3D &p_position, CVector3D &p_directionUnit, float &p_collisionDistance, CVector3D &p_collisionPointOut, CSurface3D **p_collisionSurfaceOut)
{
// p_position and p_direction are in object space
// so expect to do transforms on them before submitting them to
//    this routine
// p_direction must be a unit vector (we could unitize it here if need be)
// Careful with scaled models, distance may not evaluate
//   correctly and hence cause comparisons to be inaccurate

// This only works if the model has its normals calculated

    unsigned i;
    CVector3D collisionPoint;
	bool result = false;
    
    // Check surfaces and segments
    for (i=0; i<p_surfaceQty; i++)
	{
        if (p_surfaces[i].IsVisible())
            if (p_surfaces[i].RayCollision(p_position, p_directionUnit, p_collisionDistance, p_vertices, p_collisionPointOut))
			{
                result = true;
                *p_collisionSurfaceOut = &p_surfaces[i];
			}
    }

	return result;
}

// Returns a float between [0.0-1.0)
float Random(void)
{
	return (float(rand()) / float(RAND_MAX));
}

