// DInput.cpp: implementation of the CDInput class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "DInput.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

LPDIRECTINPUTDEVICE8 CDInput::diJoystick8;
LPDIRECTINPUT8       CDInput::di8;
      
CDInput::CDInput()
{
	di8 = NULL;
	diJoystick8 = NULL;
	valid = true;
}

CDInput::~CDInput()
{
	Destroy();
}

void CDInput::Destroy(void)
{
	DestroyJoystick();

	if (di8)
	{
		di8->Release();
	}

	di8 = NULL;
}

void CDInput::PollJoystick(void)
{
	if (!diJoystick8) return;

	HRESULT hr;

	hr = diJoystick8->Poll();
	// This code is from the dx9 sample joystick.cpp
	if (FAILED(hr))
	{
        // DInput is telling us that the input stream has been
        // interrupted. We aren't tracking any state between polls, so
        // we don't have any special reset that needs to be done. We
        // just re-acquire and try again.
        hr = diJoystick8->Acquire();
        while( hr == DIERR_INPUTLOST ) 
            hr = diJoystick8->Acquire();

        // hr may be DIERR_OTHERAPPHASPRIO or other errors.  This
        // may occur when the app is minimized or in the process of 
        // switching, so just try again later 
        return; 
	}

    // Get the input's device state
    if( FAILED( hr = diJoystick8->GetDeviceState( sizeof(DIJOYSTATE2), &joystickState ) ) )
        return; // The device should have been acquired during the Poll()
}

// Call PollJoystick before calling this to get updated values
void CDInput::GetJoystickState(strJoystick &p_jsValues)
{
	p_jsValues.axis1 = joystickState.lX;
	p_jsValues.axis2 = joystickState.lY;

	p_jsValues.button1 = ((joystickState.rgbButtons[0] & 0x80) != 0);
	p_jsValues.button2 = ((joystickState.rgbButtons[1] & 0x80) != 0);
}

void CDInput::DestroyJoystick(void)
{
	if (diJoystick8)
	{
		diJoystick8->Unacquire();
		diJoystick8->Release();
	}

	diJoystick8 = NULL;
}

bool CDInput::InitializeDInput(void)
{
	if (!valid) return false;
	if (di8) return true;

	HRESULT hr;

    if( FAILED( hr = DirectInput8Create( GetModuleHandle(NULL), DIRECTINPUT_VERSION, 
                                         IID_IDirectInput8, (VOID**)&di8, NULL ) ) )
	{
		valid = false;
        return false;
	}

	return true;
}

// The p_hWnd you send for this should be the main app hWnd, not the control
//   hWnd you send for D3D or OpenGL (which could be the same one)
bool CDInput::InitializeJoystick(HWND p_hWnd)
{
	if (!di8) InitializeDInput();

	if (!valid) return false;
	if (diJoystick8) return true;

	// This code is fromt he dx9 sample joystick.cpp

	HRESULT hr;
    // Look for a simple joystick we can use for this sample program.
    if( FAILED( hr = di8->EnumDevices( DI8DEVCLASS_GAMECTRL, 
                                         EnumJoysticksCallback,
                                         NULL, DIEDFL_ATTACHEDONLY ) ) )
        return false;

    // Make sure we got a joystick
    if( NULL == diJoystick8 )
    {
		// No joystick found
        return false;
    }

    // Set the data format to "simple joystick" - a predefined data format 
    //
    // A data format specifies which controls on a device we are interested in,
    // and how they should be reported. This tells DInput that we will be
    // passing a DIJOYSTATE2 structure to IDirectInputDevice::GetDeviceState().
    if( FAILED( hr = diJoystick8->SetDataFormat( &c_dfDIJoystick2 ) ) )
	{
		DestroyJoystick();
        return false;
	}

    // Set the cooperative level to let DInput know how this device should
    // interact with the system and with other DInput applications.
    if( FAILED( hr = diJoystick8->SetCooperativeLevel( p_hWnd, DISCL_EXCLUSIVE | 
                                                             DISCL_FOREGROUND ) ) )
	{
		DestroyJoystick();
        return false;
	}

    // Enumerate the joystick objects. The callback function enabled user
    // interface elements for objects that are found, and sets the min/max
    // values property for discovered axes.
    if( FAILED( hr = diJoystick8->EnumObjects( EnumJSObjectsCallback, 
                                                (VOID*)p_hWnd, DIDFT_ALL ) ) )
	{
		DestroyJoystick();
        return false;
	}

	// Force acquire
    hr = diJoystick8->Acquire();
    while( hr == DIERR_INPUTLOST ) 
        hr = diJoystick8->Acquire();

    return true;
}

BOOL CALLBACK CDInput::EnumJSObjectsCallback( const DIDEVICEOBJECTINSTANCE* pdidoi,
                                   VOID* pContext )
{
    HWND hWnd = (HWND)pContext;

    static int nSliderCount = 0;  // Number of returned slider controls
    static int nPOVCount = 0;     // Number of returned POV controls

    // For axes that are returned, set the DIPROP_RANGE property for the
    // enumerated axis in order to scale min/max values.
	// Set possible range to -1000/1000
    if( pdidoi->dwType & DIDFT_AXIS )
    {
        DIPROPRANGE diprg; 
        diprg.diph.dwSize       = sizeof(DIPROPRANGE); 
        diprg.diph.dwHeaderSize = sizeof(DIPROPHEADER); 
        diprg.diph.dwHow        = DIPH_BYID; 
        diprg.diph.dwObj        = pdidoi->dwType; // Specify the enumerated axis
        diprg.lMin              = -1000; 
        diprg.lMax              = +1000; 
    
        // Set the range for the axis
        if( FAILED( diJoystick8->SetProperty( DIPROP_RANGE, &diprg.diph ) ) ) 
            return DIENUM_STOP;    
    }

	// Many other possibilities here, see dx9 sample joystick.cpp

    return DIENUM_CONTINUE;
}

BOOL CALLBACK CDInput::EnumJoysticksCallback( const DIDEVICEINSTANCE* pdidInstance,
                                     VOID* pContext )
{
    HRESULT hr;

	// Code from dx9 sample joystick.cpp

    // Obtain an interface to the enumerated joystick.
    hr = di8->CreateDevice( pdidInstance->guidInstance, &diJoystick8, NULL );

    // If it failed, then we can't use this joystick. (Maybe the user unplugged
    // it while we were in the middle of enumerating it.)
    if( FAILED(hr) ) 
        return DIENUM_CONTINUE;

    // Stop enumeration. Note: we're just taking the first joystick we get. You
    // could store all the enumerated joysticks and let the user pick.
    return DIENUM_STOP;
}
