// PtlPortal.h: interface for the CPtlPortal class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_PTLPORTAL_H__CD63B8C8_18D0_42CE_8320_6C01B359B782__INCLUDED_)
#define AFX_PTLPORTAL_H__CD63B8C8_18D0_42CE_8320_6C01B359B782__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "..\math\Vector3D.h"
#include "..\api3d\Surface3D.h"
#include "..\api3d\ModelVertex3D.h"
#include "PtlFrustum.h"
#include "PtlDoor.h"
#include "..\api3d\IRender.h"

class CPtlPortal  
{
public:
	CPtlPortal();
	virtual ~CPtlPortal();
	void Destroy();

private:
	// Vertices for the portal, references only
	unsigned vertexRefQty;
	int *vertexRefs;

	// Rooms on each side of the portal
	int frontRoomID;		// Usually always set
	int backRoomID;			// Usually -1 for mirrors (mirror have no back
							//    side, not yet anyway)

	CVector3D normal;		// Normal perpendicular to portal (faces frontRoom,
							//   assuming clockwise definition)

	bool closed;	// Can we see through the portal? (overrides all other conditions)

	bool mirror;	// Is it a mirror?

	CSurface3D	*mirrorSurface;	// Which surface serves as the mirror?
								//  (for rendering)

	// Door list (doors will only be rendered and used in collisions if
	//   they are listed in the octree)
	CPtlDoorList doorList;

public:
	void Initialize(unsigned p_vertexRefQty);
	void SetVertexRef(unsigned p_vertexRefID, int p_vertexID) {vertexRefs[p_vertexRefID] = p_vertexID;}
	void SetAdjacentRoomIDs(int p_frontRoomID, int p_backRoomID)
	{
		frontRoomID = p_frontRoomID;
		backRoomID = p_backRoomID;
	}
	inline void SetAdjacentRoomIDs(int p_frontRoomID)
	{
		SetAdjacentRoomIDs(p_frontRoomID, -1);
	}
	void SetMirror(bool p_mirror) { mirror = p_mirror;}
	void SetClosed(bool p_closed) { closed = p_closed; }
	void SetMirrorSurface(CSurface3D *p_surface) {mirrorSurface = p_surface;}

	void CalculateData(CModelVertex3D *p_vertices);

	bool IsMirror(void) { return mirror; }
	bool IsClosed(void);
	int GetFrontRoomID(void) {return frontRoomID;}
	int GetBackRoomID(void) {return backRoomID;}
	CVector3D & GetNormal(void) {return normal;}
	CVector3D & GetVertex(unsigned p_id, CModelVertex3D *p_vertices) {return p_vertices[vertexRefs[p_id]].vertex;}

	bool PointInFront(CVector3D &p_point, CModelVertex3D *p_vertices)
	{
		return (((p_point - p_vertices[vertexRefs[0]].vertex) * normal) > 0);
	}
	bool MergeFrustums(CPtlFrustum &p_oldFrustum, CPtlFrustum &p_newFrustum, CModelVertex3D *p_vertices, int p_currentRoomID);
	bool IsVisible(CPtlFrustum &p_frustum, CModelVertex3D *p_vertices, int p_roomID);

	//void Render(IRender *p_render, CModelVertex3D *p_vertices);
	void RenderSpliced(IRender *p_render, CModelVertex3D *p_vertices, CPtlFrustum &p_frustum, CVector3D *p_verticesOut, unsigned &p_vertexOutQty);
	void RenderVertexArray(IRender *p_render, CVector3D *p_vertexArray, unsigned p_vertexQty, bool p_renderSurface, CAPITexture **p_lightmaps, CLightSourceRefList *p_lightSourceList);

	// Door routines
	void AnimateDoors(long p_elapsedTimeMS);
	CPtlDoorListNode * CPtlPortal::AddDoor(void);
	void RenderDoors(IRender *p_render, CLightSourceList *lightSourceList, unsigned int p_renderID);
	bool RayCollisionWithDoor(CVector3D &p_start, CVector3D &p_directionUnit, float &p_distanceOut, CSurface3D **p_collisionSurfaceOut, CVector3D &p_collisionPointOut, CPtlDoor **p_collisionDoorOut);
	bool SphereCollisionWithDoors(CVector3D &p_pos, CVector3D &p_offset, float p_radius, float &p_timeOut, CVector3D &p_collisionNormalOut, CSurface3D **p_collisionSurfaceOut, CPtlDoor **p_collisionDoorOut);

private:
	bool SplicePolygonWithPlane(CVector3D *p_vertices, unsigned p_vertexQty, CVector3D *p_verticesOut, unsigned &p_vertexQtyOut, CVector3D &p_planePoint, CVector3D &p_planeNormal);
};

#endif // !defined(AFX_PTLPORTAL_H__CD63B8C8_18D0_42CE_8320_6C01B359B782__INCLUDED_)
