// PtlRoom.cpp: implementation of the CPtlRoom class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "PtlRoom.h"

#include "PtlMap.h"	// for access to CPtlMap::renderID

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CLinkedList<CPtlDoor> CLinkedList<CPtlDoor>::DeletedList;
unsigned int CPtlRoom::currentRenderID = 0;

void CPtlDoorList::DestroyDoors()
{
	CLinkedListNode<CPtlDoor> *node = GetFirst();
	if (node)
	while (node->next != NULL)
	{
		node->Destroy();

		node = node->next;
	}
}

CPtlRoom::CPtlRoom()
{
	surfaceQty = 0;
	surfaces = NULL;

	lightmaps = NULL;
	lightmapQty = 0;

	decalLists = NULL;

	adjacentPortalQty = 0;
	adjacentPortalIDs = NULL;

	opposingRoomIDs = NULL;

	renderID = 0;

	particleSet.Initialize(1000);
}

CPtlRoom::~CPtlRoom()
{
	Destroy();
}

void CPtlRoom::Destroy(void)
{
	collisionVertexList.EmptyList();

	// Get rid of decal lists and lightmaps
	if (decalLists)
	{
		delete [] decalLists;
		decalLists = NULL;
	}

	if (lightmaps)
	{
		unsigned i;
		for (i=0; i<lightmapQty; i++)
		{
			APITextures.DeleteTexture(lightmaps[i]);
		}

		delete [] lightmaps;
		lightmaps = NULL;

		lightmapQty = 0;
	}

	if (adjacentPortalIDs)
		delete [] adjacentPortalIDs;
	adjacentPortalIDs = NULL;
	adjacentPortalQty = 0;

	if (surfaces)
		delete [] surfaces;
	surfaces = NULL;
	surfaceQty = 0;

	if (opposingRoomIDs)
		delete [] opposingRoomIDs;
	opposingRoomIDs = NULL;

	collisionVertexList.EmptyList();

	particleEmitters.EmptyList();
	particleSet.Destroy();
}

void CPtlRoom::Initialize(unsigned p_surfaceQty, unsigned p_adjacentPortalQty, unsigned p_lightmapQty, bool p_setSurfacesWhite)
{   
    unsigned i;
    
    surfaceQty = p_surfaceQty;
	surfaces = new CSurface3D[surfaceQty];
    for (i=0; i<surfaceQty; i++)
	{
        surfaces[i].SetDecalListID(i);
        surfaces[i].SetID(i);
        
        if (p_setSurfacesWhite)
			surfaces[i].SetColor(255, 255, 255);
	}
    decalLists = new CSurfaceDecalList[surfaceQty];
    opposingRoomIDs = new int[surfaceQty];
    for (i=0; i<surfaceQty; i++)
	{
        opposingRoomIDs[i] = -1;
    }
    
    adjacentPortalQty = p_adjacentPortalQty;
    if (adjacentPortalQty > 0)
        adjacentPortalIDs = new unsigned[adjacentPortalQty];
        
	lightmapQty = p_lightmapQty;
    if (lightmapQty > 0)
	{
		lightmaps = new LPAPITEXTURE[lightmapQty];
        for (i=0; i<lightmapQty; i++)
            lightmaps[i] = APITextures.CreateNewTexture();
	}    
}

void CPtlRoom::CalculateSurfaceData(CModelVertex3D *p_vertices, COctree &p_octree, float p_targetVolume, float p_minDimension)
{
	// Initialize bounds
	minPoint.SetVec(100000000, 100000000, 100000000);
	maxPoint.SetVec(-100000000, -100000000, -100000000);

    unsigned i;
    for (i=0; i<surfaceQty; i++)
	{
        surfaces[i].CalculateSurfaceData(p_vertices);

		// Why not all?
        if ((surfaces[i].IsMirror()) && (surfaces[i].GetTexture()))
            surfaces[i].CalculateTex1Axes(p_vertices);
    }
    
    // Assemble vertex list for collisions
	// Stores vertex references
    unsigned j;
	collisionVertexList.EmptyList();
    for (i=0; i<surfaceQty; i++)
	{
        // Only work with collidable surfaces
        // ?? NOTE: if a surface is made invisible later, its
        //   vertices will still be collided with!
        if (surfaces[i].IsVisible())
		{
            for (j=0; j<surfaces[i].GetVertexRefQty(); j++)
			{
				// Look for vertex ID in list.  If it's not there, add it
				CLinkedListNode<CInteger> *node = collisionVertexList.GetFirst();
				bool found = false;
				if (node)
				while (node->next != NULL)
				{
					if (node->value == surfaces[i].GetVertexRef(j))
						found = true;

					node = node->next;
				}

				if (!found)
				{
					node = collisionVertexList.NewNode();
					node->value = surfaces[i].GetVertexRef(j);

					// Also, use this point to help determine the min and max
					//   points for the room
					if (p_vertices[node->value].vertex.x < minPoint.x)
						minPoint.x = p_vertices[node->value].vertex.x;
					if (p_vertices[node->value].vertex.y < minPoint.y)
						minPoint.y = p_vertices[node->value].vertex.y;
					if (p_vertices[node->value].vertex.z < minPoint.z)
						minPoint.z = p_vertices[node->value].vertex.z;
					if (p_vertices[node->value].vertex.x > maxPoint.x)
						maxPoint.x = p_vertices[node->value].vertex.x;
					if (p_vertices[node->value].vertex.y > maxPoint.y)
						maxPoint.y = p_vertices[node->value].vertex.y;
					if (p_vertices[node->value].vertex.z > maxPoint.z)
						maxPoint.z = p_vertices[node->value].vertex.z;
				}
            }
        }
	} // for i

	// Submit bounding volume to octree for splitting
	p_octree.SplitToVolume(minPoint, maxPoint, p_targetVolume, p_minDimension);
}

bool CPtlRoom::PointInRoom(CVector3D &p_point, CModelVertex3D *p_vertices)
{
    unsigned i;
    
    // This will work even for room walls that are only segments,
    //   as long as all coords are x,y and are Ortho Window coordinates
    for (i=0; i<surfaceQty; i++)
        if (!surfaces[i].PointInFront(p_point, p_vertices))
            return false;
    
    // Couldn't find a failure
    return true;
}

void CPtlRoom::Render(CModelVertex3D *p_vertices, IRender *p_render, DWORD p_elapsedTimeMS, DWORD p_currentTime, unsigned int p_mirrorDepth, unsigned int p_renderID, CLightSourceRefList *p_lightSourceList, CColorRGB *p_ambientColor)
{
    // As a further optimization, the engine could calculate only
    //   those portions of the geometry that are not clipped by the frustum
    //  that got us here
    // At the very least, we could just not draw any polygons
    //   that are entirely outside any one of the the clipping planes
    // Or draw polygons we know to be partially visible (just like a portal
    //    surface)
    // ??

	// Render the room surfaces
    p_render->RenderSurfaces(surfaces, surfaceQty, p_vertices, NULL, 255, lightmaps, decalLists, true, p_lightSourceList);

	CColorRGB renderColor;

	/*
	// Render the objects in the room
	CLinkedListNode<CRenderObject> *rnode;
	rnode = renderObjects.GetFirst();
	if (rnode)
	while (rnode->next != NULL)
	{
		if (rnode->renderID != p_renderID)
		{
			rnode->renderID = p_renderID;

			if (p_ambientColor)
			{
				if (p_lightSourceList)
					renderColor.SetColor(p_lightSourceList->CalculateRenderColor(rnode->orient.p, *p_ambientColor));
				else
					renderColor.SetColor(*p_ambientColor);
			}
			else
				renderColor.SetColor(255,255,255);

			p_render->PushMatrix();
			p_render->TransformWorld(rnode->orient);
			p_render->Scalef(rnode->scalef.x, rnode->scalef.y, rnode->scalef.z);
			p_render->RenderModelColor(rnode->modelRef, &renderColor);
			p_render->PopMatrix();
		}

		rnode = rnode->next;
	}
	*/

	// Render the objects touching the same cells the room touches
	CLinkedListNode<COctCellRef> *cnode1;
	CLinkedListNode<CLPVoid> *vnode1;
	CRenderObject *rnode;
	CPtlDoor *dnode;

	CLightSourceRefList objectLights;

	cnode1 = cellsTouched.GetFirst();
	if (cnode1)
	while (cnode1->next != NULL)
	{
		vnode1 = cnode1->cellRef->lpVoidList.GetFirst();	// <-- object list
		if (vnode1)
		while (vnode1->next != NULL)
		{
			rnode = (CRenderObject *)(vnode1->lpVoid);

			if (rnode->renderID != p_renderID)
			if ((p_mirrorDepth > 0) || (!rnode->controlObject))
			{
				rnode->renderID = p_renderID;

				// Assemble a light source
				//   list for that specific object.
				//   since the lights affecting a large object may
				//   expand beyond the lights illuminating a room
				if (p_ambientColor)
				{
					objectLights.EmptyList();
					CLightSource::currentRenderID++;
					CLinkedListNode<COctCellRef> *cnode2;
					CLinkedListNode<CLPVoid> *vnode2;
					cnode2 = rnode->cellsTouched.GetFirst();
					if (cnode2)
					while (cnode2->next != NULL)
					{
						vnode2 = cnode2->cellRef->lpVoidList2.GetFirst();
						if (vnode2)
						while (vnode2->next != NULL)
						{
							if (((CLightSource *)(vnode2->lpVoid))->renderID != CLightSource::currentRenderID)
							{
								((CLightSource *)(vnode2->lpVoid))->renderID = CLightSource::currentRenderID;
								objectLights.NewNode()->ref = (CLightSource *)vnode2->lpVoid;
							}

							vnode2 = vnode2->next;
						}

						cnode2 = cnode2->next;
					}

					renderColor.SetColor(objectLights.CalculateRenderColor(rnode->orient.p, *p_ambientColor));
				}
				else
					renderColor.SetColor(255,255,255);

				p_render->PushMatrix();
				p_render->TransformWorld(rnode->orient);
				p_render->Scalef(rnode->scalef.x, rnode->scalef.y, rnode->scalef.z);
				p_render->RenderModelColor(rnode->modelRef, &renderColor);
				p_render->PopMatrix();
			}

			vnode1 = vnode1->next;
		}

		// Same with the doors
		vnode1 = cnode1->cellRef->lpVoidList1.GetFirst();	// <-- door list
		if (vnode1)
		while (vnode1->next != NULL)
		{
			dnode = (CPtlDoor *)(vnode1->lpVoid);

			if (dnode->renderID != p_renderID)
			{
				dnode->renderID = p_renderID;

				dnode->Render(p_render);
			}

			vnode1 = vnode1->next;
		}

		cnode1 = cnode1->next;
	}    

	// We need to disable culling, since it could be in a mirror
	p_render->EnableDepthWrite(false);
	p_render->EnableCulling(false);
	p_render->RenderParticleSet(particleSet, p_ambientColor, CPtlMap::renderID, p_lightSourceList);
	p_render->EnableCulling(true);
	p_render->EnableDepthWrite(true);

}

void CPtlRoom::RenderFloorTexture(IRender *p_render, CVector3D &p_pos, CVector3D *p_surfaceVertices, CAPITexture *p_texture, CColorRGB *p_modColor, BYTE p_modAlpha, CModelVertex3D *p_mapVertices)
{
	bool stateSet = false;
	bool renderIt;

	static CVector3D vertOut[32];
	static unsigned vertQty;

	// p_surfaceVertices is a horizontal quad waiting to be projected
	//    downward

	// This is the surface projected downward on to the map surface,
	//   it's always 4 vertices
	static CSurface3D tmpSurface1;
	static CModelVertex3D tmpVerts1[4];

	// This is the result of the surface intersection, could be variable
	//   number of vertices
	static CSurface3D tmpSurface2;
	static CModelVertex3D tmpVerts2[32];

	static init = false;
	if (!init)
	{
		tmpSurface1.Initialize(4);
		init = true;
	}
	tmpSurface1.SetTexture(p_texture);

	// Set up vertices so we can calculate tex axes (we can't use
	//   to results of projecting down onto the surface, since it
	//   might not be flat and wouldn't provide a uniform result)
	tmpVerts1[0].vertex.x = p_surfaceVertices[0].x;
	tmpVerts1[0].vertex.y = p_surfaceVertices[0].y;
	tmpVerts1[0].vertex.z = p_surfaceVertices[0].z;
	tmpVerts1[1].vertex.x = p_surfaceVertices[1].x;
	tmpVerts1[1].vertex.y = p_surfaceVertices[1].y;
	tmpVerts1[1].vertex.z = p_surfaceVertices[1].z;
	tmpVerts1[2].vertex.x = p_surfaceVertices[2].x;
	tmpVerts1[2].vertex.y = p_surfaceVertices[2].y;
	tmpVerts1[2].vertex.z = p_surfaceVertices[2].z;
	tmpVerts1[3].vertex.x = p_surfaceVertices[3].x;
	tmpVerts1[3].vertex.y = p_surfaceVertices[3].y;
	tmpVerts1[3].vertex.z = p_surfaceVertices[3].z;

	tmpSurface1.SetVertexRef(0,0,0,1);
	tmpSurface1.SetVertexRef(1,1,1,1);
	tmpSurface1.SetVertexRef(2,2,1,0);
	tmpSurface1.SetVertexRef(3,3,0,0);
	tmpSurface1.CalculateTex1Axes(tmpVerts1);

	unsigned i;
	for (i=0; i<surfaceQty; i++)
	{
		// For any surface whose normal faces up at all, project the surface
		//    straight down on it
		if (surfaces[i].GetNormal().y > 0.0f)
		// And position must be in front of the surface for it to qualify
		if (((p_pos - p_mapVertices[surfaces[i].GetVertexRef(0)].vertex) * surfaces[i].GetNormal()) > 0)
		{
			// Now set it up so it lies on the surface
			tmpVerts1[0].vertex.y = surfaces[i].CalculateYatXZ(p_surfaceVertices[0]);
			tmpVerts1[1].vertex.y = surfaces[i].CalculateYatXZ(p_surfaceVertices[1]);
			tmpVerts1[2].vertex.y = surfaces[i].CalculateYatXZ(p_surfaceVertices[2]);
			tmpVerts1[3].vertex.y = surfaces[i].CalculateYatXZ(p_surfaceVertices[3]);
			if (p_modColor)
				tmpSurface1.SetColor(*p_modColor);
			else
				tmpSurface1.SetColor(255,255,255);
			tmpSurface1.SetAlpha(p_modAlpha);
			tmpSurface1.CalculateSurfaceData(tmpVerts1);	// we need this for an accurate SurfaceIntersection
															//   (PointInside is called on tmpSurface1 if there are
															//    no segment or point intersections)

			// Intersect it with the surface (maybe we could use stencil?)
			renderIt = surfaces[i].SurfaceIntersection(tmpSurface1, tmpVerts1, p_mapVertices, vertOut, vertQty);
			//renderIt = true;

			// Render it if we get a result
			if (renderIt)
			{
				if (!stateSet)
				{
					p_render->SetZBias(true);
					p_render->EnableDepthWrite(false);
					stateSet = true;
				}

				// ?? Build tmpSurface2
				tmpSurface2.Destroy();
				tmpSurface2.Initialize(vertQty, p_texture);
				unsigned j;
				float s, t;
				for (j=0; j<vertQty; j++)
				{
					tmpVerts2[j].vertex = vertOut[j];

					tmpSurface1.CalculateTex1Coords(vertOut[j], s, t);
					tmpSurface2.SetVertexRef(j,j, s, t);
				}
				tmpSurface2.SetAlpha(tmpSurface1.GetAlpha());
				tmpSurface2.SetColor(tmpSurface1.GetColor());

				// Rendering tmpSurface1 for testing now
				p_render->RenderSurfaces(&tmpSurface2, 1, tmpVerts2, NULL, 255, NULL, NULL, true, NULL, NULL, NULL);
			}
		}
	}

	if (stateSet)
	{
		p_render->EnableDepthWrite(true);
		p_render->SetZBias(false);
	}
}

void CPtlRoom::DetermineCellsTouched(COctree &p_octree)
{
	p_octree.CellsTouched(minPoint, maxPoint, cellsTouched);

	CLinkedListNode<COctCellRef> *cnode;
	cnode = cellsTouched.GetFirst();
	if (cnode)
	while (cnode->next != NULL)
	{
		cnode->cellRef->intList.NewNode()->value = id;

		cnode = cnode->next;
	}
}

void CPtlRoom::AnimateParticles(COctree &p_octree, CPtlRoom *p_rooms, CModelVertex3D *p_vertices, DWORD p_elapsedTimeMS, DWORD p_currentTime)
{
	COctree *cell;

	// Animate the particle emitters in the room
	// It's expected that there is no need at this point to
	//   move emitters to other rooms, 
	//   that is done before rendering using PlaceParticleEmitter
	//   as any emitter changes position
	CLinkedListNode<CParticleEmitter> *node;
	node = particleEmitters.GetFirst();
	if (node)
	{
		while (node->next != NULL)
		{
			// Animate the emitter
			node->Animate(p_elapsedTimeMS, p_currentTime);

			node = node->next;
		}
	}

	// Animate and Render the particle set in the room
	particleSet.AnimateParticles(p_elapsedTimeMS, p_currentTime);

	// Now for each particle that moved, figure out which room they are
	//   in.  This is the dangerously slow part.
	// Make sure we don't do this twice in a tick
	if (particleSet.SetTimeRoomsCalculated(p_currentTime))
	{
		CParticle *particles = particleSet.GetParticles();
		CParticle *newParticle;
		if (particles)
		{
			CParticle *lpParticle = &particles[0];
			unsigned int particleQty = particleSet.GetParticleQty();

			unsigned int i;
			for (i=0; i<particleQty; i++)
			{
				if (lpParticle->oldPosition != lpParticle->position)
				{
					lpParticle->oldPosition = lpParticle->position;

					// Get the octree cell this particle is in
					cell = p_octree.PointIn(lpParticle->position);

					if (!cell)
					{
						// Cell is nowhere on the map - remove the particle
						particleSet.RemoveParticle(i);
						i--;
						particleQty--;
						lpParticle--;
					}
					else
					{
						// Get all the rooms the cell touches
						// Find the room the particle is in
						CLinkedListNode<CInteger> *inode = cell->intList.GetFirst();
						if (inode)
						{
							// If there is only one room in the list, particle
							//   belongs there
							if (inode->next->next == NULL)
							{
								// Only reassign if it's a different room than this one
								// (id is a class member)
								if (inode->value != id)
								{
									newParticle = p_rooms[inode->value].GetParticleSet().AddParticle();
									if (newParticle)
									{
										*newParticle = *lpParticle;
										// Don't let moved particle be animated again
										//   (could result in an 'orphaned' particle)
										newParticle->animated = !(p_rooms[inode->value].GetParticleSet().AlreadyAnimated(p_currentTime));
									}

									particleSet.RemoveParticle(i);
									i--;
									particleQty--;
									lpParticle--;
								}
							}
							else
							{
								// There is more than one room in the list, figure
								//   out which room we are in
								// First make sure we have moved out of our current room
								if (!PointInRoom(lpParticle->position, p_vertices))
								{
									bool found = false;
									// See if the particle passed through a portal in the current
									//   room - if it has, assign the particle to the room on the
									//   other side
									// ??

									if (!found)
									{
										// Now figure out which room we are in the long way
										while ((inode->next != NULL) && (!found))
										{
											if (inode->value != id)
											if (p_rooms[inode->value].PointInRoom(lpParticle->position, p_vertices))
											{
												found = true;

												newParticle = p_rooms[inode->value].GetParticleSet().AddParticle();
												if (newParticle)
												{
													*newParticle = *lpParticle;
													// Don't let moved particle be animated again
													//   (could result in an 'orphaned' particle)
													newParticle->animated = !(p_rooms[inode->value].GetParticleSet().AlreadyAnimated(p_currentTime));
												}

												particleSet.RemoveParticle(i);
												i--;
												particleQty--;
												lpParticle--;
											}

											inode = inode->next;
										}
									}
								} // point not in assigned room anymore
							} // more than one room in the list
						}	// room list exists
					}	// found a cell list

				}	// if position changed

				lpParticle++;

			} // for i
		}
	}
}

