// PtlRoom.h: interface for the CPtlRoom class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_PTLROOM_H__3274D7F7_2109_4D37_994C_7FB85783B304__INCLUDED_)
#define AFX_PTLROOM_H__3274D7F7_2109_4D37_994C_7FB85783B304__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "..\structs\LinkedList.h"
//#include "..\api3d\ParticleEmitter.h"
#include "PtlDoor.h"
#include "..\api3d\renderobject.h"
#include "..\structs\octree.h"

class CPtlRoom  
{
public:
	CPtlRoom();
	virtual ~CPtlRoom();
	void Destroy(void);

	unsigned int renderID;	// Used to prevent room overuse while
							//   projecting a floor texture and collisions
	static unsigned int currentRenderID;

private:
	int id;			// id in CPtlMap's array, used in DetermineCellsTouched
	unsigned surfaceQty;
	CSurface3D *surfaces;

	LPAPITEXTURE *lightmaps;
	unsigned lightmapQty;
	CSurfaceDecalList *decalLists;

	unsigned adjacentPortalQty;
	unsigned *adjacentPortalIDs;

	int *opposingRoomIDs; // One for each surface, -1 usually

	// List of vertices to check collisions against (sphere)
	CLinkedList<CInteger> collisionVertexList;

	// Particle emitters in the room
	CParticleEmitterList particleEmitters;
	CParticleSet particleSet;

	// Bound limits on vertices for octree
	CVector3D minPoint;
	CVector3D maxPoint;

	// And cells this room is touching (based on bounding volume)
	COctCellRefList cellsTouched;

public:

	inline void SetID(int p_id) {id = p_id;}

	inline void Initialize(unsigned p_surfaceQty, unsigned p_adjacentPortalQty)
	{
		Initialize(p_surfaceQty, p_adjacentPortalQty, 0, true);
	}
	inline void Initialize(unsigned p_surfaceQty, unsigned p_adjacentPortalQty, unsigned p_lightmapQty)
	{
		Initialize(p_surfaceQty, p_adjacentPortalQty, p_lightmapQty, true);
	}
	void Initialize(unsigned p_surfaceQty, unsigned p_adjacentPortalQty, unsigned p_lightmapQty, bool p_setSurfacesWhite);

	void CalculateSurfaceData(CModelVertex3D *p_vertices, COctree &p_octree, float p_targetVolume, float p_minDimension);

	void DetermineCellsTouched(COctree &p_octree);
	inline COctCellRefList & GetCellsTouched(void)
	{
		return cellsTouched;
	}

	CSurface3D & GetSurface(unsigned p_id) {return surfaces[p_id];}
	CSurface3D * GetSurfacePtr(unsigned p_id) {return &(surfaces[p_id]);}
	unsigned GetSurfaceQty(void) {return surfaceQty;}
	CSurface3D * GetSurfaces(void) {return surfaces;}

	void SetAdjacentPortalID (unsigned p_id, unsigned p_portalID)
	{
		adjacentPortalIDs[p_id] = p_portalID;
	}
	unsigned GetAdjacentPortalQty(void) {return adjacentPortalQty;}
	unsigned * GetAdjacentPortalIDs(void) {return adjacentPortalIDs;}

	CSurfaceDecalList * GetDecalLists(void) {return decalLists;}
	LPAPITEXTURE * GetLightmaps(void) {return lightmaps;}

	void SetOpposingRoom(unsigned p_surfaceID, unsigned p_opposingRoomID)
	{
		opposingRoomIDs[p_surfaceID] = p_opposingRoomID;
	}
	unsigned GetOpposingRoom(unsigned p_surfaceID){return opposingRoomIDs[p_surfaceID];}

	inline CParticleSet & GetParticleSet(void) {return particleSet;}

	CParticleEmitter * AddNewParticleEmitter(void)
	{
		CLinkedListNode<CParticleEmitter> *node = particleEmitters.NewNode();
		node->particleSetRef = &particleSet;
		node->intRef = id;
		node->oldPosition.SetVec(-100000,-100000,-100000);
		return node;
	}

	void AddDetachedParticleEmitter(CLinkedListNode<CParticleEmitter> *node)
	{
		particleEmitters.Add(node);
		node->particleSetRef = &particleSet;
		node->intRef = id;
		// Don't change this - we need the old position to place
		//   particles in the right place
		//node->oldPosition.SetVec(-100000,-100000,-100000);
	}

	void DetachParticleEmitter(CLinkedListNode<CParticleEmitter> *node)
	{
		particleEmitters.Detach(node);
		node->particleSetRef = NULL;
		node->intRef = -1;
	}

	void AnimateParticles(COctree &p_octree, CPtlRoom *p_rooms, CModelVertex3D *p_vertices, DWORD p_elapsedTimeMS, DWORD p_currentTime);

	CSurfaceDecalListNode * AddDecal(CSurface3D *p_surface)
	{
		return decalLists[p_surface->GetDecalListID()].NewNode();
	}

	// Try to avoid calling this, a BSP algorithm would be faster
	bool PointInRoom(CVector3D &p_point, CModelVertex3D *p_vertices);

	bool SphereCollisionRoom(CModelVertex3D *p_vertices, CVector3D &p_pos, CVector3D &p_offset, float p_radius, float &p_timeOut, CVector3D &p_collisionNormalOut, CSurface3D **p_collisionSurfaceOut)
	{
		return SphereCollision(surfaces, surfaceQty, p_vertices, 0, p_pos, p_offset, p_radius, p_timeOut, p_collisionNormalOut, p_collisionSurfaceOut, &collisionVertexList);
	}

	inline void Render(CModelVertex3D *p_vertices, IRender *p_render, DWORD p_elapsedTimeMS, DWORD p_currentTime, unsigned p_mirrorDepth, unsigned int p_renderID)
	{
		Render(p_vertices, p_render, p_elapsedTimeMS, p_currentTime, p_mirrorDepth, p_renderID, NULL, NULL);
	}
	void Render(CModelVertex3D *p_vertices, IRender *p_render, DWORD p_elapsedTimeMS, DWORD p_currentTime, unsigned int p_mirrorDepth, unsigned int p_renderID, CLightSourceRefList *p_lightSourceList, CColorRGB *p_ambientColor);

	void RenderFloorTexture(IRender *p_render, CVector3D &p_pos, CVector3D *p_surfaceVertices, CAPITexture *p_texture, CColorRGB *p_modColor, BYTE p_modAlpha, CModelVertex3D *p_mapVertices);
};

#endif // !defined(AFX_PTLROOM_H__3274D7F7_2109_4D37_994C_7FB85783B304__INCLUDED_)
