// ptldoor.h: interface for the CPtlDoor class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_PTLDOOR_H__AA9DBDF7_5137_420C_844D_B7A313D65A76__INCLUDED_)
#define AFX_PTLDOOR_H__AA9DBDF7_5137_420C_844D_B7A313D65A76__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "..\math\Orient3D.h"
#include "..\api3d\Model3D.h"
#include "..\globals.h"
#include "..\api3d\APITexture.h"
#include "..\structs\Octree.h"

class CPtlDoor  
{
public:
	CPtlDoor();
	virtual ~CPtlDoor();
	void Destroy(void);

	// Values defining the two states of the door
	float openValue;
	float closedValue;

	// Target value is either openValue or closedValue
	// CurrentValue affects rendering and closed state
	float currentValue;
	float targetValue;

	float speedMS;	// How fast does currentValue change?

	unsigned int renderID;	// Tag to keep door from being rendered multiple
							//   times
	unsigned int collisionID;	// Tag to keep door from being tested for
								//   collision multiple times

	static unsigned int currentCollisionID;

	// ?? Need next and prev for door lists

	// Decal lists
	CSurfaceDecalList *decalLists;	// array, one list for each surface
									//   on the door
	// Lightmaps
	LPAPITEXTURE *lightmaps;		// Same for lightmaps (array)

	CModel3D	*modelRef;		// Model to use for rendering
								// This should only be a reference to a model that
								//   the main program deallocates elsewhere,
								//   since many doors could use the same model

	COrient3D	baseOrient;	// Orientation as closedValue (usually 0)
	COrient3D	currentOrient;	// Orient to render by (result of currentValue)
	// For now, assume closedValue and openValue are angles to rotate on the u axis,
	//   like a standard door moves

	COctCellRefList	cellsTouched;	// Which octree cells is this door touching? (for rendering)

	bool IsClosed(void) {return (currentValue == closedValue);}
	void Activate(void)
	{
		// Just toggle target value between open and closed
		if (targetValue == openValue)
			targetValue = closedValue;
		else
			targetValue = openValue;
	}
	void Animate(long p_elapsedTimeMS);

	void Render(IRender *p_render);

	inline CSurfaceDecalList * GetDecalLists(void) {return decalLists;}
	inline LPAPITEXTURE * GetLightmaps(void) {return lightmaps;}

	void SetupLists(CColorRGB p_lightmapColor, float p_texelPerUnit);

	void DetermineCellsTouched(COctree &p_octree);
};

class CPtlDoorList : public CLinkedList<CPtlDoor>
{
public:
	void DestroyDoors(void);
	void EmptyList(void)
	{
		DestroyDoors();
		CLinkedList<CPtlDoor>::EmptyList();
	}
	void Delete(CLinkedListNode<CPtlDoor> *node)
	{
		node->Destroy();
		CLinkedList<CPtlDoor>::Delete(node);
	}
};

class CPtlDoorListNode : public CLinkedListNode<CPtlDoor>
{
};

#endif // !defined(AFX_PTLDOOR_H__AA9DBDF7_5137_420C_844D_B7A313D65A76__INCLUDED_)
