// Orient3D.cpp: implementation of the COrient3D class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "Orient3D.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

COrient3D::COrient3D()
{

}

COrient3D::~COrient3D()
{

}

void COrient3D::RotateV(float p_angle, CVector3D &p_vector)
{
	Rotate(p_angle, p_vector.x, p_vector.y, p_vector.z);
}

void COrient3D::Rotate(float p_angle, float p_x, float p_y, float p_z)
{
    // This affects the orientation only, not the position
    CVector3D UA;     //Unit axis
    CVector3D RotateL;
    CVector3D RotateU;
    CVector3D RotateF;
    float RA;            //radian angle
    float cosRA, sinRA, OneMinusCosRA;
    COrient3D OrientTemp;

    // Unitize the incoming axis
    UA.SetVec(p_x, p_y, p_z);

    RA = p_angle * 3.14159266f / 180.0f;

    if (!UA.Unitize()) return;

    cosRA = float(cos(RA));
    sinRA = float(sin(RA));
    OneMinusCosRA = 1.0f - cosRA;

    // Got this one off the net!
    RotateL.SetVec(cosRA + UA.x * UA.x * OneMinusCosRA,
                    -UA.z * sinRA + UA.x * UA.y * OneMinusCosRA,
                    UA.y * sinRA + UA.x * UA.z * OneMinusCosRA);
    RotateU.SetVec(UA.z * sinRA + UA.y * UA.x * OneMinusCosRA,
                    cosRA + UA.y * UA.y * OneMinusCosRA,
                    -UA.x * sinRA + UA.y * UA.z * OneMinusCosRA);
    RotateF.SetVec(-UA.y * sinRA + UA.z * UA.x * OneMinusCosRA,
                    UA.x * sinRA + UA.z * UA.y * OneMinusCosRA,
                    cosRA + UA.z * UA.z * OneMinusCosRA);

    // This came out of the OpenGL manual knowing how the
    //   matrices are multiplied and what each element represents

    // Done this way to calculate all values before copying.
    //  If we did l.x =, l.y = etc. later calculations would be thrown
    // off
    SetOrientAxes(
        l.x * RotateL.x + l.y * RotateU.x + l.z * RotateF.x, 
        l.x * RotateL.y + l.y * RotateU.y + l.z * RotateF.y, 
        l.x * RotateL.z + l.y * RotateU.z + l.z * RotateF.z, 
        u.x * RotateL.x + u.y * RotateU.x + u.z * RotateF.x, 
        u.x * RotateL.y + u.y * RotateU.y + u.z * RotateF.y, 
        u.x * RotateL.z + u.y * RotateU.z + u.z * RotateF.z, 
        f.x * RotateL.x + f.y * RotateU.x + f.z * RotateF.x, 
        f.x * RotateL.y + f.y * RotateU.y + f.z * RotateF.y, 
        f.x * RotateL.z + f.y * RotateU.z + f.z * RotateF.z);
}

COrient3D COrient3D::TransformOrientToObjectSpace(COrient3D &p_orient)
{
	COrient3D out;
	CVector3D v1;

	out.f = CVector3D::Vector3D(l*p_orient.f, u*p_orient.f, f*p_orient.f);
	out.l = CVector3D::Vector3D(l*p_orient.l, u*p_orient.l, f*p_orient.l);
	out.u = CVector3D::Vector3D(l*p_orient.u, u*p_orient.u, f*p_orient.u);
	v1 = p_orient.p - p;
	out.p = CVector3D::Vector3D(l*v1, u*v1, f*v1);

	return out;
}


// This is used to take a relative orient and combine it with
//   a world orient to return a new world orient (used for determining
//   world positions and orientations along figure joints)
// p_orient is the relative orient, *this is the world orient
COrient3D COrient3D::TransformOrient(COrient3D &p_orient)
{
	COrient3D out;

	out.p = p_orient.p + p_orient.l*p.x + p_orient.u*p.y + p_orient.f*p.z;
	out.l = p_orient.l*l.x + p_orient.u*l.y + p_orient.f*l.z;
	out.u = p_orient.l*u.x + p_orient.u*u.y + p_orient.f*u.z;
	out.f = p_orient.l*f.x + p_orient.u*f.y + p_orient.f*f.z;

	return out;
}

COrient3D COrient3D::ReflectOnPlane(CVector3D &p_point, CVector3D &p_normal, float p_adjustment)
{
    COrient3D out;
    float dp;
    
    out.f = f + (p_normal * (-2.0f * (f*p_normal)));
    out.l = l + (p_normal * (-2.0f * (l*p_normal)));
    out.u = u + (p_normal * (-2.0f * (u*p_normal)));
    
    // The p_adjustment is to avoid artifacting for mirror reflections
    dp = (p - p_point) * p_normal;
	if (dp >= 0.0f)
		out.p = p + (p_normal * (-2.0f * (dp - p_adjustment)));
	else
		out.p = p + (p_normal * (-2.0f * (dp + p_adjustment)));
    
    return out;
}

