// Orient3D.h: interface for the COrient3D class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_ORIENT3D_H__FAAFE87C_189A_46B6_BE6A_E89AF6F9DB06__INCLUDED_)
#define AFX_ORIENT3D_H__FAAFE87C_189A_46B6_BE6A_E89AF6F9DB06__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "Vector3D.h"

class COrient3D  
{
public:
	COrient3D();
	virtual ~COrient3D();

	// l = u x f
	// u = f x l
	// f = l x u

	CVector3D l;
	CVector3D u;
	CVector3D f;
	CVector3D p;

	// Didn't need this
	// Orient3D & operator = (const COrient3D &p_orient)
	//{
	//	l = p_orient.l;
	//	u = p_orient.u;
	//	f = p_orient.f;
	//	p = p_orient.p;
		// Would this be faster?
		//CopyMemory((void *)p_orient.l, (void *)&l, 48);	// Would this work?

	//	return *this;
	//}

	void LoadIdentity()
	{
		float matrix[] = { 1.0f, 0.0f, 0.0f, 0.0f,
							0.0f, 1.0f, 0.0f, 0.0f,
							0.0f, 0.0f, 1.0f, 0.0f,
							0.0f, 0.0f, 0.0f, 1.0f};

		CopyMemory (&l.x, &matrix[0], 12);
		CopyMemory (&u.x, &matrix[4], 12);
		CopyMemory (&f.x, &matrix[8], 12);
		CopyMemory (&p.x, &matrix[12], 12);
	}

	void Rotate(float p_angle, float p_x, float p_y, float p_z);
	void RotateV(float p_angle, CVector3D &p_vector);

	CVector3D TransformVector(CVector3D &p_vector)
	{
		return (p + l*p_vector.x + u*p_vector.y + f*p_vector.z);
	}

	CVector3D TransformOffset(CVector3D &p_vector)
	{
		return (l*p_vector.x + u*p_vector.y + f*p_vector.z);
	}

	CVector3D TransformVectorToObjectSpace(CVector3D &p_vector)
	{
		return CVector3D::Vector3D(l*p_vector, u*p_vector, f*p_vector);
	}

	COrient3D TransformOrientToObjectSpace(COrient3D &p_orient);
	COrient3D TransformOrient(COrient3D &p_orient);

	COrient3D ReflectOnPlane(CVector3D &p_point, CVector3D &p_normal, float p_adjustment);

	void SetOrientAxes(float p_lx, float p_ly, float p_lz,
						float p_ux, float p_uy, float p_uz,
						float p_fx, float p_fy, float p_fz)
	{
		l.SetVec(p_lx, p_ly, p_lz);
		u.SetVec(p_ux, p_uy, p_uz);
		f.SetVec(p_fx, p_fy, p_fz);
	}

	inline void Normalize(void)
	{
		f = f * (1.0f / f.MgnTud());
		u = f.CrsProd(l);
		u = u * (1.0f / u.MgnTud());
		l = u.CrsProd(f);
	}
};

#endif // !defined(AFX_ORIENT3D_H__FAAFE87C_189A_46B6_BE6A_E89AF6F9DB06__INCLUDED_)
