// Vector3D.cpp: implementation of the CVector3D class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "Vector3D.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

#include "..\structs\LinkedList.h"

// Putting this here, since multiple locations could use a vector list
CLinkedList<CVector3D> CLinkedList<CVector3D>::DeletedList;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CVector3D::CVector3D()
{

}

CVector3D::CVector3D(float p_x, float p_y, float p_z)
{
	x = p_x;
	y = p_y;
	z = p_z;
}

//CVector3D::~CVector3D()
//{
//
//}

CVector3D CVector3D::CrsProd(CVector3D &p_vector)
{
	static CVector3D out;

	out.x = y * p_vector.z - z * p_vector.y;
	out.y = z * p_vector.x - x * p_vector.z;
	out.z = x * p_vector.y - y * p_vector.x;

	return out;
}

bool CVector3D::Unitize(void)
{
	static float mgntudsqr;
	static float invmgntud;

	mgntudsqr = x*x + y*y + z*z;
	if (mgntudsqr > 0.0f)
	{
		invmgntud = 1.0f / float(sqrt(mgntudsqr));
		x = x * invmgntud;
		y = y * invmgntud;
		z = z * invmgntud;

		return true;
	}
	else
		return false;
}

// Collision between sphere and the vector as a point in space
bool CVector3D::SphereCollision(CVector3D &p_pos, CVector3D &p_offset, float p_radius, float &p_timeOut, CVector3D &p_collisionNormalOut)
{

	float collisionTime;
	float tmp;
	CVector3D tmpPos;

	float a, b, c;
	float b1;
	float k1, k2, k;

	// Make sure sphere is moving in a direction that would make it collide
	tmp = (*this - p_pos) * p_offset;
	// Tolerance for up-close collisions and float truncation
//	if (tmp <= 0.0001) return false;
	if (tmp < 0.0f) return false;
	//if (fabs(tmp) < 0.00001f) return false;	// Almost parallel?

	// This one's a quadratic, sadly
	a = p_offset.MgnTudSqr();
	b = 2.0f * (p_offset * (p_pos - *this));
	c = (p_pos - *this).MgnTudSqr() - p_radius * p_radius;

	if (fabs(a) < 1e-11)
		return false;
	b1 = b * b - 4 * a * c;
	if (b1 < 0)
		return false;
	b1 = float(sqrt(b1));

	k1 = (-b + b1) / (2.0f * a);
	k2 = (-b - b1) / (2.0f * a);

	k = 100;
	// Tolerance for up-close collisions
	if ((k1 > -p_radius / 1000.0f) && (k1 <= 1.0f))
		k = k1;
	if ((k2 > -p_radius / 1000.0f) && (k2 <= 1.0f) && (k2 < k))
	    k = k2;

	collisionTime = k;

	// Tolerance for up-close collisions and float truncation
//	if ((collisionTime > -p_radius / 1000.0f) && (collisionTime <= 1.0f))
//	{
		if (collisionTime < p_timeOut)
		{
			p_timeOut = collisionTime;
        
			// Calculate the collision normal
			tmpPos = p_pos + (p_offset * p_timeOut) - *this;
			tmpPos.Unitize();
			p_collisionNormalOut = tmpPos;

			return true;
		}
//	}

	return false;
}
