// Octree.cpp: implementation of the COctree class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "Octree.h"
#include "..\api3d\IRender.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CLinkedList<COctCellRef> CLinkedList<COctCellRef>::DeletedList;

COctree::COctree()
{
	hasChildren = false;

	octFUR = NULL;
	octFLR = NULL;
	octFUL = NULL;
	octFLL = NULL;
	octBUR = NULL;
	octBLR = NULL;
	octBUL = NULL;
	octBLL = NULL;

	totalVolume = 0;
}

COctree::~COctree()
{
	Destroy();
}

void COctree::Destroy(void)
{
	if (octFUR) delete octFUR;
	if (octFLR) delete octFLR;
	if (octFUL) delete octFUL;
	if (octFLL) delete octFLL;
	if (octBUR) delete octBUR;
	if (octBLR) delete octBLR;
	if (octBUL) delete octBUL;
	if (octBLL) delete octBLL;

	octFUR = NULL;
	octFLR = NULL;
	octFUL = NULL;
	octFLL = NULL;
	octBUR = NULL;
	octBLR = NULL;
	octBUL = NULL;
	octBLL = NULL;

	hasChildren = false;
}

void COctree::Split(unsigned int p_qty, float p_minDimension)
{
	float splitQty;

	if (!hasChildren)
	if (((maxX-minX)>=p_minDimension)||((maxY-minY)>=p_minDimension)||((maxZ-minZ)>=p_minDimension))
	{
		// Just in case, initialize
		splitQty = 0.0f;

		// Figure out how we are going to split
		// Give priority to F, U and L cells
		if ((maxX-minX)>=p_minDimension)
		{
			if ((maxY-minY)>=p_minDimension)
			{
				if ((maxZ-minZ)>=p_minDimension)
				{
					// Split in all directions
					octFUL = new COctree;
					octFUR = new COctree;
					octFLL = new COctree;
					octFLR = new COctree;
					octBUL = new COctree;
					octBUR = new COctree;
					octBLL = new COctree;
					octBLR = new COctree;
					splitQty = 8.0f;
				}
				else
				{
					// Split only in x and y directions
					octFUL = new COctree;
					octFUR = new COctree;
					octFLL = new COctree;
					octFLR = new COctree;
					splitQty = 4.0f;
				}
			}
			else
			{
				if ((maxZ-minZ)>=p_minDimension)
				{
					// Split only in x and z directions
					octFUL = new COctree;
					octFUR = new COctree;
					octBUL = new COctree;
					octBUR = new COctree;
					splitQty = 4.0f;
				}
				else
				{
					// Split only in x direction
					octFUL = new COctree;
					octFUR = new COctree;
					splitQty = 2.0f;
				}
			}
		}
		else
		{
			if ((maxY-minY)>=p_minDimension)
			{
				if ((maxZ-minZ)>=p_minDimension)
				{
					// Split in y and z directions
					octFUL = new COctree;
					octFLL = new COctree;
					octBUL = new COctree;
					octBLL = new COctree;
					splitQty = 4.0f;
				}
				else
				{
					// Split only in y direction
					octFUL = new COctree;
					octFLL = new COctree;
					splitQty = 2.0f;
				}
			}
			else
			{
				if ((maxZ-minZ)>=p_minDimension)
				{
					// Split only in z direction
					octFUL = new COctree;
					octBUL = new COctree;
					splitQty = 2.0f;
				}
				else
				{
					// No splitting
					splitQty = 0.0f;
				}
			}
		}

		if (splitQty > 0) hasChildren = true;

		if (octFUL)
		{
			if ((maxX-minX)>=p_minDimension)
				octFUL->minX = midX;
			else
				octFUL->minX = minX;
			if ((maxY-minY)>=p_minDimension)
				octFUL->minY = midY;
			else
				octFUL->minY = minY;
			if ((maxZ-minZ)>=p_minDimension)
				octFUL->minZ = midZ;
			else
				octFUL->minZ = minZ;
			octFUL->maxX = maxX;
			octFUL->maxY = maxY;
			octFUL->maxZ = maxZ;
			octFUL->midX = (octFUL->minX + octFUL->maxX)/2.0f;
			octFUL->midY = (octFUL->minY + octFUL->maxY)/2.0f;
			octFUL->midZ = (octFUL->minZ + octFUL->maxZ)/2.0f;
		}

		if (octFUR)
		{
			octFUR->minX = minX;
			if ((maxY-minY)>=p_minDimension)
				octFUR->minY = midY;
			else
				octFUR->minY = minY;
			if ((maxZ-minZ)>=p_minDimension)
				octFUR->minZ = midZ;
			else
				octFUR->minZ = minZ;
			if ((maxX-minX)>=p_minDimension)
				octFUR->maxX = midX;
			else
				octFUR->maxX = maxX;
			octFUR->maxY = maxY;
			octFUR->maxZ = maxZ;
			octFUR->midX = (octFUR->minX + octFUR->maxX)/2.0f;
			octFUR->midY = (octFUR->minY + octFUR->maxY)/2.0f;
			octFUR->midZ = (octFUR->minZ + octFUR->maxZ)/2.0f;
		}

		if (octFLL)
		{
			if ((maxX-minX)>=p_minDimension)
				octFLL->minX = midX;
			else
				octFLL->minX = minX;
			octFLL->minY = minY;
			if ((maxZ-minZ)>=p_minDimension)
				octFLL->minZ = midZ;
			else
				octFLL->minZ = minZ;
			octFLL->maxX = maxX;
			if ((maxY-minY)>=p_minDimension)
				octFLL->maxY = midY;
			else
				octFLL->maxY = maxY;
			octFLL->maxZ = maxZ;
			octFLL->midX = (octFLL->minX + octFLL->maxX)/2.0f;
			octFLL->midY = (octFLL->minY + octFLL->maxY)/2.0f;
			octFLL->midZ = (octFLL->minZ + octFLL->maxZ)/2.0f;
		}

		if (octFLR)
		{
			octFLR->minX = minX;
			octFLR->minY = minY;
			if ((maxZ-minZ)>=p_minDimension)
				octFLR->minZ = midZ;
			else
				octFLR->minZ = minZ;
			if ((maxX-minX)>=p_minDimension)
				octFLR->maxX = midX;
			else
				octFLR->maxX = maxX;
			if ((maxY-minY)>=p_minDimension)
				octFLR->maxY = midY;
			else
				octFLR->maxY = maxY;
			octFLR->maxZ = maxZ;
			octFLR->midX = (octFLR->minX + octFLR->maxX)/2.0f;
			octFLR->midY = (octFLR->minY + octFLR->maxY)/2.0f;
			octFLR->midZ = (octFLR->minZ + octFLR->maxZ)/2.0f;
		}

		if (octBUL)
		{
			if ((maxX-minX)>=p_minDimension)
				octBUL->minX = midX;
			else
				octBUL->minX = minX;
			if ((maxY-minY)>=p_minDimension)
				octBUL->minY = midY;
			else
				octBUL->minY = minY;
			octBUL->minZ = minZ;
			octBUL->maxX = maxX;
			octBUL->maxY = maxY;
			if ((maxZ-minZ)>=p_minDimension)
				octBUL->maxZ = midZ;
			else
				octBUL->maxZ = maxZ;
			octBUL->midX = (octBUL->minX + octBUL->maxX)/2.0f;
			octBUL->midY = (octBUL->minY + octBUL->maxY)/2.0f;
			octBUL->midZ = (octBUL->minZ + octBUL->maxZ)/2.0f;
		}

		if (octBUR)
		{
			octBUR->minX = minX;
			if ((maxY-minY)>=p_minDimension)
				octBUR->minY = midY;
			else
				octBUR->minY = minY;
			octBUR->minZ = minZ;
			if ((maxX-minX)>=p_minDimension)
				octBUR->maxX = midX;
			else
				octBUR->maxX = maxX;
			octBUR->maxY = maxY;
			if ((maxZ-minZ)>=p_minDimension)
				octBUR->maxZ = midZ;
			else
				octBUR->maxZ = maxZ;
			octBUR->midX = (octBUR->minX + octBUR->maxX)/2.0f;
			octBUR->midY = (octBUR->minY + octBUR->maxY)/2.0f;
			octBUR->midZ = (octBUR->minZ + octBUR->maxZ)/2.0f;
		}

		if (octBLL)
		{
			if ((maxX-minX)>=p_minDimension)
				octBLL->minX = midX;
			else
				octBLL->minX = minX;
			octBLL->minY = minY;
			octBLL->minZ = minZ;
			octBLL->maxX = maxX;
			if ((maxY-minY)>=p_minDimension)
				octBLL->maxY = midY;
			else
				octBLL->maxY = maxY;
			if ((maxZ-minZ)>=p_minDimension)
				octBLL->maxZ = midZ;
			else
				octBLL->maxZ = maxZ;
			octBLL->midX = (octBLL->minX + octBLL->maxX)/2.0f;
			octBLL->midY = (octBLL->minY + octBLL->maxY)/2.0f;
			octBLL->midZ = (octBLL->minZ + octBLL->maxZ)/2.0f;
		}

		if (octBLR)
		{
			octBLR->minX = minX;
			octBLR->minY = minY;
			octBLR->minZ = minZ;
			if ((maxX-minX)>=p_minDimension)
				octBLR->maxX = midX;
			else
				octBLR->maxX = maxX;
			if ((maxY-minY)>=p_minDimension)
				octBLR->maxY = midY;
			else
				octBLR->maxY = maxY;
			if ((maxZ-minZ)>=p_minDimension)
				octBLR->maxZ = midZ;
			else
				octBLR->maxZ = maxZ;
			octBLR->midX = (octBLR->minX + octBLR->maxX)/2.0f;
			octBLR->midY = (octBLR->minY + octBLR->maxY)/2.0f;
			octBLR->midZ = (octBLR->minZ + octBLR->maxZ)/2.0f;
		}

		// Just an approximation, but it will work
		if (splitQty > 0)
		{
			if (octFUL) octFUL->totalVolume = totalVolume/splitQty;
			if (octFUR) octFUR->totalVolume = totalVolume/splitQty;
			if (octFLL) octFLL->totalVolume = totalVolume/splitQty;
			if (octFLR) octFLR->totalVolume = totalVolume/splitQty;
			if (octBUL) octBUL->totalVolume = totalVolume/splitQty;
			if (octBUR) octBUR->totalVolume = totalVolume/splitQty;
			if (octBLL) octBLL->totalVolume = totalVolume/splitQty;
			if (octBLR) octBLR->totalVolume = totalVolume/splitQty;
		}
	}

	if (p_qty > 1)
	{
		if (octFUR)
			octFUR->Split(p_qty-1, p_minDimension);
		if (octFLR)
			octFLR->Split(p_qty-1, p_minDimension);
		if (octFUL)
			octFUL->Split(p_qty-1, p_minDimension);
		if (octFLL)
			octFLL->Split(p_qty-1, p_minDimension);
		if (octBUR)
			octBUR->Split(p_qty-1, p_minDimension);
		if (octBLR)
			octBLR->Split(p_qty-1, p_minDimension);
		if (octBUL)
			octBUL->Split(p_qty-1, p_minDimension);
		if (octBLL)
			octBLL->Split(p_qty-1, p_minDimension);
	}
}

COctree * COctree::PointIn(float p_x, float p_y, float p_z)
{
	if (p_x >= minX)
	if (p_x < maxX)
	if (p_y >= minY)
	if (p_y < maxY)
	if (p_z >= minZ)
	if (p_z < maxZ)
	{
		if (!hasChildren)
		{
			return this;
		}
		else
		{
			// Parse children (avoid recursive call stack)
			COctree *cell = this;
			while (cell)
			{
				if (cell->hasChildren)
				{
					if (p_x >= cell->midX)
					{
						if (p_y >= cell->midY)
						{
							if (p_z >= cell->midZ)
							{
								if (cell->octFUL)
									cell = cell->octFUL;
							}
							else
							{
								if (cell->octBUL)
									cell = cell->octBUL;
								else
									cell = cell->octFUL;	// no z split
							}
						}
						else
						{
							if (p_z >= cell->midZ)
							{
								if (cell->octFLL)
									cell = cell->octFLL;
								else
									cell = cell->octFUL;	// no y split
							}
							else
							{
								if (cell->octBLL)
									cell = cell->octBLL;
								else
									if (cell->octBUL)
										cell = cell->octBUL;
									else
										if (cell->octFLL)
											cell = cell->octFLL;
										else
											cell = cell->octFUL;
							}
						}
					}
					else
					{
						if (p_y >= cell->midY)
						{
							if (p_z >= cell->midZ)
							{
								if (cell->octFUR)
									cell = cell->octFUR;
								else
									cell = cell->octFUL;
							}
							else
							{
								if (cell->octBUR)
									cell = cell->octBUR;
								else
									if (cell->octBUL)
										cell = cell->octBUL;
									else
										if (cell->octFUR)
											cell = cell->octFUR;
										else
											cell = cell->octFUL;
							}
						}
						else
						{
							if (p_z >= cell->midZ)
							{
								if (cell->octFLR)
									cell = cell->octFLR;
								else
									if (cell->octFUR)
										cell = cell->octFUR;
									else
										if (cell->octFLL)
											cell = cell->octFLL;
										else
											cell = cell->octFUL;
							}		
							else
							{
								if (cell->octBLR)
									cell = cell->octBLR;
								else
									if (cell->octBUR)
										cell = cell->octBUR;
									else
										if (cell->octBLL)
											cell = cell->octBLL;
										else
											if (cell->octFLR)
												cell = cell->octFLR;
											else
												cell = cell->octFUL;
							}
						}
					}
				}
				else
					return cell;

			}	// shouldn't get an infinite loop (I hope)
		}
	}

	return NULL;
}

void COctree::DebugParse(char *p_pad)
{
	TRACE("%sBounds: (%g,%g,%g)  (%g,%g,%g)  mid:(%g,%g,%g)\n", p_pad, minX, minY, minZ, maxX, maxY, maxZ, midX, midY, midZ);
	char pad[256];
	if (hasChildren)
	{
		sprintf(pad, "%s>>-", p_pad);
		if (octFUL)
			octFUL->DebugParse(pad);
		if (octFUR)
			octFUR->DebugParse(pad);
		if (octFLL)
			octFLL->DebugParse(pad);
		if (octFLR)
			octFLR->DebugParse(pad);
		if (octBUL)
			octBUL->DebugParse(pad);
		if (octBUR)
			octBUR->DebugParse(pad);
		if (octBLL)
			octBLL->DebugParse(pad);
		if (octBLR)
			octBLR->DebugParse(pad);
	}
	else
	{
		TRACE("%sObjects in list 0: %5li, Objects in list 1: %5li\n", p_pad, lpVoidList.GetNodeCount(), lpVoidList1.GetNodeCount());
	}
}

void COctree::CellsTouched(float p_minX, float p_minY, float p_minZ, float p_maxX, float p_maxY, float p_maxZ, COctCellRefList &p_cellList)
{
	p_cellList.EmptyList();

	CellsTouchedRecurse(p_minX, p_minY, p_minZ, p_maxX, p_maxY, p_maxZ, p_cellList);
}

void COctree::CellsTouchedRecurse(float p_minX, float p_minY, float p_minZ, float p_maxX, float p_maxY, float p_maxZ, COctCellRefList &p_cellList)
{
	if (p_minX >= maxX) return;
	if (p_maxX < minX) return;
	if (p_minY >= maxY) return;
	if (p_maxY < minY) return;
	if (p_minZ >= maxZ) return;
	if (p_maxZ < minZ) return;

	if (hasChildren)
	{
		if (octFUL)
			octFUL->CellsTouchedRecurse(p_minX, p_minY, p_minZ, p_maxX, p_maxY, p_maxZ, p_cellList);
		if (octFUR)
			octFUR->CellsTouchedRecurse(p_minX, p_minY, p_minZ, p_maxX, p_maxY, p_maxZ, p_cellList);
		if (octFLL)
			octFLL->CellsTouchedRecurse(p_minX, p_minY, p_minZ, p_maxX, p_maxY, p_maxZ, p_cellList);
		if (octFLR)
			octFLR->CellsTouchedRecurse(p_minX, p_minY, p_minZ, p_maxX, p_maxY, p_maxZ, p_cellList);
		if (octBUL)
			octBUL->CellsTouchedRecurse(p_minX, p_minY, p_minZ, p_maxX, p_maxY, p_maxZ, p_cellList);
		if (octBUR)
			octBUR->CellsTouchedRecurse(p_minX, p_minY, p_minZ, p_maxX, p_maxY, p_maxZ, p_cellList);
		if (octBLL)
			octBLL->CellsTouchedRecurse(p_minX, p_minY, p_minZ, p_maxX, p_maxY, p_maxZ, p_cellList);
		if (octBLR)
			octBLR->CellsTouchedRecurse(p_minX, p_minY, p_minZ, p_maxX, p_maxY, p_maxZ, p_cellList);
	}
	else
	{
		CLinkedListNode<COctCellRef> *node = p_cellList.NewNode();
		if (node)
		{
			node->cellRef = this;
		}
	}
}

void COctree::SplitToVolume(float p_minX, float p_minY, float p_minZ, float p_maxX, float p_maxY, float p_maxZ, float p_targetVolume, float p_minDimension)
{
	if (p_targetVolume < 0.001f) return;

	// Make sure some portion of the bounds are in this cell
	if (p_minX >= maxX) return;
	if (p_maxX < minX) return;
	if (p_minY >= maxY) return;
	if (p_maxY < minY) return;
	if (p_minZ >= maxZ) return;
	if (p_maxZ < minZ) return;

	if (!hasChildren)
	{
		// Test for sent bounds to take up target volume in this cell
		// If it does, perform a split

		// Wanted to use 'min' and 'max', but those names are taken
		float loX, loY, loZ;
		float hiX, hiY, hiZ;

		loX = minX; loY = minY, loZ = minZ;
		hiX = maxX; hiY = maxY, hiZ = maxZ;
		if (loX < p_minX) loX = p_minX;
		if (loY < p_minY) loY = p_minY;
		if (loZ < p_minZ) loZ = p_minZ;
		if (hiX > p_maxX) hiX = p_maxX;
		if (hiY > p_maxY) hiY = p_maxY;
		if (hiZ > p_maxZ) hiZ = p_maxZ;

		// Increment the volume in this cell
		// Make sure total volume doesn't go higher than the actual
		//   size of the cell
		totalVolume += ((hiX-loX)*(hiY-loY)*(hiZ-loZ));
		if (totalVolume > ((maxX-minX)*(maxY-minY)*(maxZ-minZ)))
			totalVolume = (maxX-minX)*(maxY-minY)*(maxZ-minZ);

		// hi and lo x,y,z how describe the volume that exists in this cell
		if (totalVolume > p_targetVolume)
			Split(1, p_minDimension);
	}

	if (hasChildren)
	{
		if (octFUL)
			octFUL->SplitToVolume(p_minX, p_minY, p_minZ, p_maxX, p_maxY, p_maxZ, p_targetVolume, p_minDimension);
		if (octFUR)
			octFUR->SplitToVolume(p_minX, p_minY, p_minZ, p_maxX, p_maxY, p_maxZ, p_targetVolume, p_minDimension);
		if (octFLL)
			octFLL->SplitToVolume(p_minX, p_minY, p_minZ, p_maxX, p_maxY, p_maxZ, p_targetVolume, p_minDimension);
		if (octFLR)
			octFLR->SplitToVolume(p_minX, p_minY, p_minZ, p_maxX, p_maxY, p_maxZ, p_targetVolume, p_minDimension);
		if (octBUL)
			octBUL->SplitToVolume(p_minX, p_minY, p_minZ, p_maxX, p_maxY, p_maxZ, p_targetVolume, p_minDimension);
		if (octBUR)
			octBUR->SplitToVolume(p_minX, p_minY, p_minZ, p_maxX, p_maxY, p_maxZ, p_targetVolume, p_minDimension);
		if (octBLL)
			octBLL->SplitToVolume(p_minX, p_minY, p_minZ, p_maxX, p_maxY, p_maxZ, p_targetVolume, p_minDimension);
		if (octBLR)
			octBLR->SplitToVolume(p_minX, p_minY, p_minZ, p_maxX, p_maxY, p_maxZ, p_targetVolume, p_minDimension);
	}
}

void COctree::Render(void *p_render)
{
	IRender *render = (IRender *)p_render;

	if (hasChildren)
	{
		if (octFUL) octFUL->Render(p_render);
		if (octFUR) octFUR->Render(p_render);
		if (octFLL) octFLL->Render(p_render);
		if (octFLR) octFLR->Render(p_render);
		if (octBUL) octBUL->Render(p_render);
		if (octBUR) octBUR->Render(p_render);
		if (octBLL) octBLL->Render(p_render);
		if (octBLR) octBLR->Render(p_render);
	}
	else
		render->RenderWireframeCube(minX, minY, minZ, maxX, maxY, maxZ);
}
