// Octree.h: interface for the COctree class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_OCTREE_H__C4705075_503D_46A1_A187_29EC61D343C2__INCLUDED_)
#define AFX_OCTREE_H__C4705075_503D_46A1_A187_29EC61D343C2__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "LinkedList.h"
#include "..\math\Vector3D.h"

class COctree;
class COctCellRef
{
public:
	COctree *cellRef;
};

class COctCellRefList : public CLinkedList<COctCellRef>
{
};

class COctree  
{
public:
	COctree();
	virtual ~COctree();
	void Destroy();

private:
	// NOTE: bounds are [min, max)
	float minX, minY, minZ;	// max's must be numerically large than min's
	float maxX, maxY, maxZ;

	float midX, midY, midZ;	// midpoints stored for faster compares

	float totalVolume;		// Total volume submitted for this cell

	bool hasChildren;		// Is this cell split?

	// Abbreviation is ZYX
	//  for Z, F is +, B is -; for Y, U is +, L is -; for X, L is +, R is -
	// "Front upper left", "Back lower right", etc.
	COctree *octFLL;	// x, -y, z
	COctree *octFLR;	// -x, -y, z
	COctree *octFUL;	// x, y, z
	COctree *octFUR;	// -x, y, z
	COctree *octBLL;	// x, -y, -z
	COctree *octBLR;	// -x, -y, -z
	COctree *octBUL;	// x, y, -z
	COctree *octBUR;	// -x, y, -z

public:
	CLinkedList<CLPVoid> lpVoidList;	// objects
	CLinkedList<CLPVoid> lpVoidList1;	// doors
	CLinkedList<CLPVoid> lpVoidList2;	// lights

	CLinkedList<CInteger> intList;		// room IDs
	CLinkedList<CInteger> intList1;

	inline void Split(void) {Split(1, 0);}
	inline void Split(unsigned int p_qty) {Split(p_qty, 0);}
	void Split(unsigned int p_qty, float p_minDimension);

	inline void SplitToVolume(CVector3D &p_min, CVector3D &p_max, float p_targetVolume)
	{
		SplitToVolume(p_min.x, p_min.y, p_min.z, p_max.x, p_max.y, p_max.z, p_targetVolume, 0.0f);
	}
	inline void SplitToVolume(CVector3D &p_min, CVector3D &p_max, float p_targetVolume, float p_minDimension)
	{
		SplitToVolume(p_min.x, p_min.y, p_min.z, p_max.x, p_max.y, p_max.z, p_targetVolume, p_minDimension);
	}
	inline void SplitToVolume(float p_minX, float p_minY, float p_minZ, float p_maxX, float p_maxY, float p_maxZ, float p_targetVolume)
	{
		SplitToVolume(p_minX, p_minY, p_minZ, p_maxX, p_maxY, p_maxZ, p_targetVolume, 0.0f);
	}
	void SplitToVolume(float p_minX, float p_minY, float p_minZ, float p_maxX, float p_maxY, float p_maxZ, float p_targetVolume, float p_minDimension);

	inline COctree * PointIn(CVector3D &p_vector)
	{
		return PointIn(p_vector.x, p_vector.y, p_vector.z);
	}
	COctree * PointIn(float p_x, float p_y, float p_z);
	void SetBounds(CVector3D &p_minPoint, CVector3D &p_maxPoint)
	{
		minX = p_minPoint.x;
		maxX = p_maxPoint.x;
		minY = p_minPoint.y;
		maxY = p_maxPoint.y;
		minZ = p_minPoint.z;
		maxZ = p_maxPoint.z;
		midX = (minX + maxX)/2.0f;
		midY = (minY + maxY)/2.0f;
		midZ = (minZ + maxZ)/2.0f;
	}
	void SetBounds(float p_minX, float p_minY, float p_minZ, float p_maxX, float p_maxY, float p_maxZ)
	{
		minX = p_minX;
		maxX = p_maxX;
		minY = p_minY;
		maxY = p_maxY;
		minZ = p_minZ;
		maxZ = p_maxZ;
		midX = (p_minX + p_maxX)/2.0f;
		midY = (p_minY + p_maxY)/2.0f;
		midZ = (p_minZ + p_maxZ)/2.0f;
	}

	void GetBounds(float &p_minXOut, float &p_minYOut, float &p_minZOut, float &p_maxXOut, float &p_maxYOut, float &p_maxZOut)
	{
		p_minXOut = minX;
		p_maxXOut = maxX;
		p_minYOut = minY;
		p_maxYOut = maxY;
		p_minZOut = minZ;
		p_maxZOut = maxZ;
	}

	void GetChildren(COctree **p_FUL, COctree **p_FUR, COctree **p_FLL, COctree **p_FLR,
						COctree **p_BUL, COctree **p_BUR, COctree **p_BLL, COctree **p_BLR)
	{
		*p_FUL = octFUL;
		*p_FUR = octFUR;
		*p_FLL = octFLL;
		*p_FLR = octFLR;
		*p_BUL = octBUL;
		*p_BUR = octBUR;
		*p_BLL = octBLL;
		*p_BLR = octBLR;
	}

	inline COctree * GetFUL(void) {return octFUL;}
	inline COctree * GetFUR(void) {return octFUR;}
	inline COctree * GetFLL(void) {return octFLL;}
	inline COctree * GetFLR(void) {return octFLR;}
	inline COctree * GetBUL(void) {return octBUL;}
	inline COctree * GetBUR(void) {return octBUR;}
	inline COctree * GetBLL(void) {return octBLL;}
	inline COctree * GetBLR(void) {return octBLR;}

	void DebugParse(char *p_pad);
	void Render(void *p_render);

	inline void CellsTouched(CVector3D &p_minPoint, CVector3D &p_maxPoint, COctCellRefList &p_cellList)
	{
		CellsTouched(p_minPoint.x, p_minPoint.y, p_minPoint.z, p_maxPoint.x, p_maxPoint.y, p_maxPoint.z, p_cellList);
	}
	void CellsTouched(float p_minX, float p_minY, float p_minZ, float p_maxX, float p_maxY, float p_maxZ, COctCellRefList &p_cellList);
private:
	void CellsTouchedRecurse(float p_minX, float p_minY, float p_minZ, float p_maxX, float p_maxY, float p_maxZ, COctCellRefList &p_cellList);
};

#endif // !defined(AFX_OCTREE_H__C4705075_503D_46A1_A187_29EC61D343C2__INCLUDED_)
