// QuadTree.cpp: implementation of the CQuadTree class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "QuadTree.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CLinkedList<CQuadCellRef> CLinkedList<CQuadCellRef>::DeletedList;

CQuadTree::CQuadTree()
{
	quadLL = NULL;
	quadLR = NULL;
	quadUL = NULL;
	quadUR = NULL;
}

CQuadTree::~CQuadTree()
{
	Destroy();
}

void CQuadTree::Destroy(void)
{
	if (quadLL) delete quadLL;
	if (quadLR) delete quadLR;
	if (quadUL) delete quadUL;
	if (quadUR) delete quadUR;

	quadLL = NULL;
	quadLR = NULL;
	quadUL = NULL;
	quadUR = NULL;
}

void CQuadTree::Split(unsigned int p_qty)
{
	if (!quadLL)
	{
		quadLL = new CQuadTree;
		quadLL->maxX = maxX;
		quadLL->minX = midX;
		quadLL->maxY = midY;
		quadLL->minY = minY;
		quadLL->midX = (quadLL->minX + quadLL->maxX)/2.0f;
		quadLL->midY = (quadLL->minY + quadLL->maxY)/2.0f;

		quadLR = new CQuadTree;
		quadLR->maxX = midX;
		quadLR->minX = minX;
		quadLR->maxY = midY;
		quadLR->minY = minY;
		quadLR->midX = (quadLR->minX + quadLR->maxX)/2.0f;
		quadLR->midY = (quadLR->minY + quadLR->maxY)/2.0f;

		quadUL = new CQuadTree;
		quadUL->maxX = maxX;
		quadUL->minX = midX;
		quadUL->maxY = maxY;
		quadUL->minY = midY;
		quadUL->midX = (quadUL->minX + quadUL->maxX)/2.0f;
		quadUL->midY = (quadUL->minY + quadUL->maxY)/2.0f;

		quadUR = new CQuadTree;
		quadUR->maxX = midX;
		quadUR->minX = minX;
		quadUR->maxY = maxY;
		quadUR->minY = midY;		
		quadUR->midX = (quadUR->minX + quadUR->maxX)/2.0f;
		quadUR->midY = (quadUR->minY + quadUR->maxY)/2.0f;

		if (p_qty > 1)
		{
			quadLL->Split(p_qty-1);
			quadLR->Split(p_qty-1);
			quadUL->Split(p_qty-1);
			quadUR->Split(p_qty-1);
		}
	}
}

CQuadTree * CQuadTree::PointIn(float p_x, float p_y)
{
	if (p_x >= minX)
	if (p_x < maxX)
	if (p_y >= minY)
	if (p_y < maxY)
	{
		if (!quadLL)
		{
			return this;
		}
		else
		{
			// Parse children (avoid recursive call stack)
			CQuadTree *cell = this;
			while (cell)
			{
				if (cell->quadLL)
				{
					if (p_x >= cell->midX)
					{
						if (p_y >= cell->midY)
						{
							cell =  cell->quadUL;
						}
						else
						{
							cell =  cell->quadLL;
						}
					}
					else
					{
						if (p_y >= cell->midY)
						{
							cell =  cell->quadUR;
						}
						else
						{
							cell =  cell->quadLR;
						}
					}
				}
				else
					return cell;

			}	// shouldn't get an infinite loop (I hope)
		}
	}

	return NULL;
}

void CQuadTree::DebugParse(char *p_pad)
{
	TRACE("%sBounds: (%g,%g)  (%g,%g)  mid:(%g,%g)\n", p_pad, minX, minY, maxX, maxY, midX, midY);
	if (quadLL)
	{
		char pad[50];
		sprintf(pad, "%s>>-", p_pad);

		quadLL->DebugParse(pad);
		quadLR->DebugParse(pad);
		quadUL->DebugParse(pad);
		quadUR->DebugParse(pad);
	}
}

void CQuadTree::CellsTouched(float p_minX, float p_minY, float p_maxX, float p_maxY, CQuadCellRefList &p_cellList)
{
	p_cellList.EmptyList();

	CellsTouchedRecurse(p_minX, p_minY, p_maxX, p_maxY, p_cellList);
}

void CQuadTree::CellsTouchedRecurse(float p_minX, float p_minY, float p_maxX, float p_maxY, CQuadCellRefList &p_cellList)
{
	if (p_minX >= maxX) return;
	if (p_maxX < minX) return;
	if (p_minY >= maxY) return;
	if (p_maxY < minY) return;

	if (quadLL)
	{
		quadLL->CellsTouchedRecurse(p_minX, p_minY, p_maxX, p_maxY, p_cellList);
		quadLR->CellsTouchedRecurse(p_minX, p_minY, p_maxX, p_maxY, p_cellList);
		quadUL->CellsTouchedRecurse(p_minX, p_minY, p_maxX, p_maxY, p_cellList);
		quadUR->CellsTouchedRecurse(p_minX, p_minY, p_maxX, p_maxY, p_cellList);
	}
	else
	{
		CLinkedListNode<CQuadCellRef> *node = p_cellList.NewNode();
		if (node)
		{
			node->cellRef = this;
		}
	}
}
