// QuadTree.h: interface for the CQuadTree class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_QUADTREE_H__4BF111FC_B367_4024_A8A0_8D3955E63F6F__INCLUDED_)
#define AFX_QUADTREE_H__4BF111FC_B367_4024_A8A0_8D3955E63F6F__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "LinkedList.h"

class CQuadTree;
class CQuadCellRef
{
public:
	CQuadTree *cellRef;
};

class CQuadCellRefList : public CLinkedList<CQuadCellRef>
{
};

class CQuadTree  
{
public:
	CQuadTree();
	virtual ~CQuadTree();
	void Destroy();

private:
	// NOTE: bounds are [min, max)
	float minX, minY;	// max's must be numerically large than min's
	float maxX, maxY;

	float midX, midY;	// midpoints stored for faster compares

	CQuadTree *quadLL;	// x, -y
	CQuadTree *quadLR;	// -x, -y
	CQuadTree *quadUL;	// x, y
	CQuadTree *quadUR;	// -x, y


public:
	CLinkedList<CLPVoid> lpVoidList;
	CLinkedList<CInteger> intList;

	inline void Split(void) {Split(1);}
	void Split(unsigned int p_qty);

	CQuadTree * PointIn(float p_x, float p_y);
	void SetBounds(float p_minX, float p_minY, float p_maxX, float p_maxY)
	{
		minX = p_minX;
		maxX = p_maxX;
		minY = p_minY;
		maxY = p_maxY;
		midX = (p_minX + p_maxX)/2.0f;
		midY = (p_minY + p_maxY)/2.0f;
	}

	void GetBounds(float &p_minXOut, float &p_minYOut, float &p_maxXOut, float &p_maxYOut)
	{
		p_minXOut = minX;
		p_maxXOut = maxX;
		p_minYOut = minY;
		p_maxYOut = maxY;
	}

	void GetChildren(CQuadTree **p_UL, CQuadTree **p_UR, CQuadTree **p_LL, CQuadTree **p_LR)
	{
		*p_UL = quadUL;
		*p_UR = quadUR;
		*p_LL = quadLL;
		*p_LR = quadLR;
	}

	inline CQuadTree * GetUL(void) {return quadUL;}
	inline CQuadTree * GetUR(void) {return quadUR;}
	inline CQuadTree * GetLL(void) {return quadLL;}
	inline CQuadTree * GetLR(void) {return quadLR;}

	void DebugParse(char *p_pad);

	void CellsTouched(float p_minX, float p_minY, float p_maxX, float p_maxY, CQuadCellRefList &p_cellList);
private:
	void CellsTouchedRecurse(float p_minX, float p_minY, float p_maxX, float p_maxY, CQuadCellRefList &p_cellList);
};

#endif // !defined(AFX_QUADTREE_H__4BF111FC_B367_4024_A8A0_8D3955E63F6F__INCLUDED_)
